//==================================================================================
//
// OpenDCC
//
// Copyright (c) 2006-2011 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//----------------------------------------------------------------------------------
//
// file:      bidib_client_if.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2011-02-28 V0.01 kw start
//
//----------------------------------------------------------------------------------
//
// purpose:   client for BiDiB (standard AVR)
//
// content:   routines for RS485 Interface
//            see also following references:
//            -> www.bidib.org
// here:      RX fifo and TX buffer
//            Control of line direction
//
//==================================================================================
// upstream interface (howto):
//
// to send something on bidib:
//   a) create message to be sent
//      (this must include header and data, the crc checksum
//      is calculated here)
//   b) copy this message to bidib_tx_fifo
//
// to receive a message:
//   a) all incoming messages are forwarded to Rx_Fifo
//   b) bidib_client_if collects the incomming message and handles
//      slot usage and message tracking
//   c) see bidib_client_parser.c for processing of messages
//
// init:
//   a) be sure to set up timer correctly
//   b) define the usart and direction port
//   c) call init_bidib_client_if();
//
// run_time:
//   a) periodically call run_bidib_client() / or appropriate task;
//      this updates states
//
//==================================================================================
// used hardware resources (see bidib_client_hardware.h):
//
//   a) UART:   may be the only UART, UART0 or UART1, see below and 
//              definition header (see also config.h)
//   b) TIMER2: this must run at something between 0.2us to 2us clockrate.
//              The timer must run in wrap around mode, we use OCR2
//              and the corresponding compare interrupt
//              Note: in PWM Modes OCR is double buffered - we cant use these modes
//   c) PORT:   a port, which controls the direction setting.
//
//==================================================================================
// How it works:
//
//  ISR:       | content
//-------------|-----------------------------------------
//  RS485_RX   | get data and put to fifo
//             | check for call to our slot_id
//             | if called: enable timer int (20us)
//             |
//  TIMER      | disable timer int
//             | check tx_request
//             | if yes: prepare message
//             |         enable tx
//             |         enable rs485_tx
//             |
//  RS485_TX   | calculate crc
//             | transmit data
//             | if last data, disable tx
//             |
//  TX_NO_DATA | switch rs485 to rx
//             |
//==================================================================================

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <util/parity.h>
#include <string.h>

#include "config.h"                 // general structures and definitions

#include "hardware.h"               // general structures and definitions
#include "cortos.h"                 // make consumer ready
#include "prbs8.h"                  // random number for backoff
#include "bidib_client_hardware.h"
#include "bidib_client_if.h"
#include "bidib_client_parser.h"    // unique_id
#include "bidib_messages.h"

#include "crc_8bit.h"


// Some Triggermark - for debugging with a scope
// 
#define TRIGA    { LED_MSG_ON(); _mydelay_us(5); LED_MSG_OFF(); }
#define TRIGB    { LED_S88_ON(); _mydelay_us(5); LED_S88_OFF(); }
#define TRIGC    { LED_BIDIB_ON(); _mydelay_us(5); LED_BIDIB_OFF(); }


//==================================================================================
// check and expand hardware definitions
//==================================================================================


#ifndef BIDIB_UART
   #warning SEVERE no UART defined, see bidib_client_hardware.h
#endif

#if (BIDIB_UART == UART)
    #define my_UDR      UDR
    #define my_UCSRA    UCSRA
        #define my_RXC      RXC
        #define my_TXC      TXC
        #define my_U2X      U2X
        #define my_FE       FE
        #define my_DOR      DOR
        #define my_UDRE     UDRE
    #define my_UCSRB    UCSRB
        #define my_RXCIE    RXCIE
        #define my_TXCIE    TXCIE
        #define my_UDRIE    UDRIE
        #define my_RXEN     RXEN
        #define my_TXEN     TXEN
        #define my_UCSZ2    UCSZ2
        #define my_RXB8     RXB8
        #define my_TXB8     TXB8
    #define my_UCSRC    UCSRC
        #ifdef URSEL
        #define my_URSEL    URSEL
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL
        #define my_UPM1     UPM1
        #define my_UPM0     UPM0
        #define my_USBS     USBS
        #define my_UCSZ1    UCSZ1
        #define my_UCSZ0    UCSZ0
        #define my_UCPOL    UCPOL
    #define my_UBRRH    UBRRH
    #define my_UBRRL    UBRRL

#elif (BIDIB_UART == UART0)
    #define my_UDR      UDR0
    #define my_UCSRA    UCSR0A
        #define my_RXC      RXC0
        #define my_TXC      TXC0
        #define my_U2X      U2X0
        #define my_FE       FE0
        #define my_DOR      DOR0
        #define my_UDRE     UDRE0
    #define my_UCSRB    UCSR0B
        #define my_RXCIE    RXCIE0
        #define my_TXCIE    TXCIE0
        #define my_UDRIE    UDRIE0
        #define my_RXEN     RXEN0
        #define my_TXEN     TXEN0
        #define my_UCSZ2    UCSZ02
        #define my_RXB8     RXB80
        #define my_TXB8     TXB80
    #define my_UCSRC    UCSR0C
        #ifdef URSEL0
         #define my_URSEL    URSEL0
        #else
         #define my_URSEL    7          // create a dummy
        #endif
        #ifdef UMSEL00
         #define my_UMSEL    UMSEL00
        #else
         #define my_UMSEL    UMSEL0
        #endif
        #define my_UPM1     UPM01
        #define my_UPM0     UPM00
        #define my_USBS     USBS0
        #define my_UCSZ1    UCSZ01
        #define my_UCSZ0    UCSZ00
        #define my_UCPOL    UCPOL0
    #define my_UBRRH    UBRR0H
    #define my_UBRRL    UBRR0L

#elif (BIDIB_UART == UART1)
    #define my_UDR      UDR1
    #define my_UCSRA    UCSR1A
        #define my_RXC      RXC1
        #define my_TXC      TXC1
        #define my_U2X      U2X1
        #define my_FE       FE1
        #define my_DOR      DOR1
        #define my_UDRE     UDRE1
    #define my_UCSRB    UCSR1B
        #define my_RXCIE    RXCIE1
        #define my_TXCIE    TXCIE1
        #define my_UDRIE    UDRIE1
        #define my_RXEN     RXEN1
        #define my_TXEN     TXEN1
        #define my_UCSZ2    UCSZ12
        #define my_RXB8     RXB81
        #define my_TXB8     TXB81
    #define my_UCSRC    UCSR1C
        #ifdef URSEL1
        #define my_URSEL    URSEL1
        #else
        #define my_URSEL    7          // create a dummy
        #endif
        #define my_UMSEL    UMSEL1
        #define my_UPM1     UPM11
        #define my_UPM0     UPM10
        #define my_USBS     USBS1
        #define my_UCSZ1    UCSZ11
        #define my_UCSZ0    UCSZ10
        #define my_UCPOL    UCPOL1
    #define my_UBRRH    UBRR1H
    #define my_UBRRL    UBRR1L
#else
  #warning (severe) wrong BIDIB_UART !
#endif // BIDIB_UART


//=================================================================================
//
// Data Section
//
//=================================================================================

// general:

unsigned char my_bidib_node_addr;               // our assigned local address, 
                                                // where we are allowed to send
                                                // this address is assigned during logon

unsigned char g_bidib_backoff;                  // backoff during logon


// out-fifo handling

#define BIDIB_TX_BUF_HEADROOM   32                  // abzuprfender, notwendiger freier Platz, 
                                                    // damit Spontan bzw. feature gesendet werden kann
#define BIDIB_TX_BUF_SIZE       64                  // fifo for outgoing messages
                                                    // MUST be 2^^n !!!

unsigned char  bidib_tx_buf[BIDIB_TX_BUF_SIZE];

volatile unsigned char bidib_tx_buf_read;           // points to next read
volatile unsigned char bidib_tx_buf_write;          // points to next write
                                                    // rd = wr: queue empty
                                                    // rd = wr + 1: queue full

volatile unsigned char bidib_tx_fill;               // aktueller Stand noch nicht 'geladener' Nachrichten
                                                    // nur atomic access!

volatile unsigned char bidib_tx_remaining;          // noch zu bertragen (also bereits geladen)

volatile unsigned char bidib_tx_crc;                // laufende CRC

// in-fifo handling

// a note on buffer size: bidib runs at 500k, thus one byte is 22us;
// 64 bytes = 1.4ms;
// 
#define BIDIB_RX_BUF_SIZE       48                  // mind. 64 / max 255
unsigned int bidib_rx_buf[BIDIB_RX_BUF_SIZE];

unsigned char bidib_rx_buf_read = 0;                // point to next read
unsigned char bidib_rx_buf_write = 0;               // point to next write
unsigned char bidib_rx_fill = 0;

//=================================================================================
//
// Code Section
//
//=================================================================================

//===============================================================================
//
// BiDiBus Messages (Logon)
//
//===============================================================================

// unsigned char bidib_msg_logon[] =
//   {10,             // length of message
//    0,              // message addr
//    0,              // message num
//    MSG_LOGON,      // message type
//    1,2,3,4,5,6,7,  // t_bidib_unique_id
//  };

static void bidib_send_logon(void)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = 0;
    message.header.msg_type = MSG_LOGON;
    memcpy(message.data, &my_unique_id, sizeof(my_unique_id));
    message.header.size = 3+sizeof(my_unique_id);     // + 7
    bidib_tx_fifo_put((unsigned char *)&message);
  }

// hier mal schnell eine Codekontrolle, ob beim define der unique_id alles okay ist.
#if (BIDIB_SIZE_OF_LOGON_MSG != 4+7 )   // MLENGTH, ADDR, MNUM, TYPE, sizeof(my_unique_id)
  #error BIDIB_SIZE_OF_LOGON_MSG mismatch
#endif


void bidib_flush_rx(void)                 // Flush Receive-Buffer
  {
    uint8_t sreg = SREG;
    cli();
    do
      {
        my_UDR;
      }
    while (my_UCSRA & (1 << my_RXC));

    // Ruecksetzen von Receive Complete-Flags
    my_UCSRA |= (1 << my_RXC);
    bidib_rx_buf_read = 0;
    bidib_rx_buf_write = 0;
    bidib_rx_fill = 0;
    SREG = sreg;
  }

void bidib_flush_tx(void)                 // Flush Transmit-Buffer
  {
    uint8_t sreg = SREG;
    cli();
    bidib_tx_buf_read = 0;
    bidib_tx_buf_write = 0;
    bidib_tx_fill = 0;
    SREG = sreg;
    bidib_send_logon();                   // eine Logon-Message in den Transmit-Buffer legen
  }


inline void set_bidib_to_receive(void) __attribute__((always_inline));
void set_bidib_to_receive(void)
  {
    HARDWARE_SET_BIDIB_RECEIVE;        // see bidib_client_hardware.h
  }

inline void set_bidib_to_transmit(void) __attribute__((always_inline));
void set_bidib_to_transmit(void)
  {
    HARDWARE_SET_BIDIB_TRANSMIT;        // see bidib_client_hardware.h
  }


void init_bidib_client_if(void)
  {
    uint16_t ubrr;
    uint8_t sreg = SREG;

    set_bidib_to_receive();

    cli();

    my_UCSRB = 0;                  // stop everything

    // we calculate 100* and add 50 to get correct integer cast;
    // normal speed - pre divider 16 (high speed mode - pre divider 8)

    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*62500L) - 1);
    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*625L) - 100L + 50L) / 100);
    my_UBRRH = (uint8_t) (ubrr>>8);
    my_UBRRL = (uint8_t) (ubrr);
    my_UCSRA = (1 << my_RXC) | (1 << my_TXC) | (0 << my_U2X);  // U2X: Speed Mode

    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*500000L) - 1);
    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*5000L) - 100L + 50L) / 100);
    my_UBRRH = (uint8_t) (ubrr>>8);
    my_UBRRL = (uint8_t) (ubrr);
    my_UCSRA = (1 << my_RXC) | (1 << my_TXC) | (0 << my_U2X);  // U2X: Speed Mode


    // FIFOs fuer Ein- und Ausgabe initialisieren

    bidib_flush_tx();

    // UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 9N1, asynchron

    my_UCSRB = (1 << my_RXCIE)      // enable rec_interrupt
           | (0 << my_TXCIE)
           | (0 << my_UDRIE)
           | (1 << my_RXEN)
           | (1 << my_TXEN)
           | (1 << my_UCSZ2)        // set frame length to 9bit;
           | (0 << my_RXB8)
           | (0 << my_TXB8);        // set 9th bit to 0

    my_UCSRC = (1 << my_URSEL)      // must be 1 (selector)
           | (0 << my_UMSEL)        // 0 = asyn mode
           | (0 << my_UPM1)         // 00 = parity disabled
           | (0 << my_UPM0)         //
           | (0 << my_USBS)         // 1 = tx with 2 stop bits
           | (1 << my_UCSZ1)        // 11 = 8 or 9 bits
           | (1 << my_UCSZ0)
           | (0 << my_UCPOL);

    bidib_flush_rx();

    my_bidib_node_addr = 0xFF;

    my_UCSRA |= (1 << my_TXC);         // clear tx

    SREG = sreg;
}

void stop_bidib_client_if(void)
  {
    uint8_t sreg = SREG;

    set_bidib_to_receive();

    cli();

    my_UCSRB = 0;                  // stop everything

    my_UCSRA |= (1 << my_TXC);      // clear any pending tx complete flag

    SREG = sreg;
  }


//----------------------------------------------------------------------------------
// BiDiB receive ISR
//   - received chars are feed to input fifo, 
//     consumer is the task run_bidib_client
//   - enable this task
//   - 16 bit interface
//   - no overrun check!
//   - If callbyte to our address:
//             - start timer
//             - after timer: check message and switch to tx
//
#if (BIDIB_UART == UART)
ISR(USART_RXC_vect)
#elif (BIDIB_UART == UART0)
 #ifndef USART0_RXC_vect
 #define USART0_RXC_vect USART0_RX_vect
 #endif
ISR(USART0_RXC_vect)
#elif (BIDIB_UART == UART1)
 #ifndef USART1_RXC_vect
 #define USART1_RXC_vect USART1_RX_vect
 #endif
ISR(USART1_RXC_vect)
#else
   #warning ISR undefined
#endif
  {
    unsigned char resh, resl;

    if (my_UCSRA & (1<< my_FE))
	  { // Frame Error
        resh = my_UCSRB;
        resl = my_UDR;                // at least: grep data
      }
    else
      {
        if (my_UCSRA & (1<< my_DOR))
          { // DATA Overrun -> Fatal
            // !!!
          }

        resh = my_UCSRB;
        resh = (resh >> my_RXB8) & 0x01;
        resl = my_UDR;

        sei();                              // Achtung: Interrupt freigeben, damit Soft-PWM nicht hakt.

        if (resh)                           // it's a token
          {
            // token parity checking wird hier nicht gemacht,
            // da beim Ablegen der Adresse das Parity gleich dazugerechnet wird.
            // also reicht hier ein einfacher Compare.
            
            if ((resl & 0x40) == 0x40)                          //-------->> it's a system command
              {
                // if (resl == BIDIBUS_POWER_UP_par) 
                // Hinweis: der Busreset wird erst im Parser behandelt - hier ist
                //          es wichtiger, dass die ISR klein (und damit schnell) bleibt.
                // Update:  direkter Reset wird nicht mehr behandelt -> Fehlergefahr durch Falschtrigger
                if (resl == BIDIBUS_LOGON_par)
                  {
                    if (g_bidib_connect == BIDIB_DISCONNECTED)
                      {
                        if (g_bidib_backoff)
                          {
                            g_bidib_backoff--;
                          }
                        else
                          {
                            // hier kommt jetzt zuerst collision avoidance:
                            // Der Knoten wartet eine spezifische Wartezeit (von 0.31us),
                            // debei wird die RX-Leitung beobachtet 
                            // und fallweise _nicht_ gesendet!
                            unsigned char i;
                            for (i=0; i<g_bidib_ca_time; i++)
                              {
                                if (HARDWARE_GET_RX == 0) return; 
                                _delay_us(1);
                              }
                            
                            // hier wird jetzt ein LOGON versucht
                            g_bidib_connect = BIDIB_APPLIED;
                   
                            // das passiert mit den normalen Mittel, allerdings muessen
                            // wir vorher am buffer tricksen ....
                            bidib_tx_buf_read = 0;          // Achtung: Buffer rewind!
                                                            // im Buffer liegt unsere logon message
                                                            // das ist hier 'dreckig',
                                                            // aber wir sind DISCONNECTED
                            bidib_tx_fill = BIDIB_SIZE_OF_LOGON_MSG;
                         
                            set_bidib_to_transmit();
                            my_UCSRA |= (1 << my_TXC);          // clear any pending tx complete flag
                            my_UCSRB |= (1 << my_TXEN);         // enable TX
                            my_UDR = bidib_tx_fill;             // transmit size
                     
                            bidib_tx_crc = read_crc_array(bidib_tx_fill);
                            bidib_tx_remaining = bidib_tx_fill;
                            bidib_tx_fill = 0;

                            my_UCSRB |= (1 << my_UDRIE);        // enable TxINT
                            my_UCSRB |= (1 << my_TXCIE);        // enable TX complete
                          } 
                      }
                    else if (g_bidib_connect == BIDIB_APPLIED)
                      {
                        // neuer Logon, offenbar ist unser APPLY schief gegangen
                        // wir muessen hier einen dynamischen backoff einhalten
                        
                        unsigned char mytemp;
                        
                        mytemp = static_prbs8() & 0xF;      // Zufallszahl holen
                        mytemp += g_bidib_last_backoff;
                        mytemp  &= 0x3F;                    // modulo 32
                        g_bidib_backoff= mytemp;
                        g_bidib_last_backoff = mytemp;
                        
                        g_bidib_connect = BIDIB_DISCONNECTED;
                      }
                  }
              }
            else                                                //-------->> it's a normal poll
              {
                if (resl == my_bidib_node_addr)             // komplett mit parity
                  {
                    // Hinweis; wir haben hier einen Zugriff auf unsere Node-ID, also
                    //          sollte man hier einen Update auf connected machen.
                    //          Auch das ist aus Performancegrnden in den Parser verlagert 
                    //          (dort aber nur als generelle Hostaktivitt abgebildet)
                    
                    // open tx if message is present
                    if (bidib_tx_fill)                      // check tx (atomic access)
                      {
                        set_bidib_to_transmit();
                        
                        my_UCSRA |= (1 << my_TXC);          // clear any pending tx complete flag
                        my_UCSRB |= (1 << my_TXEN);         // enable TX
                        my_UDR = bidib_tx_fill;             // transmit size of paket
                     
                        bidib_tx_crc = read_crc_array(bidib_tx_fill);
                        bidib_tx_remaining = bidib_tx_fill;
                        bidib_tx_fill = 0;
                        
                        my_UCSRB |= (1 << my_UDRIE);        // enable TxINT
                        my_UCSRB |= (1 << my_TXCIE);        // enable TX complete
                      }
                    else                                    // This is a Ping
                      {
                         set_bidib_to_transmit();
                        
                         my_UCSRA |= (1 << my_TXC);         // clear any pending tx complete flag
                         my_UCSRB |= (1 << my_TXEN);        // enable TX
                         my_UDR = BIDIBUS_NODE_READY;       // 0, hier immer ready
                      // my_UCSRB |= (1 << my_UDRIE);       // enable TxINT (hier nicht, da nur ein Wort gesendet werden soll)
                         my_UCSRB |= (1 << my_TXCIE);       // enable TX complete
                      }
                  }
              }
          }

        // always: push to fifo for analysis
        //         Das ist notwendig, weil einige Dinge (Reset, Ping)
        //         auf Parserebene erledigt werden
        
       	bidib_rx_buf[bidib_rx_buf_write] = (resh << 8) | resl;
        bidib_rx_buf_write++;
        if (bidib_rx_buf_write == BIDIB_RX_BUF_SIZE) bidib_rx_buf_write = 0;
        bidib_rx_fill++;          // no check for full, we just do an overrun!

        // make consumer ready

        #ifdef _CORTOS_H_
            set_task_ready(TASK_BIDIB);
        #else
            #warning cortos call disabled, use in polling environment only
        #endif
      }
  }

//----------------------------------------------------------------------------------
//
// TX Messages are put in a fifo
//
//----------------------------------------------------------------------------------

// return: 1 if enough room
//         0 if stuck

bool bidib_tx_fifo_healthy(void)
  {
    if (bidib_tx_fill < (BIDIB_TX_BUF_SIZE - BIDIB_TX_BUF_HEADROOM)) return(1); 
     return(0); 
  }


// Hier werden nur die Messages ins Fifo gelegt, Paketbndelung macht dann der
// Leseprozess.

// message = LENGTH MSG_ADDR MSG_NUM MSG_TYPE DATA 
// return: 1 if failed
//         0 if success

bool bidib_tx_fifo_put(unsigned char *new_message)
  {
    unsigned char i;
    unsigned char size;
    signed char j;

    size = *new_message + 1;                        // total = message size + 1;

    j = bidib_tx_buf_read - bidib_tx_buf_write - 1;
    j &= (BIDIB_TX_BUF_SIZE-1);                     // modulo

    if (size >= j) return(1);

    for (i = 0; i < size; i++)
      {
        bidib_tx_buf[bidib_tx_buf_write] = *new_message++;
        bidib_tx_buf_write++;
        bidib_tx_buf_write &= (BIDIB_TX_BUF_SIZE-1);
      }

    cli();
    bidib_tx_fill += size;                          // set flag for TX
    sei();
    
    return(0);
  }

/*
bool bidib_out_buf_is_full(void)
  {
    unsigned char i;
    
    // check for full
    i = bidib_tx_buf_write+1;
    if (i == BIDIB_TX_BUF_SIZE) i = 0;
    if (i == bidib_tx_buf_read) return(1);        // totally full
    return(0);
  }
*/


//----------------------------------------------------------------------------------
// TX: transmit one byte from bidib_out_buf to uart
//     if transmitted: a new UART_DATA-IRQ will be created
//     if all data is transmitted, we send the crc
// We transmit 9 bits, but the 8th bit is allways zero.

#if (BIDIB_UART == UART)
ISR(USART_UDRE_vect)
#elif (BIDIB_UART == UART0)
ISR(USART0_UDRE_vect)
#elif (BIDIB_UART == UART1)
ISR(USART1_UDRE_vect)
#else
   #warning ISR undefined
#endif
  {
    unsigned char txdat;
    if (bidib_tx_remaining > 0)
      {
        // my_UCSRA |= (1 << my_TXC);              // writing a one clears any existing tx complete flag
        // my_UCSRB &= ~(1<<my_TXB8);               // clear bit 8

        txdat = bidib_tx_buf[bidib_tx_buf_read];
        my_UDR = txdat;
        bidib_tx_crc = read_crc_array(bidib_tx_crc ^ txdat);
        bidib_tx_buf_read++;
        if (bidib_tx_buf_read == BIDIB_TX_BUF_SIZE) bidib_tx_buf_read = 0;
        bidib_tx_remaining--;
      }
    else
      {
        my_UDR = bidib_tx_crc;
        my_UCSRB &= ~(1 << my_UDRIE);           // disable further TxINT
      }
  }

//----------------------------------------------------------------------------------
// Halbduplex: auf TXC Complete schalten wir sofort die Richtung um
// Das als NAKED um es wirklich schnell zu bekommen.
// Wenn Optimize wegfaellt, dann darf NAKED nicht mehr benutzt werden - Flags!
// Vermutlich wird alles andere als ISR_NOBLOCK laufen muessen!
//

#if (BIDIB_UART == UART)
  #define BIDIB_TXC_vect USART_TXC_vect
#elif (BIDIB_UART == UART0)
  #ifndef USART0_TXC_vect
    #define USART0_TXC_vect USART0_TX_vect
  #endif
  #define BIDIB_TXC_vect USART0_TXC_vect
#elif (BIDIB_UART == UART1)
  #ifndef USART1_TXC_vect
    #define USART1_TXC_vect USART1_TX_vect
  #endif
  #define BIDIB_TXC_vect USART1_TXC_vect
#else
   #warning ISR undefined
#endif

/*
ISR(BIDIB_TXC_vect)
  {
    set_bidib_to_receive();
  }
*/

ISR(BIDIB_TXC_vect, ISR_NAKED)
  {
    set_bidib_to_receive();     // das ist hier nur ein einzelnes Statement ohne
                                // Flags - darum als naked moeglich
    reti();                     // wenn naked, dann auch RETI erforderlich
  }

//==================================================================================
// Upstream Interface
//----------------------------------------------------------------------------------
// TX:
bool bidib_tx_ready (void)
  {
    if (bidib_tx_fill == 0)
      {
        return(1);                        // yes, we can take a message
      }
    else
      {
        return(0);
      }
  }

//----------------------------------------------------------------------------------
// RX:
bool bidib_rx_ready (void)
{
    if (bidib_rx_buf_read != bidib_rx_buf_write)
      {
        return(1);     // there is something
      }
    else
      {
        return(0);
      }
}

//----------------------------------------------------------------------------------
// rx_fifo_read gets one char from the input fifo
//
// there is no check whether a char is ready, this must be
// done before calling with a call to rx_fifo_ready();

unsigned int bidib_rx_read (void)
  {
    unsigned int retval;

    retval = bidib_rx_buf[bidib_rx_buf_read];
    bidib_rx_buf_read++;
    if (bidib_rx_buf_read == BIDIB_RX_BUF_SIZE) bidib_rx_buf_read=0;

    unsigned char mysreg = SREG;
    cli();
    bidib_rx_fill--;
    SREG = mysreg;

    return(retval);
  }

// BiDiB 'vertraeglich' abschalten, hierzu zuerst RX disablen
// (mit der Gefahr, dass eine laufende Nachricht verstmmelt wird),
// dann eine Zeit busy warten (falls gerade TX laeuft), dann TX kontrollieren.

void bidib_soft_stop (void)
  {
	// disable further RxINT
	cli();
    my_UCSRB &= ~(1 << my_RXCIE);
	sei();
    _delay_us(50);
    // cli();
    // while((BIDIB_UART.STATUS & (USART_TXCIF_bm | USART_DREIF_bm)) != (USART_TXCIF_bm | USART_DREIF_bm)){}
  }

void bidib_soft_start (void)
  {
	cli();

    // UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 9N1, asynchron

    my_UCSRA |= (1 << my_TXC)      // clear tx
               |(1 << my_RXC) ;    // clear rx

    my_UCSRB = (1 << my_RXCIE)      // enable rec_interrupt
           | (0 << my_TXCIE)
           | (0 << my_UDRIE)
           | (1 << my_RXEN)
           | (1 << my_TXEN)
           | (1 << my_UCSZ2)        // set frame length to 9bit;
           | (0 << my_RXB8)
           | (0 << my_TXB8);        // set 9th bit to 0
    sei();
  }

//----------------------------------------------------------------------------------
// testroutinen zu Parsertest
// Ersatzroutine, um ein Zeichen in den Empfangsbuffer abzulegen
// Sendepuffer kann im Simulator mit einen Databreakpoint auf UDR �berwacht werden.
void X_push_to_rx(unsigned int d)
  {
	bidib_rx_buf[bidib_rx_buf_write] = d;

    bidib_rx_buf_write++;
    if (bidib_rx_buf_write == BIDIB_RX_BUF_SIZE) bidib_rx_buf_write=0;
    bidib_rx_fill++;
  }


//==================================================================================
// security checks for misconfiguration
//==================================================================================


