//=================================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010-2012 Wolfgang Kufer
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//
//---------------------------------------------------------------------------------
//
// file:      bidib_client_parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-12-19 V0.01 kw  added BiDi-CV Read
//            2011-09-18 V0.03 bidib_send_sys_sw_version -> bidib_send_sys_pversion
//            2011-01-14       MSG_LOCAL_PONG added
//            2012-03-13 V0.04 kw  Bugfix Length
//            2012-03-18 V0.05 kw  feature in chunks. tx_fifo_healthy
//            2012-03-19 V0.06 kw  Umstellung auf neue feature und node transmit
//            2012-05.04 V0.07 kw  Config Port erweitert.
//            2012-07-13 V0.08 kw  PONG erweitert
//            2012-07-23 V0.09 kw  Vendor-Befehle
//            2012-08-07 V0.10 kw  MSG_SYS_CLOCK
//            2012-09-27 V0.11 kw  MSG_LC_WAIT wird hier nicht untersttzt
//                                 (bei Dimmvorgngen nicht wichtig) 
//            2012-10-14 V0.12 kw  MSG_ACCESSORY_SET/GET
//            2012-10-31 V0.13 kw  acc fixed
//            2012-12-03 V0.14 kw  FEATURE_ACCESSORY_MACROMAPPED= NUM_OF_ASPECTS_PER_ACCESSORY
//            2013-03-02 V0.16 kw  added if (g_bidib_connect == BIDIB_CONNECTED) for broadcast
//            2013-05-07 V0.17 kw  bugfix in vendor_eval
//            2013-07-24 V0.18 kw  bugfix in process_bidib_message - unintended decoding of messages
//            2013-12-20 V0.19 kw  MSG_STRING
//
//---------------------------------------------------------------------------------
//
// purpose:   BiDiB-Client
//            
// ====== >>>> config.h  is the central definition file for the project
//             bidib_client_hardware.h is our hardware
//
//=================================================================================
//
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/parity.h>
#include <string.h>

#include "config.h"                 // general definitions
#include "datatypes.h"              // general definitions
#include "hardware.h"
#include "timer.h"                  // access to systick for surveillance
#include "cortos.h"                 // fifo access
#include "keycodes.h"               
#include "keyboard.h"
#include "led.h"

#include "crc_8bit.h"
#include "prbs8.h"
#include "bidib_messages.h"            // general include for BiDiB (taken from www.bidib.org!)
#include "bidib_client_hardware.h"     // port definitions for target
#include "bidib_client_if.h"           // this is the low level backend
#include "bidib_client_parser.h"       // our header


#include "lights.h"                 // downstream
#include "macro.h"                  // downstream



#define DEBUG_RX   0
#if (DEBUG != 0)                    // generell debug allowed
    #define DEBUG_RX   0            // 0: no debug messages
                                    // 1: dump down messages to our slot
                                    // 2: dump first received byte
#endif
#if (DEBUG_RX == 1)
    #warning debug printout of BiDiB RX messages enabled
#endif


#define LED_CONTROL   1             // 1: set / reset LEDs
#if (LED_CONTROL == 1)
    // typically, we use 3 LEDs
        

#endif
//===============================================================================
//
// Data Definitions / Configuration
// 
//===============================================================================
// 
// with BIDIB, we have different Classes of NODES, it is essential to define
// the correct classes for your node

#define   BIDIB_VENDOR_ID     13

#define   BIDIB_PRODUCT_ID   108


#define   CLASS_OCCUPANCY      0                // 0: no occupancy messages
                                                // 1: occupancy messages included

#define   CLASS_ACCESSORY      1                // 0: no accessory messages, 
                                                // 1: accessory messages included

#define   CLASS_SWITCH         1                // 0: no switch messages
                                                // 1: switch messages included

//-------------------------------------------------------------------------------


unsigned char g_bidib_spontan_enabled;          // the node is allowed to send occupancy

unsigned char g_bidib_last_error;               // the last error

unsigned char g_bidib_identify;                 // identify mode

static unsigned char bidib_rx_msg_num = 0;      // counter for rx sequence
static unsigned char bidib_tx0_msg_num;         // this is our running message index

static unsigned char bidib_feature_transmit;    // flag, ob wir aktuell features senden
static unsigned char bidib_feature2send;        // und da sind wir gerade
static unsigned char bidib_node2send;           // counter for node transmission


const unsigned char bidib_sys_magic[] =         // this is our answer for SYS_MAGIC
  { 5,              // size
    0x00,           // addr
    0x00,           // msg_num
    MSG_SYS_MAGIC,
    BIDIB_SYS_MAGIC & 0xFF,
    (BIDIB_SYS_MAGIC >> 8) & 0xFF,
  };

t_bidib_unique_id my_unique_id =
  {
    {
      {
        .class_switch = CLASS_SWITCH,
        .class_booster = 0,
        .class_accessory = CLASS_ACCESSORY,
        .class_dcc_prog = 0,
        .class_dcc_main = 0,
        .class_ui = 0,
        .class_occupancy = CLASS_OCCUPANCY,     // Belegtmelder-Klasse anmelden
        .class_bridge = 0,
      },
    },
   .classx_id = 0,                              // reserviert
   .dcc_vendor = BIDIB_VENDOR_ID,
    {
      {
        .product_id = BIDIB_PRODUCT_ID,
        .product_serial = 7,
      }
    }
  };

// Some Triggermark - for debugging with a scope
// 
#define DEBUG_TRIGGER 0
#if (DEBUG_TRIGGER == 1)
    #warning DEGUG_TRIGGER is on
    #define TRIGA    { LED_MSG_ON(); _delay_us(5); LED_MSG_OFF(); }
    #define TRIGB    { LED_S88_ON(); _delay_us(5); LED_S88_OFF(); }
    #define TRIGC    { LED_BIDIB_ON(); _delay_us(5); LED_BIDIB_OFF(); }
#else
    #define TRIGA    {  }
    #define TRIGB    {  }
    #define TRIGC    {  }
#endif
//=================================================================================
//
// Upstream Messages
//
//=================================================================================


#if (CLASS_OCCUPANCY == 1)
    // local data / headers (Belegtmeldung)

    signed int last_bm_sent;            // Zeitpunkt der letzten Sendung

    static void get_belegtmeldungen(unsigned char start, unsigned char ende);
    static void set_belegt_ack(unsigned char *msg);
    static void set_single_free_ack(unsigned char index);
    static void set_single_belegt_ack(unsigned char index);
#endif

//-------------------------------------------------------------------------------
//
// basic output routines
//
//-------------------------------------------------------------------------------
//
// Single Message TX, no concatenation of messages
// parameter: message: pointer to complete message (no crc, no delimiter)
bool send_bidib_message(unsigned char *message)
  {
    return (bidib_tx_fifo_put(message));
  }

//-- helper functions

static unsigned char get_tx_num(void)
  {
    unsigned char retval;
    retval = bidib_tx0_msg_num;
    bidib_tx0_msg_num++;
    if (bidib_tx0_msg_num == 0) bidib_tx0_msg_num++;
    return(retval);
  }

/*
static void bidib_send_single_msg(unsigned char message)
  {
    t_node_message_header header;
    header.node_addr = 0;
    header.index = get_tx_num();
    header.msg_type = message;
    header.size = 3+0;

    send_bidib_message((unsigned char *)&header);
  }
*/

bool bidib_send_onepara_msg(unsigned char msg_type, unsigned char dat)
  {
    t_node_message1 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = msg_type;
    message.data = dat;
    message.header.size = 3+1;

    return (send_bidib_message((unsigned char *)&message));
  }


//===============================================================================

void bidib_send_sys_identify(unsigned char state)
  {
    bidib_send_onepara_msg(MSG_SYS_IDENTIFY_STATE, state);
  }

static void bidib_send_sys_magic(void)
  {
    send_bidib_message((unsigned char *) bidib_sys_magic);
    bidib_tx0_msg_num = 0;
  }

static void bidib_send_sys_pong(unsigned char para)
  {
    // bidib_send_single_msg(MSG_SYS_PONG);
    bidib_send_onepara_msg(MSG_SYS_PONG, para);
  }

static void bidib_send_sys_pversion(void)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_SYS_P_VERSION;
    message.data[0] = BIDIB_VERSION & 0xFF;
    message.data[1] = (BIDIB_VERSION >> 8) & 0xFF;
    message.header.size = 3+2;
    
    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_sys_unique_id(void)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_SYS_UNIQUE_ID;
    memcpy(message.data, &my_unique_id, sizeof(my_unique_id));
    message.header.size = 3+sizeof(my_unique_id);     // + 7

    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_sys_sw_version(void)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_SYS_SW_VERSION;
    message.data[0] = COMPILE_RUN;
    message.data[1] = SUB_VERSION;
    message.data[2] = MAIN_VERSION;
    message.header.size = 3+3;

    send_bidib_message((unsigned char *)&message);
  }


static void bidib_send_error(unsigned char error_num, unsigned char error_para)
  {
    t_node_message2 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_SYS_ERROR;
    message.data[0] = error_num;
    message.data[1] = error_para;
    message.header.size = 3+2;

    send_bidib_message((unsigned char *)&message);
  }


static void bidib_send_node_na(unsigned char wrong_node)
  {
    bidib_send_onepara_msg(MSG_NODE_NA, wrong_node);
  }


static void bidib_send_pkt_capacity(void)
  {
    bidib_send_onepara_msg(MSG_PKT_CAPACITY, 64);   // gilt fuer FLAT und BiDiBus
  }

static void bidib_send_node_count(unsigned char length)
  {
    bidib_send_onepara_msg(MSG_NODETAB_COUNT, length);
  }

static void bidib_send_node_tab_empty(void)
  {
    if (bidib_node2send == 0)
      {
        bidib_node2send++;
        t_node_message18 message;
        message.header.node_addr = 0;
        message.header.index = get_tx_num();
        message.header.msg_type = MSG_NODETAB;
        message.data[0] = 1;    // Version
        message.data[1] = 0;    // local addr
        memcpy(&message.data[2], &my_unique_id, sizeof(my_unique_id));
        message.header.size = 3+2+sizeof(my_unique_id);  // 3+9, 12 in total

        send_bidib_message((unsigned char *)&message);
      }
    else
      {
        bidib_send_onepara_msg(MSG_NODE_NA, 255);
      }
  }

//-------------------------------------------------------------------------------
// Bootloader

void bidib_send_bootloader_state(uint8_t stat, uint8_t timeout)
  {
    t_node_message2 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_FW_UPDATE_STAT;
    message.data[0] = stat;
    message.data[1] = timeout;
    message.header.size = 3+2;                          // 3= sizeof(t_node_message_header)-1
    
    send_bidib_message((unsigned char *)&message);
  }

//-------------------------------------------------------------------------------
// Feature-Handling
//
// Fuer features haben wir hier eine Tabelle, in der wir das angefragte feature
// suchen: wenn es vorhanden ist, dann wird es zurueckmeldet.
// Beim Setzen eines Features vergleichen wir auf die Grenzen (min, max) und rufen
// bei Veraenderungen die notify_function auf, damit die Veraenderung auch wirkt.
//
// --> Zum Hinzufuegen weiterer features einfach die Tabelle erweitern!
//     Es gibt einen #define (mit der Endung _idx), dieser mu durchlaufend
//     numeriert sein, ber diesen
//     define werden die features lokal angesprochen. Fr den remote-Zugriff ist
//     die verpflichtende Featurenummer als erster Eintrag abgespeichert.
// --> Nicht benoetige Features kann man einfach weglassen

typedef struct
  {
    unsigned char num;              // feature index 
    unsigned char value;            // der aktuelle Wert
    unsigned char min;              // unsere Grenzen
    unsigned char max;
    void (*notify_function)(void);  // Diese Funktion wird gerufen, wenn das Feature beschrieben wird
  } t_feature;

#if (CLASS_OCCUPANCY == 1)
    #define FEATURE_BM_SIZE_idx                     0
    #define FEATURE_BM_ON_idx                       1
    #define FEATURE_BM_SECACK_AVAILABLE_idx         2
    #define FEATURE_BM_SECACK_ON_idx                3
    #define FEATURE_BM_ADDR_DETECT_AVAILABLE_idx    4
    #define FEATURE_BM_ADDR_DETECT_ON_idx           5
    #define FEATURE_BM_ADDR_AND_DIR_idx             6     
    #define FEATURE_BM_CV_AVAILABLE_idx             7
    #define FEATURE_BM_CV_ON_idx                    8
    #define FEATURE_STRING_SIZE_idx                 9
    #define FEATURE_table_size                     10


    // am Node 0 habe wir wenig features ... nur Melderbits
    t_feature feature[] =
      { // index                                    num                                value min max notify_function (if changed)
        [FEATURE_BM_SIZE_idx]                   = { FEATURE_BM_SIZE,                   24,   0,  72, NULL},  // 0
        [FEATURE_BM_ON_idx]                     = { FEATURE_BM_ON,                     1,    0,  1,  NULL},  // 1
        [FEATURE_BM_SECACK_AVAILABLE_idx]       = { FEATURE_BM_SECACK_AVAILABLE,       1,    0,  1,  NULL},  // 2
        [FEATURE_BM_SECACK_ON_idx]              = { FEATURE_BM_SECACK_ON,              0,    0,  1,  NULL},  // 3
        [FEATURE_BM_ADDR_DETECT_AVAILABLE_idx]  = { FEATURE_BM_ADDR_DETECT_AVAILABLE,  1,    0,  1,  NULL},  // 8
        [FEATURE_BM_ADDR_DETECT_ON_idx]         = { FEATURE_BM_ADDR_DETECT_ON,         1,    0,  1,  NULL},  // 9
        [FEATURE_BM_ADDR_AND_DIR_idx]           = { FEATURE_BM_ADDR_AND_DIR,           1,    0,  1,  NULL},  // 10
        [FEATURE_BM_CV_AVAILABLE_idx]           = { FEATURE_BM_CV_AVAILABLE,           1,    0,  1,  NULL},  // 13
        [FEATURE_BM_CV_ON_idx]                  = { FEATURE_BM_CV_ON,                  1,    0,  1,  NULL},  // 14
        [FEATURE_STRING_SIZE_idx]               = { FEATURE_STRING_SIZE,              BIDIB_STRING_MAX, BIDIB_STRING_MAX, BIDIB_STRING_MAX,     NULL},  // 14
      };
#endif // (CLASS_OCCUPANCY == 1)
#if (CLASS_SWITCH == 1)
    #define FEATURE_CTRL_LPORT_COUNT_idx          0
    #define FEATURE_CTRL_INPUT_COUNT_idx          1
    #define FEATURE_CTRL_INPUT_NOTIFY_idx         2
    #define FEATURE_CTRL_MAC_LEVEL_idx            3
    #define FEATURE_CTRL_MAC_SAVE_idx             4
    #define FEATURE_CTRL_MAC_COUNT_idx            5
    #define FEATURE_CTRL_MAC_SIZE_idx             6
    #define FEATURE_CTRL_MAC_START_MAN_idx        7
    #define FEATURE_ACCESSORY_COUNT_idx           8
    #define FEATURE_ACCESSORY_SURVEILLED_idx      9
    #define FEATURE_ACCESSORY_MACROMAPPED_idx    10
    #define FEATURE_STRING_SIZE_idx              11
    #define FEATURE_table_size                   12
    

    t_feature feature[] =
      { // index                                    num                            value min max notify_function (if changed)
        [FEATURE_CTRL_LPORT_COUNT_idx]        = { FEATURE_CTRL_LPORT_COUNT,        24,   24, 24,  NULL},  // 0
        [FEATURE_CTRL_INPUT_COUNT_idx]        = { FEATURE_CTRL_INPUT_COUNT,         2,    2,  2,  NULL},  // 1
        [FEATURE_CTRL_INPUT_NOTIFY_idx]       = { FEATURE_CTRL_INPUT_NOTIFY,        1,    0,  1,  NULL},  // 2
        [FEATURE_CTRL_MAC_LEVEL_idx]          = { FEATURE_CTRL_MAC_LEVEL,           2,    2,  2,  NULL},  // 3
        [FEATURE_CTRL_MAC_SAVE_idx]           = { FEATURE_CTRL_MAC_SAVE,            1,    1,  1,  NULL},  // 4
        [FEATURE_CTRL_MAC_COUNT_idx]          = { FEATURE_CTRL_MAC_COUNT,           NUM_OF_MACROS,    0,  NUM_OF_MACROS,  NULL},  // 3
        [FEATURE_CTRL_MAC_SIZE_idx]           = { FEATURE_CTRL_MAC_SIZE,            NUM_OF_ITEMS_PER_MACRO,    0,  NUM_OF_ITEMS_PER_MACRO,  NULL},  // 8
        [FEATURE_CTRL_MAC_START_MAN_idx]      = { FEATURE_CTRL_MAC_START_MAN,       1,    0,  1,  NULL},  // 7
        [FEATURE_ACCESSORY_COUNT_idx]         = { FEATURE_ACCESSORY_COUNT,          NUM_OF_ACCESSORY,    NUM_OF_ACCESSORY,  NUM_OF_ACCESSORY,  NULL},  // 10
        [FEATURE_ACCESSORY_SURVEILLED_idx]    = { FEATURE_ACCESSORY_SURVEILLED,      0,    0,  0,  NULL},  // 10
        [FEATURE_ACCESSORY_MACROMAPPED_idx]   = { FEATURE_ACCESSORY_MACROMAPPED,    NUM_OF_ASPECTS_PER_ACCESSORY,    NUM_OF_ASPECTS_PER_ACCESSORY,  NUM_OF_ASPECTS_PER_ACCESSORY,  NULL},  // 14
        [FEATURE_STRING_SIZE_idx]             = { FEATURE_STRING_SIZE,              BIDIB_STRING_MAX, BIDIB_STRING_MAX, BIDIB_STRING_MAX,     NULL},  // 14
      };
#endif // (CLASS_SWITCH == 1)


#if (NUM_OF_FEATURES < FEATURE_table_size)
   #error size im eeprom does not fit, see cv_define.h and cv_data.h
#endif

static void init_features_from_eeprom(void)
  {
    unsigned char i;
    unsigned char f_value;

    for (i=0; i<sizeof(feature)/sizeof(feature[0]); i++)
      {
        f_value = eeprom_read_byte(&CV.feature[i]);
        if (f_value < feature[i].min) f_value = feature[i].min;
        if (f_value > feature[i].max) f_value = feature[i].max;
        feature[i].value = f_value;
      }
  }

static void bidib_send_feature(unsigned char fnum, unsigned char f_value)
  {
    t_node_message2 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_FEATURE;
    message.data[0] = fnum;
    message.data[1] = f_value;
    message.header.size = 3+2;
    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_feature_na(unsigned char fnum)
  {
    bidib_send_onepara_msg(MSG_FEATURE_NA, fnum);
  }

static void bidib_send_feature_count(unsigned char fcount)
  {
    bidib_send_onepara_msg(MSG_FEATURE_COUNT, fcount);
  }

static void set_feature(unsigned char fnum, unsigned char f_value)
  {
    unsigned char i;
    void (*fp)(void);                               // declare pointer to function

    for (i=0; i<sizeof(feature)/sizeof(feature[0]); i++)
      {
        if (fnum == feature[i].num)
          {
            if (f_value < feature[i].min) f_value = feature[i].min;
            if (f_value > feature[i].max) f_value = feature[i].max;
            if (feature[i].value != f_value)
              {
                feature[i].value = f_value;
                fp = feature[i].notify_function;
                if (fp) {(*fp)();}                    // call this function, if not NULL
              }
            bidib_send_feature(fnum, feature[i].value);    // feature zuruecksenden
            return;
          }
      }
    bidib_send_feature_na(fnum);                            // kein feature gefunden
  }

static void get_feature(unsigned char fnum)
  {
    unsigned char i;

    for (i=0; i<sizeof(feature)/sizeof(feature[0]); i++)
      {
        if (fnum == feature[i].num)
          {
            bidib_send_feature(fnum, feature[i].value);    // feature zuruecksenden
            return;
          }
      }
    bidib_send_feature_na(fnum);                            // kein feature gefunden
  }

static void bidib_send_next_feature(void)
  {
    if (bidib_feature2send < sizeof(feature)/sizeof(feature[0]))
      {
        bidib_send_feature(feature[bidib_feature2send].num, feature[bidib_feature2send].value);// feature zurcksenden
        bidib_feature2send++;
      }
    else
      {
        bidib_send_feature_na(255);                           // alle gesendet
      }
  }

#if (CLASS_OCCUPANCY == 1)
void set_melderbits(unsigned char bits)
  {
        feature[FEATURE_BM_SIZE_idx].value = bits;   // FEATURE_BM_SIZE
  }
#endif

#if (CLASS_SWITCH == 1)
unsigned char bidib_feature_key_send_isset(void)
  {
    if (feature[FEATURE_CTRL_INPUT_NOTIFY_idx].value)       // wurde freigeschaltet
        return(1);
    else return(0);
  }

unsigned char bidib_feature_macro_start_man_isset(void)
  {
    if (feature[FEATURE_CTRL_MAC_START_MAN_idx].value)       // wurde freigeschaltet
        return(1);
    else return(0);
  }

#endif



//-------------------------------------------------------------------------------
// Vendor Config

unsigned char g_vendor_config = 0;

static void enable_vendor_config(void)
  {
    g_vendor_config = 1;
    bidib_send_onepara_msg(MSG_VENDOR_ACK, g_vendor_config);
  }
static void disable_vendor_config(void)
  {
    g_vendor_config = 0;
    bidib_send_onepara_msg(MSG_VENDOR_ACK, g_vendor_config);
  }

// addr gibt die CV-Adresse 'menschlich' an, d.h. ab 1 gezaehlt
// intern agieren wir ab 0, d.h. hier wird 1 addiert
static void bidib_send_vendor_cv(unsigned int addr, unsigned char data)
  {
    unsigned char i,j;
    t_node_message18 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_VENDOR;
    
    utoa(addr+1, (char *) &message.data[1], 10);   
    i = strlen((char *) &message.data[1]);
    message.data[0] = i;
    
    utoa(data, (char *) &message.data[i+2], 10);   
    j = strlen((char *) &message.data[i+2]);
    message.data[i+1] = j;
    
    message.header.size = 3+i+1+j+1;      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

// wir werten hier nur numerische Vendorconfigs aus; das mit atoi, d.h. nur bis 32767.
static unsigned int vendor_eval(unsigned char* str)
  {
    char para[6]; unsigned char size;
    unsigned int addr;
    
    size = *str++;
    if (size > (sizeof(para)-1)) size = sizeof(para)-1;
    memcpy(para,str,size);   // die ersten 5 Zeichen nehmen.
    para[size] = 0;
    addr = atoi(para);
    return(addr);
  }

static void vendor_get(unsigned char* msg)
  {
    unsigned int addr;
    unsigned char data;
    unsigned char *ramptr;

    addr = vendor_eval(msg);
    if (addr > 0) addr--;
    ramptr = (uint8_t *)&CVram + addr;
    data = *ramptr;
    bidib_send_vendor_cv(addr, data);
  }

static void vendor_set(unsigned char* msg)
  {
    unsigned int addr;
    unsigned char i, data;
    unsigned char *ramptr;

    i = *msg;
    addr = vendor_eval(msg);
    if (addr > 0) addr--;
    msg += i+1;      // auf zweitem Parameter
    data = vendor_eval(msg);

    eeprom_write_byte((uint8_t *)&CV + addr, data);
    ramptr = (uint8_t *)&CVram + addr;
    *ramptr = data;

    bidib_send_vendor_cv(addr, data);
  }

//--------------------------------------- 0123456789ABCDEF
    const char vendor_string[] PROGMEM = "MobaList 32";
    
#if (BIDIB_STRING_MAX < 16)
  #error BIDIB_STRING_MAX too small, string doesnt fit
#endif

static void bidib_send_string_vendor(void)
  {
    t_node_message28 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_STRING;
    message.data[0] = 0;                        // Name space
    message.data[1] = 0;                        // string_id
    message.data[2] = sizeof(vendor_string)-1;  // size
    memcpy_P(&message.data[3], vendor_string, sizeof(vendor_string)-1);
    message.header.size = 3+3+sizeof(vendor_string)-1;      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_string_username(void)
  {
    t_node_message28 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_STRING;
    message.data[0] = 0;                        // Name space
    message.data[1] = 1;                        // string_id
    message.data[2] = strlen((char*)CVram.bidib.usernick);   // size
    if (message.data[2] > BIDIB_STRING_MAX) message.data[2] = BIDIB_STRING_MAX;
    memcpy(&message.data[3], CVram.bidib.usernick, message.data[2]);
    message.header.size = 3+3+message.data[2];      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

static void user_string_get(unsigned char* msg)
  {
    if (*msg == 0)
      {
        if (*(msg+1) == 0) { bidib_send_string_vendor();     return;  }
        if (*(msg+1) == 1) { bidib_send_string_username();   return;  }
      }
    t_node_message28 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_STRING;
    message.data[0] = *msg;                        // Name space
    message.data[1] = *(msg+1);                    // string_id
    message.data[2] = 0;                           // size
    message.header.size = 3+3;                     // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

static void user_string_set(unsigned char* msg)
  {
    if (*msg == 0)
      {
        if (*(msg+1) == 1)
          {
            unsigned char size, i;
            size = *(msg + 2);
            if (size > BIDIB_STRING_MAX) size = BIDIB_STRING_MAX;             
            for (i=0; i<size; i++)
              {
                CVram.bidib.usernick[i] = *(msg + 3 + i);
                eeprom_write_byte(&CV.bidib.usernick[i], *(msg + 3 + i));
              }
            CVram.bidib.usernick[size] = 0;
            eeprom_write_byte(&CV.bidib.usernick[size], 0);
            bidib_send_string_username();   
            return;
          }
      }
    t_node_message28 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_STRING;
    message.data[0] = *msg;                        // Name space
    message.data[1] = *(msg+1);                    // string_id
    message.data[2] = 0;                           // size
    message.header.size = 3+3;                     // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }


//-------------------------------------------------------------------------------
#if (CLASS_ACCESSORY == 1)
//-------------------------------------------------------------------------------

static unsigned char _get_accessory_total(unsigned char acc)
  {
    unsigned char i=0;
    while (i<NUM_OF_ASPECTS_PER_ACCESSORY)
      {
        if (CVram.accessory_map[acc].mapped_macro[i] == 255) return(i);
        i++;
      }
    return(i);
  }

void bidib_send_accessory_state(unsigned char acc, unsigned char fb_state, unsigned char wait)
  {
    unsigned char aspect, total;
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_ACCESSORY_STATE;
    if (acc >= NUM_OF_ACCESSORY) 
      {
        message.data[0] = 255;
        message.data[1] = 0;
        message.data[2] = 0;
        message.data[3] = BIDIB_ACC_STATE_NO_FB_AVAILABLE + BIDIB_ACC_STATE_DONE;   // kein feedback / fertig
        message.data[4] = wait;                 // no wait
      }
    else
      {
        message.data[0] = acc;
        total = _get_accessory_total(acc);
        aspect = CVram.accessory_map[acc].current_aspect;
        message.data[1] = aspect;
        if (aspect >= total)  message.data[1] = 255;
        message.data[2] = total;
        message.data[3] = fb_state;
        message.data[4] = wait;                 // no wait
      }
    message.header.size = 3+5;      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

void bidib_send_accessory_mapping(unsigned char *data)
  {
    unsigned char acc, i=0;
    t_node_message18 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_ACCESSORY_PARA;
    acc = data[0];
    if (data[1] != BIDIB_ACCESSORY_PARA_MACROMAP)
      {
        message.data[0] = acc;
        message.data[1] = 255;       // not supported
        message.header.size = 3+2;      // 3= sizeof(t_node_message_header)-1
        send_bidib_message((unsigned char *)&message);
      }
    else
      {
        message.data[0] = acc;
        message.data[1] = BIDIB_ACCESSORY_PARA_MACROMAP;
        while ( (i<NUM_OF_ASPECTS_PER_ACCESSORY) &&
                (CVram.accessory_map[acc].mapped_macro[i] != 255) )
          {
            message.data[2+i] = CVram.accessory_map[acc].mapped_macro[i];
            i++;
          }
        message.data[2+i] = 255;
        message.header.size = 3+3+i;      // 3= sizeof(t_node_message_header)-1
        send_bidib_message((unsigned char *)&message);
      }
  }

void bidib_eval_accessory_mapping(unsigned char *data)
  {
    unsigned char acc, i=0;
    acc = data[0];
    if (data[1] == BIDIB_ACCESSORY_PARA_MACROMAP)
      {
        while (i<NUM_OF_ASPECTS_PER_ACCESSORY)                  // einfach alles kopieren - bei 255 ist Schluss
          {
            CVram.accessory_map[acc].mapped_macro[i] = data[2+i];
            eeprom_write_byte(&CV.accessory_map[acc].mapped_macro[i], data[2+i]); 
            i++;
          }
      }
    bidib_send_accessory_mapping(data);
  }

#endif // (CLASS_ACCESSORY == 1)
//-------------------------------------------------------------------------------
#if (CLASS_SWITCH == 1)
//-------------------------------------------------------------------------------

static void bidib_send_port_na(t_bidib_port* myport)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_NA;
    message.data[0] = myport->type;
    message.data[1] = myport->portnum;
    message.header.size = 3+2;      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }
// hier wird der Status 'as is' gesendet, keine interne Speicherung
static void bidib_send_port_stat(t_bidib_port* myport)
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_STAT;
    message.data[0] = myport->type;
    message.data[1] = myport->portnum;
    switch(myport->type)
      {
        case BIDIB_OUTTYPE_SPORT:               // standard port
            bidib_send_port_na(myport);
            break;
            
        case BIDIB_OUTTYPE_LPORT:
            if (myport->portnum >= NUM_OF_LIGHTS) bidib_send_port_na(myport);
            else
              {
                message.data[2] = myport->portstat;  // hier immer sofort melden - sonst abfragen
                message.header.size = 3+3;      // 3= sizeof(t_node_message_header)-1
                send_bidib_message((unsigned char *)&message);
              }
            break;
            
        case BIDIB_OUTTYPE_SOUND:
            bidib_send_port_na(myport);
            break;
            
        case BIDIB_OUTTYPE_MOTOR:
            bidib_send_port_na(myport);
            break;
            
        case BIDIB_OUTTYPE_ANALOG:
            bidib_send_port_na(myport);
            break;
            
        default:
            bidib_send_port_na(myport);
            break;
      }
  }

static void bidib_send_port_config(t_bidib_port* port)
  {
    t_node_message10 message;
    t_bidib_lport_cfg myport;

    switch(port->type)
      {
        case BIDIB_OUTTYPE_SPORT:  break;    // standard port
        case BIDIB_OUTTYPE_LPORT:            // light port
            myport.portnum = port->portnum;
            if (config_light_query(&myport))
              {
                message.header.node_addr = 0;
                message.header.index = get_tx_num();
                message.header.msg_type = MSG_LC_CONFIG;
                message.data[0] = port->type;
                memcpy(&message.data[1], &myport, sizeof(t_bidib_lport_cfg));
                message.header.size = 3+1+sizeof(t_bidib_lport_cfg);      // 3= sizeof(t_node_message_header)-1
    
                send_bidib_message((unsigned char *)&message);
                return;
              }
        case BIDIB_OUTTYPE_SERVO:  break;    // servo port
        case BIDIB_OUTTYPE_SOUND:  break;    // sound port
        case BIDIB_OUTTYPE_MOTOR:  break;    // motor port
        case BIDIB_OUTTYPE_ANALOG:  break;    // analog port
        default:  break;
      }
    bidib_send_port_na(port);
  }

void bidib_send_keystroke(unsigned char key, unsigned char state)
  {
    t_node_message2 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_KEY;
    message.data[0] = key;
    message.data[1] = state;
    message.header.size = 3+2;                          // 3= sizeof(t_node_message_header)-1
    
    send_bidib_message((unsigned char *)&message);
  }

// -- Macros

static void bidib_send_macro_point(t_macro_point_cfg* point) 
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_MACRO;
    memcpy(message.data, point, sizeof(t_macro_point_cfg));
    message.header.size = 3+sizeof(t_macro_point_cfg);      // 3= sizeof(t_node_message_header)-1
    
    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_macro_para(t_macro_para* para) 
  {
    t_node_message10 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_MACRO_PARA;
    memcpy(message.data, para, sizeof(t_macro_para));
    message.header.size = 3+sizeof(t_macro_para);      // 3= sizeof(t_node_message_header)-1
    send_bidib_message((unsigned char *)&message);
  }

static void bidib_send_macro_state(uint8_t macro_i, uint8_t opcode)
  {
    t_node_message2 message;
    message.header.node_addr = 0;
    message.header.index = get_tx_num();
    message.header.msg_type = MSG_LC_MACRO_STATE;
    message.data[0] = macro_i;
    message.data[1] = opcode;
    message.header.size = 3+2;                          // 3= sizeof(t_node_message_header)-1
    
    send_bidib_message((unsigned char *)&message);
  }

#endif // (CLASS_SWITCH == 1)

//===============================================================================
//
// Parser
//
//===============================================================================

// this routine processes one single message
// parameter: pointer to message (i.e. pointing at length-field of message)
//            message: length, addr, msgnum, msg_type, data
// return:    length of processed data; 128 in case of length error
static unsigned char process_bidib_message (unsigned char *bidib_rx_msg)
  {
    unsigned char length;
    unsigned char *msg_type;

    length = *bidib_rx_msg++;
    if ( (length == 0) || (length & (1<<7)) )
      {
        bidib_send_error(BIDIB_ERR_SIZE, bidib_rx_msg_num); // currently we do not support X-long pakets
        return(128);                                        // pakets with length 0 are void!
      }

    // check address
    
    if (*bidib_rx_msg == 0) 
      {
        bidib_rx_msg++;                                     // broadcast; now point to MNUM
      }
    else if (*bidib_rx_msg == (my_bidib_node_addr & 0x1F))
      {
        bidib_rx_msg++;                            
        if (*bidib_rx_msg == 0)                             // we are the target of this message 
          {
            bidib_rx_msg++;                                 // now point to MNUM

            // check sequence (only if directly addressed)
    
            bidib_rx_msg_num++;
            if (bidib_rx_msg_num == 0) bidib_rx_msg_num++;      // 1,2,..,255,1,2,...
            if (*bidib_rx_msg == 0) bidib_rx_msg_num = 0;       // in case of no MNUM: clear our MNUM
            if (*bidib_rx_msg != bidib_rx_msg_num)
              {
                bidib_send_error(BIDIB_ERR_SEQUENCE, bidib_rx_msg_num);
                bidib_rx_msg_num = *bidib_rx_msg;               // now take this number as reference
              }
          } 
        else
          { 
            bidib_send_node_na(bidib_rx_msg[2]);            // es wurde ein sub-Node adressiert,
            return(length+1);                                 // aber wir haben keinen. -> Fehlermeldung
          }
      }
    else
      {
        return(length+1);                                 // we are not addressed
      }
    // check command

    bidib_rx_msg++;
    msg_type = bidib_rx_msg;
        
    switch(*msg_type)
      {
        default: break;
        //------------------------------------------------- Sys Answers
        case MSG_SYS_GET_MAGIC:
            bidib_send_sys_magic();
            break;
        case MSG_SYS_GET_P_VERSION:
            bidib_send_sys_pversion();
            break;
        case MSG_SYS_ENABLE:
            if (g_bidib_connect == BIDIB_CONNECTED) g_bidib_spontan_enabled = 1;
            break;
        case MSG_SYS_DISABLE:
            if (g_bidib_connect == BIDIB_CONNECTED) g_bidib_spontan_enabled = 0;
            break;
        case MSG_SYS_GET_UNIQUE_ID:
            bidib_send_sys_unique_id();
            break;
        case MSG_SYS_GET_SW_VERSION:
            bidib_send_sys_sw_version();
            break;
        case MSG_SYS_PING:
            bidib_send_sys_pong(msg_type[1]);
            break;
        case MSG_SYS_IDENTIFY:
            bidib_set_sys_identify(msg_type[1]);               // antworten und anzeigen
            bidib_send_onepara_msg(MSG_SYS_IDENTIFY_STATE, msg_type[1]);
            break;
        case MSG_SYS_RESET:                                     // system reset
            stop_bidib_client();
            for (uint8_t i=0; i < 100; i++) _delay_ms(10);
            restart_bidib_client();
            break;
        case MSG_NODETAB_GETALL:
            bidib_node2send = 0;
            bidib_send_node_count(1);
            break;
        case MSG_NODETAB_GETNEXT:
            bidib_send_node_tab_empty();
            break;
        case MSG_GET_PKT_CAPACITY:
            bidib_send_pkt_capacity();
            break;
        case MSG_NODE_CHANGED_ACK:
            // ignored on interface
            break;
        case MSG_SYS_GET_ERROR:
            bidib_send_onepara_msg(MSG_SYS_ERROR, g_bidib_last_error);
            g_bidib_last_error = 0;
            break;
        case MSG_FW_UPDATE_OP:
            bidib_send_bootloader_state(BIDIB_MSG_FW_UPDATE_STAT_EXIT, 0);
            break;
        //------------------------------------------------- Features and Vendor-Config
        case MSG_FEATURE_GETALL:
            bidib_feature2send = 0;
            bidib_send_feature_count(sizeof(feature)/sizeof(feature[0]));
            break;
        case MSG_FEATURE_GETNEXT:
            bidib_send_next_feature();
            break;
        case MSG_FEATURE_GET:
            get_feature(*(msg_type+1));
            break;
        case MSG_FEATURE_SET:
            set_feature(*(msg_type+1), *(msg_type+2));
            break; 
        case MSG_VENDOR_ENABLE:
            if (memcmp (msg_type+1, &my_unique_id, sizeof(my_unique_id)) )
                disable_vendor_config();
            else
                enable_vendor_config();
            break;
        case MSG_VENDOR_DISABLE:
            disable_vendor_config();
            break;
        case MSG_VENDOR_SET:
            vendor_set(msg_type+1);
            break;
        case MSG_VENDOR_GET:
            vendor_get(msg_type+1);
            break;
        case MSG_SYS_CLOCK:
            macro_eval_timecode(msg_type+1);
            break;
        case MSG_STRING_SET:
            user_string_set(msg_type+1);
            break;
        case MSG_STRING_GET:
            user_string_get(msg_type+1);
            break;
        //------------------------------------------------- Class Besetztmelder
        #if (CLASS_OCCUPANCY == 1)
        case MSG_BM_GET_RANGE: 
            // Parameters: start, ende
            get_belegtmeldungen(*(msg_type+1), *(msg_type+2));
            break;
        case MSG_BM_MIRROR_FREE:
            // *(msg_type+1) = Melderadresse
            set_single_free_ack(*(msg_type+1));
            break;
        case MSG_BM_MIRROR_OCC:
            // *(msg_type+1) = Melderadresse
            set_single_belegt_ack(*(msg_type+1));
            break;
        case MSG_BM_MIRROR_MULTIPLE:
            // *(msg_type+1) = Basisadresse
            // *(msg_type+2) = Size  Wertebereich 8.. 128
            // *(msg_type+3)... = Melderdaten: 1...16 Bytes, LSB first
            set_belegt_ack(msg_type+1);
            // jetzt die not acknowledged (=nacked) durchsuchen und melden 
            break;
        case MSG_BM_ADDR_GET_RANGE: 
            // start = *(msg_type+1), ende = *(msg_type+2)
            // get_adressmeldungen(*(msg_type+1), *(msg_type+2));
            // das knnen wir hier nicht
            break;
        #endif
        //------------------------------------------------- Class Accessory
        #if (CLASS_ACCESSORY == 1)
        case MSG_ACCESSORY_SET:
            bidib_do_accessory(*(msg_type+1), *(msg_type+2));
            break;
        case MSG_ACCESSORY_GET:
            bidib_send_accessory_state(*(msg_type+1), BIDIB_ACC_STATE_NO_FB_AVAILABLE, 0);
            break;
        case  MSG_ACCESSORY_PARA_SET:       // 1:anum, 2:para_num, 3..n: data 
            bidib_eval_accessory_mapping(msg_type+1);
            break;
        case  MSG_ACCESSORY_PARA_GET:       // 1:anum, 2:para_num
            bidib_send_accessory_mapping(msg_type+1);
            break;
        #endif
         //------------------------------------------------- Class Switch
        #if (CLASS_SWITCH == 1)
        case MSG_LC_OUTPUT:
            // *(msg_type+1) = PTYPE
            // *(msg_type+2) = PORTNUM
            // *(msg_type+3) = PORTSTAT
              {
                t_macro_ctrl myctrl;
                convert_bidib2macro((t_bidib_port*) (msg_type+1), &myctrl);
                xperform(myctrl.lstate,myctrl.lvalue);
                bidib_send_port_stat((t_bidib_port*) (msg_type+1));
              }
            break;
        case MSG_LC_CONFIG_SET:
            // *(msg_type+1) = Typ
            // *(msg_type+2) = Port
            // *(msg_type+3) = value off
            // *(msg_type+4) = value on
            // *(msg_type+5) = dimm to off
            // *(msg_type+6) = dimm to on
            switch(*(msg_type+1))        // *(msg_type+1) = type
              {
                case BIDIB_OUTTYPE_SPORT:  break;    // standard port
                case BIDIB_OUTTYPE_LPORT: config_light((t_bidib_lport_cfg*) (msg_type+2)); break;    // light port
                case BIDIB_OUTTYPE_SERVO:  break;    // servo port
                case BIDIB_OUTTYPE_SOUND:  break;    // sound port
                case BIDIB_OUTTYPE_MOTOR:  break;    // motor port
                case BIDIB_OUTTYPE_ANALOG:  break;    // analog port
                default:  break;
              }
            bidib_send_port_config((t_bidib_port*) (msg_type+1));
            break;
        case MSG_LC_CONFIG_GET:
            // *(msg_type+1) = Typ
            // *(msg_type+2) = Port
            bidib_send_port_config((t_bidib_port*) (msg_type+1));
            break;
        case MSG_LC_KEY_QUERY:
            // *(msg_type+1) = Port
            bidib_send_keystroke(*(msg_type+1), query_key_state(*(msg_type+1)));
            break;
        case MSG_LC_MACRO_HANDLE:
              { unsigned char result;
                // *(msg_type+1) = Macro
                // *(msg_type+2) = Opcode: BIDIB_MACRO_OFF, _RUN,_SAVE, _DELETE,
                result = macro_handle(*(msg_type+1), *(msg_type+2));
                bidib_send_macro_state(*(msg_type+1),result);
              }
            break;
        case MSG_LC_MACRO_SET:
            // *(msg_type+1) = Macro
            // *(msg_type+2) = index
            // *(msg_type+3) = delay
            // *(msg_type+4) = port
            // *(msg_type+5) = command
              {
                t_macro_point_cfg point;         // nicht casten (wenn da die Parameter fehlen wuerden - crash)
                memcpy(&point, (msg_type+1), sizeof(t_macro_point_cfg));
                macro_set_point(&point);
                bidib_send_macro_point(&point);
              }
            break;
        case MSG_LC_MACRO_GET:
              {
                t_macro_point_cfg point;            // allocate here
                point.macro_i = *(msg_type+1);
                point.index = *(msg_type+2);
                macro_get_point(&point);
                bidib_send_macro_point(&point);
              }
            break;
        case MSG_LC_MACRO_PARA_SET:
              {
                t_macro_para para;
                memcpy(&para, (msg_type+1), sizeof(t_macro_para));
                macro_para_set(&para);
                bidib_send_macro_para(&para);
              }
            break;
        case MSG_LC_MACRO_PARA_GET:
              {
                t_macro_para para;            // allocate here
                para.macro_i = *(msg_type+1);
                para.index = *(msg_type+2);
                macro_para_get(&para);
                bidib_send_macro_para(&para);
              }
            break;
        #endif
        //------------------------------------------------- BiDiBus (local)
        case MSG_LOGON_ACK:  
            // Es folgt ein Byte mit der lokalen Adresse (NODE_ADDR) bidib_rx_msg[4]
            // und 7 Bytes mit der UniqueID bidib_rx_msg[5] .... 
            // Nur wenn der Knoten die emfangene UniqueID identisch zu 
            // seiner internen UniqueID geprft hat, darf er seine lokale 
            // Adresse auf die empfangene NODE_ADDR einstellen. 
            // Das hier ist unabhaengig von 'Applied', damit kann der Host
            // (sofern er sich das gemerkt hat) die Knoten schnell zuweisen.
            if (memcmp (msg_type+2, &my_unique_id, sizeof(my_unique_id)) == 0)
              {
                unsigned char temp;                      // token parity checking 
                temp = *(msg_type+1) ^ (*(msg_type+1) << 4);
                temp = temp ^ (temp << 2);
                temp = temp ^ (temp << 1);
                temp = temp & 0x80;         // parity im MSB
                my_bidib_node_addr = *(msg_type+1) | temp;
                set_bidib_state(BIDIB_CONNECTED);
              }
            break;
        case MSG_LOGON_REJECTED:
            if (memcmp (msg_type+1, &my_unique_id, sizeof(my_unique_id)) == 0)
              {
                set_bidib_state(BIDIB_REJECTED);
              }
            break;
            break;
      }
    return(length + 1);  // das Laengenbyte selbst mit beruecksichtigen
  }

#if (1==0)
#warning
void test_process_bidib_message(unsigned char *test)
  {
    my_bidib_node_addr = 1;
    test[1] = my_bidib_node_addr;
    process_bidib_message(test);
  }
#endif

//===============================================================================
//
// RX Message Collector
//
//===============================================================================
static unsigned char bidib_rx_paket[64];
static unsigned char bidib_rx_index = 0;  // the location where we store next data
static unsigned char bidib_rx_total = 0;  // the samples we got 
static unsigned char bidib_rx_crc = 0;    // accumulated crc


//===============================================================================
//
// Messages for Occupancy Class (Query and Push Service)
//
//===============================================================================

//===============================================================================
//
// Messages (Upstream Service)
//
//===============================================================================



//===============================================================================
//
// Client Task
//
//===============================================================================

// internal and static
enum bidib_rx_states
  {                                     // actual state for the BiDiB Task
    BIDIB_IDLE,
    BIDIB_GET_LEN,
    BIDIB_COLLECT_MESSAGE,
  } bidib_rx_state;

static int bidib_active_timestamp;


//------------------------------------------------------------------------
//
//  Connect Status
//
//------------------------------------------------------------------------
//
void bidib_set_sys_identify(unsigned char state)
  {
    g_bidib_identify = state;
    if (state)
      {
        led_fast_blinking(LED_PROG);                   // show identify
      }
    else
      {
        led_half(LED_PROG);                    // normal mode
      }
  }

// Hinweis: bidib_connect wird auch im client_if manipuliert, hier nur
//          die richtige Reaktion auf disconnected
void set_bidib_state(bidib_connect_t neu)
  {
    if (neu != g_bidib_connect)
      {
        g_bidib_connect = neu;
        // cr_fifo_put(&key_fifo, KEY_CONNECT);
        // cr_fifo_put(&key_fifo, KEY_PARSER_STATE);
        switch (g_bidib_connect)
          {
            default:
            case BIDIB_DISCONNECTED:      
                my_bidib_node_addr = 0xFF;              
                g_bidib_spontan_enabled = 0;
                bidib_flush_rx();                       // Empfangspuffer
                bidib_rx_state = BIDIB_IDLE;
                bidib_flush_tx();                       // Sendepuffer leeren und
                                                        // eine LOGON-message vorbereiten
                bidib_tx0_msg_num = 0;
                bidib_feature_transmit = 0;
                bidib_set_sys_identify(0);
                g_bidib_backoff = (unsigned char)(seed+2) & 0x3F;          // 2..63
                g_bidib_last_backoff = g_bidib_backoff;
                g_bidib_ca_time = seed & 0x1F;          // 0..31
                led_off(LED_MSG);
                break;
            case BIDIB_APPLIED:
                break;
            case BIDIB_CONNECTED:
                led_on(LED_MSG);
                break;
            case BIDIB_REJECTED:
                led_XooooXoooo(LED_MSG);
          }
      }
  }

//------------------------------------------------------------------------
//
//  Messagehandling
//
//------------------------------------------------------------------------
//
// Diese Routine parst ein komplettes Paket, bestehend aus
// PLENGTH {MLENGTH ADDR MNUM TYPE data} 
static uint8_t bidib_parser(void)
  {
    uint8_t Actioned = 0;

    // we got a paket with messages in it, lets loop thro all messages
    
    bidib_rx_total = bidib_rx_paket[0];

    #if (DEBUG_RX == 1)
        uint8_t i;
        for (i=0; i<= bidib_rx_total; i++)
          {
            pc_send_hex(bidib_rx_paket[i]);
            pc_send('.');
          }
        pc_send_crlf();
    #endif

    bidib_rx_index = 1;  // skip paket_length

    while(bidib_rx_index < bidib_rx_total)
      {
        bidib_rx_index += process_bidib_message(&bidib_rx_paket[bidib_rx_index]);
      }
    return(Actioned);
  }


//------------------------------------------------------------------------
//  This is a cooperative task, see cortos.c
//  It does:
//    1. get as many as available samples from bidib
//    2. trace down the link layer (see bidib_rx_state)
//    3. if a valid packet for our ID is detected:
//       call the parser
//
//  It should be called as often as possible
//------------------------------------------------------------------------


t_cr_task run_bidib_client(void)
  {
    t_data16 xdata;


    if ((signed int)(get_systick_int16() - bidib_active_timestamp) > (500000L / SYSTICK_PERIOD) )
      {
        // Busmaster ist seit mehr als 500ms tot - wir loggen aus
        set_bidib_state(BIDIB_DISCONNECTED);
        bidib_active_timestamp = get_systick_int16();
      }

    #warning fifo full, msg_stall?
    if (!bidib_tx_fifo_healthy()) return(0);                    // Ausstieg: unser upstream ist geblockt
                                                                // wir koennen nichts beantworten

    while (bidib_rx_ready())
      {
        // Hinweis: Hier ueberwachen wir die allgemeine Hostaktivitaet (und nicht
        //          das ansprechen unserer Nodeadresse)
        bidib_active_timestamp = get_systick_int16(); 
        // set_bidib_state(BIDIB_CONNECTED);          
        
        xdata.as_uint16 = bidib_rx_read();                      // reads one byte
        if (xdata.high8)
          { // new poll byte
            if (xdata.low8 & 0x40)
              {
                // TRIGA;
                // System Byte: not processed here - see client if 
                #if (1==0)
                switch (xdata.low8)
                  {
                    case BIDIBUS_POWER_UP_par:                   // Bus Reset (including parity)
                        //set_bidib_state(BIDIB_DISCONNECTED);     // not processed anyway!
                        break;
                    case BIDIBUS_LOGON_par:                      // Logon Prompt (including parity)
                        switch (g_bidib_connect)
                          {
                            default:
                            case BIDIB_REJECTED:
                                break;
                            case BIDIB_DISCONNECTED:
                                // we should try to apply -> this is done in client_if
                                break;
                            case BIDIB_APPLIED:
                                // we have applied, but there was no ACK for us 
                                // backoff neu berechnen und wieder DISCONNECTED
                                break;
                            case BIDIB_CONNECTED:
                                break;
                          }
                        break;
                  }
                #endif
              }
            else
              {
                // unsigned char bidib_current_node;               // das ist die Quelle der aktuellen Nachricht
                // bidib_current_node = xdata.low8 & 0x1F;
                if (xdata.low8 == 0) // if (bidib_current_node == 0) // including parity
                  {
                    bidib_rx_state = BIDIB_GET_LEN;     // message from host, grab it
                  }
                else
                  {
                    bidib_rx_state = BIDIB_IDLE;        // other slave to host, ignore it
                  }
              }
          }
        else // standard byte
          {
            switch(bidib_rx_state)
              {
                case BIDIB_IDLE:
                    // hier den Empfang ignorieren, wir sind 'off sync'
                    break;
                case BIDIB_GET_LEN:
                    bidib_rx_crc = read_crc_array(xdata.low8);
                    bidib_rx_paket[0] = xdata.low8;
                    bidib_rx_total = xdata.low8 + 1;
                    bidib_rx_index = 1;
                    bidib_rx_state = BIDIB_COLLECT_MESSAGE;
                    break;
                case BIDIB_COLLECT_MESSAGE:
                    bidib_rx_crc = read_crc_array(xdata.low8 ^ bidib_rx_crc);
                    if (bidib_rx_index == bidib_rx_total)
                      {
                        if (bidib_rx_crc)
                          {
                            // Nachricht hatte falsche CRC
                            // Hinweis: das ist eine Nachricht vom Host. Da wir
                            // nicht wissen, wer der Adressat ist, werfen wir das weg.
                            bidib_rx_state = BIDIB_IDLE;
                          }
                        else
                          {
                            bidib_parser();
                          }
                      }
                    else
                      {
                        bidib_rx_paket[bidib_rx_index++] = xdata.low8;
                        
                        if (bidib_rx_index == sizeof(bidib_rx_paket))
                          {
                            // Nachricht ist zu lang, einfach verwerfen!
                            bidib_rx_state = BIDIB_IDLE;
                          }
                      }
              }
          }
      }
    return(1); // everything processed - no more ready
                // The incoming fifo will issue a wakeup.
  }



void restart_bidib_client(void)
  {
    init_bidib_client_if();                         // init low level
    bidib_rx_state = BIDIB_IDLE;

    g_bidib_connect = 0xFF;                         // set void to force a change of state
    set_bidib_state(BIDIB_DISCONNECTED);
    bidib_active_timestamp = get_systick_int16();   // reset timing control
  }

void stop_bidib_client(void)
  {
    // disable interrupts
    bidib_soft_stop();
    stop_bidib_client_if();
    // the consumer task will terminate itself
  }

void init_bidib_client(void)
  {
    init_features_from_eeprom();                    // features aus eeprom lesen

    restart_bidib_client();
  }



