//==================================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2012 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      config.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-03-01 V2.00 kw start
//            2008-08-08 V2.03 kw added BiDi
//            2012-10-16 V2.04 kw added accessory
//            2013-04-13 V2.05 kw added MSG_REJECTED
//            2013-07-24 V2.06 kw changed version to 1.4.0
//            2013-10-10 V2.07 kw version now 1.5.0 -> accessory_feedback
//------------------------------------------------------------------------
//
// purpose:   flexible general purpose decoder for dcc
//            This is the central definition file for the project
//
//            Note: for Atmega162 and Bootloader: code limit is 14436
//------------------------------------------------------------------------
//
// content:   1. Project settings
//               1.a) General
//               1.b) Configuration of Software Modules
//
//            2. EEPROM definitions (CV's)
//            3. Global variables
//            4. Useful inline routines (delay, Semaphor-operations)
//
//==================================================================================
//
#ifndef _CONFIG_H_
#define _CONFIG_H_

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "datatypes.h"

#include "version.h"  // this brings us: MAIN_VERSION, SUB_VERSION, COMPILE_RUN
                      //                 BUILD_YEAR, BUILD_MONTH, BUILD_DAY




#if (__AVR_ATmega162__)
    #define DCC_ACK         PB0       // output, sending 1 makes an ACK
    #define DCC_ACK_OFF     PORTB &= ~(1<<DCC_ACK)
    #define DCC_ACK_ON      PORTB |= (1<<DCC_ACK)

    #define DCC_BIT         PD2     // must be located on INT0
    #define DCC_PORT        PORTD   // must be defined to have portable code
    #define DCC_PORT_IN     PIND    // must be defined to have portable code

    #define INT_CONF_REG    MCUCR

    #define EEPROM_SIZE     512     // define to allocate serial-ID
#elif (__AVR_ATmega32__)
    #define DCC_ACK         PB0       // output, sending 1 makes an ACK
    #define DCC_ACK_OFF     PORTB &= ~(1<<DCC_ACK)
    #define DCC_ACK_ON      PORTB |= (1<<DCC_ACK)

    #define DCC_BIT         PD2     // must be located on INT0
    #define DCC_PORT        PORTD   // must be defined to have portable code
    #define DCC_PORT_IN     PIND    // must be defined to have portable code

    #define INT_CONF_REG    MCUCR

    #define EEPROM_SIZE     1024     // define to allocate serial-ID
#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)
    #define EEPROM_SIZE     2048     // define to allocate serial-ID
#endif

//==================================================================================
//
// 1. Project Definitions
//
//==================================================================================
//
// 1.a) General
//



#define SERIAL           0x0001   // This is a 16-bit serial number used to identify different boxes.

#define FALSE  0
#define TRUE   (!FALSE)

//
// 1.b) Simulation and Debugging
// Debugswitches:  0: no debugging, interrupts running -> alive!
//                 1: no ints, some lowlevel test

#define SIMULATE_BIDIB_LOGGED_ON    0  // if > 0, we act as already logged on

#define DEBUG  0


// GLOBALS

//-----------------------------------------------------------------
//
// 1.b) Configuration of installed hardware / software
//
//-----------------------------------------------------------------

// Fuses:
//      Extended: FB
//      High:     D9
//      Low:      EE


#define LCD_EA204           1       // fully supported
#define LCD_REICHELT162     2       // not supported
#define LCD_EA128x64        3       // fully supported


#define DISPLAY             LCD_EA204
// #define DISPLAY             LCD_EA128x64


#define LCD_IF_SPI          1       // connect on SPI bus
#define LCD_IF_4BIT         2

#define DISPLAY_INTERFACE   LCD_IF_SPI



//==================================================================================
//
// 2. EEPROM Definitions (CV's)
//
//==================================================================================
//
// content is defined in config.c

#define STANDARD                    1       // any number
#define TURNOUT                     2       // any number


#ifndef MOBALIST_MODEL
    #warning no MOBALIST_MODEL defined, use STANDARD as default
    // note: this is supplied with a -D statement in the commandline 
    #define MOBALIST_MODEL  STANDARD
#endif


#if (MOBALIST_MODEL == STANDARD)
    #define NUM_OF_MACROS                12      // Modell 'Standard'
    #define NUM_OF_SAVED_MACROS          12
    #define NUM_OF_ITEMS_PER_MACRO       18
    #define NUM_OF_ACCESSORY              8  // max. 8 Signale
    #define NUM_OF_ASPECTS_PER_ACCESSORY  8  // max. 8 Begriffe je Signale
#elif  (MOBALIST_MODEL == TURNOUT)
    #define NUM_OF_MACROS                24      // Modell 'Turnout' (Weiche)
    #define NUM_OF_SAVED_MACROS          24
    #define NUM_OF_ITEMS_PER_MACRO        6
    #define NUM_OF_ACCESSORY             12  // max. 12 Accessory
    #define NUM_OF_ASPECTS_PER_ACCESSORY  8  // max. 8 Begriffe je Signale
#else
    #error MOBALIST_MODEL is void
#endif


#define NUM_OF_FEATURES            20      // so viele features im eeprom moeglich

#define BIDIB_STRING_MAX           24       // Zeichenlnge fr Strings -> macht dann 25 CVs


//==================================================================================
//
// 3. Global configs
//
//==================================================================================




//----------------------------------------------------------------------------

typedef struct
  {
    unsigned char minute; 
    unsigned char hour;
    unsigned char day_of_week;
    unsigned char ratio;
  } t_fast_clock;

t_fast_clock  fast_clock;


extern volatile signed char timerval;    // gets incremented in the timetick



//==================================================================================
// 4. Platform
//==================================================================================
//
// Board

#define  BOARD_MFT          1       // Version 1.0 or 1.1 of MyFirstThrottle
#define  BOARD_GATEWAY0     100 
#define  BOARD_GATEWAY1     101 
#define  BOARD_XP_MELDER    13 
#define  BOARD_BIDIB_S88    14 
#define  BOARD_MOBALIST     15 

#define  HW_VERSION         BOARD_MOBALIST


#define BOARD BOARD_MOBALIST   // XP/BIDIB connected to UART (there is onyl one UART)


#define T0_PRESCALER   8    // may be 1, 8, 64, 256, 1024


//-------------------------------------------------------------------------
// 4.a) Definitions for inter process communication
//-------------------------------------------------------------------------
// These routines are *inline*, so we keep them in the header.
//
// usage:   semaphor_set(flag, state) to do the communication.
//          semaphor_get(flag)

#define SEMAPHOR_IO 1

#if (SEMAPHOR_IO == 1)
  #define Communicate TWBR                // use the TWBR for Flags
#else
  extern volatile unsigned char Communicate;
#endif

#define C_Received        0    // a new DCC was received - issued by ISR(Timer1)
                               //                          cleared by main
#define C_DoSave          1    // a new PORT state should be saved
                               //                        - issued by action
                               //                          cleared by main
  
#define C_PWM_Update      2    // new light[i].volumes are there

#define C_PWM_Buffer      3    // Flag, welcher Buffer von der ISR benutzt wird


//========================================================================
// 4.b) Useful inline code
//========================================================================

static inline unsigned char semaphor_query(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_query(unsigned char flag)
  {
    return (Communicate & (1<<flag));
  }
               
static inline void semaphor_set(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_set(unsigned char flag)
  {
    cli();
    Communicate |= (1<<flag);
    sei();
  }
       
static inline unsigned char semaphor_get(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_get(unsigned char flag)
  {
    unsigned char value;
    cli();
    value = Communicate & (1<<flag);
    Communicate &= (~(1<<flag));
    sei();
    return(value);
  }

static inline unsigned char isr_semaphor_get(unsigned char flag) 
       __attribute__((always_inline));

unsigned char isr_semaphor_get(unsigned char flag)
  {
    unsigned char value;
    value = Communicate & (1<<flag);
    Communicate &= (~(1<<flag));
    return(value);
  }
        

//------------------------------------------------------------------------
// Delay-Macro (all values in us) -> this is busy waiting
//------------------------------------------------------------------------
//
// same macro as in util/delay.h, but use up to some ms.
// The maximal possible delay is 262.14 ms / F_CPU in MHz.
// This is 16ms for 16MHz; longest used delay: 1000us

#ifndef _UTIL_DELAY_H_
  #include <util/delay.h>
#endif

#define nop() __asm volatile ("nop") 

static inline void _mydelay_us(double __us) __attribute__((always_inline));
void
_mydelay_us(double __us)
{
    uint16_t __ticks;
    double __tmp = ((F_CPU) / 4e6) * __us;
    if (__tmp < 1.0)
        __ticks = 1;
    else if (__tmp > 65535)
        __ticks = 0;    /* i.e. 65536 */
    else
        __ticks = (uint16_t)__tmp;
    _delay_loop_2(__ticks);
}   

static inline void _restart(void) __attribute__((always_inline));
void
_restart(void)
{
    cli();
                    
    // laut diversen Internetseiten sollte folgender Code laufen -
    // tuts aber nicht, wenn man das Assemblerlistung ansieht.
    // void (*funcptr)( void ) = 0x0000;    // Set up function pointer
    // funcptr();                        // Jump to Reset vector 0x0000
    
    __asm__ __volatile 
    (
       "ldi r30,0"  "\n\t"
       "ldi r31,0"  "\n\t"
       "icall" "\n\t"
     );
}

//------------------------------------------------------------------------
// Config - EEPROM
//------------------------------------------------------------------------
//
// GW uses EEPROM for Configuration
//
// To achieve a fixed address schema, new sections are introduced - these
// must be allocated during linker run.
// -> see config file of AVR Studio, Memory Sections
//
// ATTENTION: There is a bug in AVR Studio - the memory options are not exported
//            to avr-objcopy, therefore the new sections appear in opendcc.hex,
//            not in opendcc.eep.
// SOLUTION:  manually edit makefile:
//
//             ## Intel Hex file production flags
//             HEX_FLASH_FLAGS = -R .eeprom
//             HEX_FLASH_FLAGS = -R .ee_my_new_sect    <-- add this line
//
// Sections:    Size:   Location:   Content:
// EEMEM                810000      General Config 
// EE_***        128    810080      

typedef struct
  {
    unsigned char ee_version;          //  1  Version (of eeprom content)
    unsigned char vid;                 //  2  Vendor ID (0x0D = DIY Decoder, 0x12 = JMRI, 0x3E = TAMS)
    unsigned char pid_l;               //  3  Product ID
    unsigned char pid_h;               //  4  Product ID
    unsigned char main_version;        //  5  software
    unsigned char sub_version;         //  6  software
    unsigned char compile_run;         //  7  software
    unsigned char usernick[BIDIB_STRING_MAX + 1];       //  8 .. 32
    unsigned char res1;                // 33  reserved
    unsigned char res2;                // 34  reserved
    unsigned char res3;                // 35  reserved
    unsigned char res4;                // 36  reserved
    unsigned char res5;                // 37  reserved
    unsigned char res6;                // 38  reserved
    unsigned char res7;                // 39  reserved
    unsigned char res8;                // 40  reserved
  } t_bidib_cv;

typedef struct
  {
    unsigned char current_aspect;
    unsigned char mapped_macro[NUM_OF_ASPECTS_PER_ACCESSORY];
  } t_accessory_map;


typedef struct
  {
    t_bidib_cv bidib;

    unsigned char feature[NUM_OF_FEATURES];          //features

    t_accessory_map accessory_map[NUM_OF_ACCESSORY];
  }
t_cv_record;

extern t_cv_record CVram;

extern t_cv_record CV EEMEM;

extern const t_cv_record CV_PRESET PROGMEM;

#endif   // _config_h_
