//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//
//------------------------------------------------------------------------------
//
// OpenDCC - Simple Cooperative Task Scheduler
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//-----------------------------------------------------------------------------
//
// file:      cortos.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-01-12 V0.01 kw  started
//            2009-03-02 V0.02 kw  LED control with compile-switch
//            2009-10-09 V0.03 kw  default task state is blocked
//
//-----------------------------------------------------------------------------
//
// purpose:   A small and tiny cooperative scheduler
//
//------------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>

#include "datatypes.h"
#include "timer.h"          				// access to systick
#include "cortos.h"

#if (DEBUG != 0)                    // generell debug allowed
    #define DEBUG_CORTOS   0        // 0: no debug messages
                                    // 1: LED turns on if not sleeping 
#endif
//------------------------------------------------------------------------------
//
// Fifo
//
//------------------------------------------------------------------------------

void cr_fifo_init (cr_fifo_t *f, uint8_t *buffer, const uint8_t size)
{
    f->count = 0;
    f->task_to_wakeup = 0;
    f->pread = f->pwrite = buffer;
    f->read2end = f->write2end = f->size = size;
}

uint8_t cr_fifo_put (cr_fifo_t *f, const uint8_t data)
{
    return _inline_fifo_put (f, data);
}

uint8_t cr_fifo_get_wait (cr_fifo_t *f)
{
    while (!f->count);
    
    return _inline_fifo_get (f);    
}

int cr_fifo_get_nowait (cr_fifo_t *f)
{
    if (!f->count)      return -1;
        
    return (int) _inline_fifo_get (f);  
}

uint8_t cr_fifo_filled (cr_fifo_t *f)
{
    return (f->count);  
}

void cr_fifo_set_comsumer (cr_fifo_t *f, t_task_id consumer)
{
    f->task_to_wakeup = consumer;
}

//------------------------------------------------------------------------------
//
// Task
//
//------------------------------------------------------------------------------
// the list of all tasks is found in cortos_tasklist.c


#ifndef set_task_ready
#warning this should be a define
void set_task_ready(t_task_id ready_to_run)
  {
    tasklist[ready_to_run].ready = 1;
    tasklist[ready_to_run].wakeup = get_systick_int16();  // ready now!
  }
#endif

#ifndef set_task_blocked
#warning this should be a define
void set_task_blocked(t_task_id ready_to_run)
  {
    tasklist[ready_to_run].ready = 0;
    // attention: do not use set_task_blocked from inside the task which is to be blocked
    //            the normal return value will supersede this!
  }
#endif


volatile signed int my_systick;


// this checks all task in a loop (top-down) and calls all tasks which are
// marked as ready; typically the weakest task is 'idle' and sets the processor
// to sleep mode.
// When a interrupt (normally the timer) occurs, the processor will wake up,
// terminate the idle and will run again.

void scheduler(void)
  {
    signed char t_id;
    // unsigned char any_task_ready = 0;
    signed int new_delay;

    // signed int my_systick;

    my_systick = get_systick_int16();  // atomic!

    for (t_id = (sizeof(tasklist) / sizeof(tasklist[0]) -1); t_id >= 0; t_id--)
      {
        if (tasklist[t_id].ready)
          {
            // any_task_ready = 1;
            if ((my_systick - tasklist[t_id].wakeup) >= 0)
              {
                // before calling the task: we take away the ready.
                // reason: if a wake up occurs with an ISR just in the moment
                //         when we get the negativ return value - we could loose an event

                tasklist[t_id].ready = 0;
                new_delay = tasklist[t_id].my_call();
                if (new_delay < 0)                      // if a task returns negativ - it get blocked
                  {
                    // do nothing - task is already blocked
                  }
                else if (new_delay == 0)
                  {
                    tasklist[t_id].ready = 1;
                    tasklist[t_id].wakeup = my_systick;
                  }
                else
                  {
                    tasklist[t_id].ready = 1;
                    tasklist[t_id].wakeup = tasklist[t_id].wakeup + new_delay;
                  }
              }
          }
      }
  }

#include "hardware.h"

t_cr_task task_idle (void)
  {
    set_sleep_mode(SLEEP_MODE_IDLE);
    #if (DEBUG_CORTOS == 1) 
        // LED_D_OFF;
    #endif
    sleep_mode(); 
    #if (DEBUG_CORTOS == 1) 
        // LED_D_RED;
    #endif
    return(0);     // ready forever
  }
