//------------------------------------------------------------------------------
//
// OpenDCC - Cooperative Simple Task Scheduler
//
// Copyright (c) 2008, 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//-----------------------------------------------------------------------------
//
// file:      cortos.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-12 V0.1 kw started
//
//-----------------------------------------------------------------------------
//
// purpose:   A small and tiny cooperative scheduler
//
//------------------------------------------------------------------------------
// Interface Guide:
//  #include "cortos.h"
//  #define BUF_SIZE 8
//  uint8_t buffer[BUF_SIZE];
//  cr_fifo_t my_fifo;
//  ...
//  fifo_init (&my_fifo, buffer, BUF_SIZE);       
//  ...
//  Producer:   uint8_t zeichen = ...;
//              cr_fifo_put (&fifo, zeichen);
//  Consumer:   zeichen = cr_fifo_get_wait (&fifo);
//------------------------------------------------------------------------------


#ifndef _CORTOS_H_
#define _CORTOS_H_

#include <avr/io.h>
#include <avr/interrupt.h>

#include "datatypes.h"

//
// NOW include the project parts
//
#include "cortos_tasklist.h"
#include "timer.h"

//------------------------------------------------------------------------------
//
// Tasks
//
//------------------------------------------------------------------------------

// this is generic pointer to a task
// param: void
// ret value: delay in systick; 
// if < 0: task wants to sleep forever (is not ready anymore)
//         (this requires that the task is woken up by 'set_task_ready')
// if = 0: task is ready again and will be called again asap.
// if > 0: task wants to sleep for some time. 
//         This value defines a grid, not an actual delay.
//         The grid is systick (typically 1,6ms).
//         i.e. for 50ms:  'return(50000L / SYSTICK_PERIOD);'

typedef signed int t_cr_task;      // Tasks are returning signed int.

typedef t_cr_task (*Ptr_To_Task)(void);

typedef struct 
  {
    unsigned char ready;     // if true: ready to run
    signed int wakeup;       // desired wakeup time
    Ptr_To_Task my_call;     // Task
  } t_task;

extern t_task tasklist[size_of_tasklist];

//-------------------------------------------------------------------
// calls:
// whenever somebody fills a queue or fifo,
// it must enable the consumer_task (see fifo routines)
//

#define set_task_ready(ready_to_run)              \
  {                                                         \
    tasklist[ready_to_run].wakeup = get_systick_int16();      \
    tasklist[ready_to_run].ready = 1;                       \
  }

#define isr_set_task_ready(ready_to_run)              \
  {                                                         \
    tasklist[ready_to_run].wakeup = isr_get_systick_int16();      \
    tasklist[ready_to_run].ready = 1;                       \
  }

#define set_task_blocked(ready_to_run)  tasklist[ready_to_run].ready = 0


// void set_task_ready(t_task_id ready_to_run);
// void set_task_blocked(t_task_id ready_to_run);



void scheduler(void);             // to be called in a loop


//------------------------------------------------------------------------------
// Fifo (for byte objects)
//------------------------------------------------------------------------------

typedef struct
{
    uint8_t volatile count;       // # Zeichen im Puffer
	uint8_t size;                 // Puffer-Groesse
	uint8_t *pread;               // Lesezeiger
	uint8_t *pwrite;              // Schreibzeiger
	uint8_t read2end, write2end;  // # Zeichen bis zum Ueberlauf Lese-/Schreibzeiger
    uint8_t task_to_wakeup;       // ID of task to set ready when fifo gets filled
} cr_fifo_t;

extern void    cr_fifo_init (cr_fifo_t*, uint8_t* buf, const uint8_t size);
extern uint8_t cr_fifo_put (cr_fifo_t*, const uint8_t data);
extern uint8_t cr_fifo_get_wait (cr_fifo_t*);
extern int     cr_fifo_get_nowait (cr_fifo_t*);
extern uint8_t cr_fifo_filled (cr_fifo_t *f);
extern void    cr_fifo_set_comsumer (cr_fifo_t *f, t_task_id consumer);


static inline uint8_t
_inline_fifo_put (cr_fifo_t *f, const uint8_t data)
{
	if (f->count >= f->size)
		return 0;
		
	uint8_t * pwrite = f->pwrite;
	
	*(pwrite++) = data;
	
	uint8_t write2end = f->write2end;
	
	if (--write2end == 0)
	{
		write2end = f->size;
		pwrite -= write2end;
	}
	
    f->write2end = write2end;
	f->pwrite = pwrite;

	uint8_t sreg = SREG;
	cli();
	f->count++;
    isr_set_task_ready(f->task_to_wakeup);

	SREG = sreg;
	
	return 1;
}

static inline uint8_t 
_inline_fifo_get (cr_fifo_t *f)
{
	uint8_t *pread = f->pread;
	uint8_t data = *(pread++);
	uint8_t read2end = f->read2end;
	
	if (--read2end == 0)
	{
		read2end = f->size;
		pread -= read2end;
	}
	
	f->pread = pread;
	f->read2end = read2end;
	
	uint8_t sreg = SREG;
	cli();
	f->count--;
	SREG = sreg;
	
	return data;
}

//
// NOW include the project parts
//
#include "cortos_fifolist.h"


#endif /* _CORTOS_H_ */

