//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
//
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//===============================================================================
//
// file:      hardware.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-12-16 V0.01 started
//            some lines from www.mobalist.de
//
//===============================================================================
//
// purpose:   Interface s88-BiDiB
// content:   central definitions used in the project
//            all hardware project settings are done here!!
//            
//            1.   Prozessor, Timing
//            2.   IOs
//
//===============================================================================
//

#ifndef __HARDWARE_H__
#define __HARDWARE_H__

#include "datatypes.h"

//========================================================================
// 1. Processor Definitions
//========================================================================
//

// PROCESSOR: one of: __AVR_ATmega162__

// if changed: check timings, timer settings, baudrate

#if (__AVR_ATmega32__)
  // #define INTERNAL_SRAM_SIZE    1024
  // #define EEPROM_SIZE           512
#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)
#else
  #warning: severe: no supported processor  
#endif
//
#ifndef F_CPU
// prevent compiler error by supplying a default 
# warning "F_CPU not defined for <hardware.h>", set default to 16MHz 
# define F_CPU 16000000UL
// if changed: check every place where it is used (possible range overflow in preprocessor!)
#endif

// ????
//------------------------------------------------------------------------
// Fuses of ATmega162 (just for documentation)
//------------------------------------------------------------------------
// Lock Bit Byte
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  1       1       1       1       1       1       1       1       default, not locked
//
// Extended Fuse Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//                                  BOD2    BOD1    BOD0    TA0SEL  111=no BOD (*), 101=2,7V, 100=4,3V
//  1       1       1       1       0       1       0       1
//
// Fuse High Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  OCDEN   JTAGEN  SPIEN   WDTON   EESAVE  BOOTSZ1 BOOTSZ0 BOOTRST
//  1       1       0       1       1       1       0       0
//  disabl  disabl  enabl   default no save =0xFC00         enter boot
//
// Fuse Low Byte:
//  7       6       5       4       3       2       1       0 
//  ------  ------  ------  ------  ------  ------  ------  ------
//  CKDIV8  CKOUT   SUT1    SUT0    CKSEL3  CKSEL2  CKSEL1  CKSEL0
//  1       1       0       0       1       1       1       0
//  no div  no clk  crystal, BOD ena, 10 MHz    




//========================================================================
// 2. Port Definitions
//========================================================================


// Port A                                       // S88 (Master)

// LEDs 8..15

// Port B                                       // Host UART (CH0)
                                            
// LED 7.. 0 (order invers)

// Port C

// LEDs 23..16 (order invers)

// Port D                                       // XP UART (CH1)
  
#define J6          7     // in (OPT2)  -> Keyboard
#define J5          6     // in (OPT1)  -> Keyboard
#define J4          5     // in (PROG)

#define LEDp_PROG   4     // out, low=on
#define LEDp_MSG    3     // out, low=on

#define BIDIB_DIR   2     // out; 0=receive
#define BIDIB_TX    1     // out       
#define BIDIB_RX    0     // in

//------------------------------------------------------------------------
// LED-Control and IO-Macros

#define LED_MSG_OFF()        PORTD |= (1<<LEDp_MSG) 
#define LED_MSG_ON()       PORTD &= ~(1<<LEDp_MSG)

#define LED_PROG_OFF()        PORTD |= (1<<LEDp_PROG) 
#define LED_PROG_ON()       PORTD &= ~(1<<LEDp_PROG)

#define JUMPER4_FITTED()    (!(PIND & (1<<J4)))
#define JUMPER5_FITTED()    (!(PIND & (1<<J5)))
#define JUMPER6_FITTED()    (!(PIND & (1<<J6)))


//------------------------------------------------------------------------


// defines von Micha

#define F_PWM 100                   	 								// PWM-Frequenz in Hz
#define PWM_STEPS 255               	 								// PWM-Schritte pro Zyklus(1..255)
#define PWM_PORT1 PORTB                  								// Port fr PWM1
#define PWM_DDR1 DDRB                    								// Datenrichtungsregister fr PWM1
#define PWM_PORT2 PORTA                  								// Port fr PWM2
#define PWM_DDR2 DDRA                    								// Datenrichtungsregister fr PWM2
#define PWM_PORT3 PORTC                  								// Port fr PWM3
#define PWM_DDR3 DDRC                  									// Datenrichtungsregister fr PWM3


// Calls
void init_hardware(void);
void init_interrupt_controller(void);


#endif   // #ifndef __HARDWARE_H__

