//===============================================================================
//
// OpenDCC - Keyboard
//
// Copyright (c) 2010 Wolfgang Kufer
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-------------------------------------------------------------------------------
//
// file:      keyboard.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-11-27 V0.01 kw start
//            2011-01-11 V0.02 kw angepasst auf s88-BiDiB
//
//-------------------------------------------------------------------------------
//
// purpose:   keyboard
//
// content:   
//             1. here keyboard routines for two different ports
//                (configurable with define, events for press and release)
//        
//===============================================================================

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "datatypes.h"
#include "cortos.h"
#include "hardware.h"
#include "timer.h"
#include "keycodes.h"

#ifndef KEYBOARD_ENABLED
  #define KEYBOARD_ENABLED   TRUE
#endif
#if (KEYBOARD_ENABLED == TRUE)

#define SIMULATION  0               // 0: real application
                                    // 1: test init routine
                                    // 2: test timing engine
                                    // 3: test action
#if (SIMULATION != 0)
   #warning SIMULATION is on!
#endif

//------------------------------------------------------------------------------
// --------------- Port1:
#define KEY1_PART           1            // if 1: use this definition

#if (KEY1_PART == 1)
    #define KEY1_PORT           PIND         // This is the port where we read bits

    #define KEY1_MASK           0b11100000   // the keyboard task runs only on
                                             // those ports defined in the mask
                                             // Ports not in the mask are ignored

    #define KEY1_OFFSET         -5           // shift keystroke result with this number
                                             // here: shift down by 4

    #define KEY1_ACTIVE_TO_GND  TRUE         // TRUE: detect a keystroke if pin is grounded
                                             // FALSE: detect a keystroke if high is applied

    unsigned char key1_state;                // static - save the previous state of keys.

    unsigned char debounce1;                 // static - save a flag for "current in debounce"

    signed char last_key_time1;              // last timerval
#endif

// --------------- Port2:
#define KEY2_PART           0            // if 1: use this definition

#if (KEY2_PART == 1)
    #define KEY2_PORT           PIND         // This is the port where we read bits

    #define KEY2_MASK           0b00001111   // the keyboard task runs only on
                                             // those ports defined in the mask
                                             // Ports not in the mask are ignored

    #define KEY2_OFFSET         0            // shift keystroke result with this number

    #define KEY2_ACTIVE_TO_GND  TRUE         // TRUE: detect a keystroke if pin is grounded
                                             // FALSE: detect a keystroke if high is applied

    unsigned char key2_state;                // static - save the previous state of keys.

    unsigned char debounce2;                 // static - save a flag for "current in debounce"

    signed char last_key_time2;              // last timerval
#endif
//------------------------------------------------------------------------------
// CONFIG

#define KEYBUF_SIZE       8       // type ahead buffer

//------------------------------------------------------------------------------
// DATA

// from cortos:
// uint8_t key_fifo_buf[KEYBUF_SIZE];
// cr_fifo_t key_fifo;


const unsigned char key_code_table[] PROGMEM = {
    // Scan line 0
    KEY_DOWN0,          // Bit 0
    KEY_DOWN1,          // Bit 1
    KEY_DOWN2,          // Bit 2
    KEY_DOWN3,          // Bit 3
    KEY_DOWN4,          // Bit 4
    KEY_DOWN5,          // Bit 5
    KEY_DOWN6,          // Bit 6
    KEY_DOWN7,          // Bit 7
    // Scan line 1
    KEY_DOWN8,          // Bit 0
    KEY_DOWN9,          // Bit 1
    KEY_DOWN10,          // Bit 2
    KEY_DOWN11,          // Bit 3
    KEY_DOWN12,          // Bit 4
    KEY_DOWN13,          // Bit 5
    KEY_DOWN14,          // Bit 6
    KEY_DOWN15,          // Bit 7
    // Scan line 0
    KEY_UP0,          // Bit 0
    KEY_UP1,          // Bit 1
    KEY_UP2,          // Bit 2
    KEY_UP3,          // Bit 3
    KEY_UP4,          // Bit 4
    KEY_UP5,          // Bit 5
    KEY_UP6,          // Bit 6
    KEY_UP7,          // Bit 7
    // Scan line 1
    KEY_UP8,          // Bit 0
    KEY_UP9,          // Bit 1
    KEY_UP10,          // Bit 2
    KEY_UP11,          // Bit 3
    KEY_UP12,          // Bit 4
    KEY_UP13,          // Bit 5
    KEY_UP14,          // Bit 6
    KEY_UP15,          // Bit 7
  };

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Code:

void init_keyboard(void)
  {
    debounce1 = 0;
    #if (KEY1_ACTIVE_TO_GND == TRUE)
        key1_state = ~KEY1_PORT;  
    #else
        key1_state = KEY1_PORT;  
    #endif
    
    #if (KEY2_PART == 1)
        debounce2 = 0;
        #if (KEY2_ACTIVE_TO_GND == TRUE)
            key2_state = ~KEY2_PORT;  
        #else
            key2_state = KEY2_PORT;  
        #endif
    #endif
    cr_fifo_init (&key_fifo, key_fifo_buf, sizeof(key_fifo_buf));
  }


t_cr_task keyboard_task(void)
  {
    unsigned char keys, xor;

    #if (KEY1_ACTIVE_TO_GND == TRUE)
        keys = ~KEY1_PORT;  
    #else
        keys = KEY1_PORT;  
    #endif
    xor = keys ^ key1_state;
    xor = xor & KEY1_MASK;

    if (xor)
      {                         // something has changed
        unsigned char i, mask;

        mask = 1;
        for (i=0; i<8; i++)
          {
            if (xor & mask)
              {
                if (debounce1 & mask)
                  { // debounce over, now change state and report
                    debounce1 &= ~mask;
                    key1_state ^= mask;
                    if (keys & mask)
                      {
                        unsigned char key_code;
                        key_code = pgm_read_byte(&key_code_table[i]);       // pressed
                        cr_fifo_put(&key_fifo, key_code);
                      }
                    else
                      {
                        unsigned char key_code;
                        key_code = pgm_read_byte(&key_code_table[i+16]);    // release
                        cr_fifo_put(&key_fifo, key_code);
                      }
                  }
                else
                  { 
                    debounce1 |= mask;
                  }
              }
            mask = mask << 1;           
          }
      }
    else
      {
        debounce1 = 0;
      }

    #if (KEY2_PART == 1)
        #if (KEY2_ACTIVE_TO_GND == TRUE)
            keys = ~KEY2_PORT;  
        #else
            keys = KEY2_PORT;  
        #endif
        xor = keys ^ key2_state;
        xor = xor & KEY2_MASK;

        if (xor)
          {                         // something has changed
            unsigned char i, mask;

            mask = 1;
            for (i=0; i<8; i++)
              {
                if (xor & mask)
                  {
                    if (debounce2 & mask)
                      { // debounce over, now change state and report
                        debounce2 &= ~mask;
                        key1_state ^= mask;

                        if (keys & mask)
                          {
                            unsigned char key_code;
                            key_code = pgm_read_byte(&key_code_table[i+8]);
                            cr_fifo_put(&key_fifo, key_code);
                          }
                        else
                          {
                            unsigned char key_code;
                            key_code = pgm_read_byte(&key_code_table[i+24]);    // release
                            cr_fifo_put(&key_fifo, key_code);
                          }
                      }
                    else
                      { 
                        debounce2 |= mask;
                      }
                  }
                mask = mask << 1;           
              }
          }
        else
          {
            debounce2 = 0;
          }
    #endif
    return(10000L / SYSTICK_PERIOD);
  }


unsigned char keystroke_present(void)
  {
    return (cr_fifo_filled(&key_fifo));
  }

unsigned char get_wait_keystroke(void)
  {
    while (!keystroke_present()) ;     // busy waiting! -> multitask
    return (cr_fifo_get_wait(&key_fifo));
  }


unsigned char query_key_state(unsigned char nr)
  {
    // Note: here the inputs are shifted, we have only two inputs at D6 and D7

    switch (nr) 
      {
        case 0:
            if (key1_state & (1<< 6)) return(1);
            else return(0);
        case 1:
            if (key1_state & (1<< 7)) return(1);
            else return(0);
        default:
            return(0);
      }  
  }

#endif // (KEYBOARD_ENABLED == TRUE)













