//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-------------------------------------------------------------------------------
//
// file:      led.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-04-14 V0.1 kw start
// 
//===============================================================================
//
// purpose:   powerful occupancy detector, 
//            sensing with 16-channel A/D-subsystem synced to DCC
//            programmable threshold, configurable memory time
//            full BiDi incl. polarity check
//
//===============================================================================
//
// content:   routines for LED control (Status LED)
//            (blinking is driven by CORTOS as scheduler)
//            
// Howto:     
//
//===============================================================================
//
// Config:    In this module we assume that LEDs are mapped to PORTD
//
// Note: when Ports are driven by interrupt (like BiDiB does), ATOMIC access here!


//===============================================================================

#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>

#include <util/atomic.h>

#include "config.h"
#include "hardware.h"
#include "cortos.h"

#include "timer.h"
#include "led.h"


//===============================================================================
// Globals and defines

#define LEDTICK_PERIOD  10000L
volatile struct
  {
    unsigned char rest;         // time until next change in Ticks (20ms)
    unsigned char state;        // 
    unsigned char bitmask;     // 
    unsigned char ontime;       // Einschaltzeit
    unsigned char offtime;      // Ausschaltzeit
    unsigned char pause;    
    unsigned char flashes;     // Anzahl Pulse
    unsigned char act_flash;    
  } led[NUM_OF_LED];



//===============================================================================
// runtime tasks

t_cr_task task_led_status(void)
  {
    unsigned char i, temp;

    for (i=0; i<sizeof(led)/sizeof(led[0]); i++)
      {
        if (led[i].rest != 0)
          {
            if (--led[i].rest == 0)
              {
                 if (led[i].state)
                  {
                    led[i].state = 0;
                    if (led[i].act_flash == led[i].flashes)
                      {
                        led[i].rest = led[i].pause;
                        led[i].act_flash = 0;
                      }
                    else
                      {
                        led[i].rest = led[i].offtime;
                      }
                    temp = led[i].bitmask;
                    ATOMIC_BLOCK(ATOMIC_FORCEON)
                      {
                        PORTD |= temp;
                      }
                  }
                else
                  {
                    led[i].state = 1;
                    led[i].act_flash++;
                    led[i].rest = led[i].ontime;
                    temp = led[i].bitmask;
                    ATOMIC_BLOCK(ATOMIC_FORCEON)
                      {
                        PORTD &= ~temp;
                      }
                  }
              }
          }
      }
    return(LEDTICK_PERIOD / SYSTICK_PERIOD);
  }


void init_led_status(void)
  {
    led[LED_MSG].rest = 0;
    led[LED_PROG].rest = 0;

    led[LED_MSG].bitmask = (1 << LEDp_MSG);
    led[LED_PROG].bitmask = (1 << LEDp_PROG);
  }

 

//===============================================================================
// upstream interface

void led_on(unsigned char myled)
  {
    unsigned temp;
    led[myled].rest = 0;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

void led_half(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 3;
    led[myled].offtime = 3;  // 25% duty 
    led[myled].ontime  = 1;  
    led[myled].rest    = 1;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

// make a single fast pulse, then dark
void led_fast_pulse(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 0;
    led[myled].flashes = 0;
    led[myled].pause   =      0L / LEDTICK_PERIOD;
    led[myled].offtime =      0L / LEDTICK_PERIOD;
    led[myled].ontime  =      0L / LEDTICK_PERIOD;
    led[myled].rest    = 100000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

// make a single fast pulse (and return to previous state)
void led_highlight(unsigned char myled)
  {
    unsigned temp;
    if (led[myled].rest)  // if anything is active - add an extended active state
      {
        led[myled].rest   =  300000L / LEDTICK_PERIOD;
        led[myled].state = 1;
        temp = led[myled].bitmask;
        ATOMIC_BLOCK(ATOMIC_FORCEON)
          {
            PORTD &= ~temp;
          }
      }
    else led_fast_pulse(myled);
  }


void led_off(unsigned char myled)
  {
    unsigned temp;
    led[myled].rest   =  0;
    led[myled].state = 0;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD |= temp;
      }
  }


// make a single pulse
void led_pulse(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 0;
    led[myled].flashes = 0;
    led[myled].pause   =      0L / LEDTICK_PERIOD;
    led[myled].offtime =      0L / LEDTICK_PERIOD;
    led[myled].ontime  =      0L / LEDTICK_PERIOD;
    led[myled].rest    = 500000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }

  }

// make a single pulse, dimmed
void led_half_pulse(unsigned char myled)
  {
    unsigned temp;
    if (led[myled].rest) return;   // prevent reentry
    
    led[myled].act_flash = 1;      // ev. vorziehen
    led[myled].flashes = 120000L / 4 / LEDTICK_PERIOD;   // use the flasher to determine time;
    led[myled].pause   = 0;
    led[myled].offtime = 3; // 25% duty cycle
    led[myled].ontime  = 1;
    led[myled].rest    = 1;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }



// make a double pulse
void led_double_pulse(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 2;
    led[myled].pause   =      0L / LEDTICK_PERIOD;
    led[myled].offtime = 300000L / LEDTICK_PERIOD;
    led[myled].ontime  = 100000L / LEDTICK_PERIOD;
    led[myled].rest    = 100000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

// slow blinking
void led_blinking(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 600000L / LEDTICK_PERIOD;
    led[myled].offtime = 600000L / LEDTICK_PERIOD;
    led[myled].ontime  = 400000L / LEDTICK_PERIOD;
    led[myled].rest    = 400000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

// fast blinking
void led_fast_blinking(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 100000L / LEDTICK_PERIOD;
    led[myled].offtime = 100000L / LEDTICK_PERIOD;
    led[myled].ontime  =  30000L / LEDTICK_PERIOD;
    led[myled].rest    =  30000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }


void led_XoooXooo(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 500000L / LEDTICK_PERIOD;
    led[myled].offtime = 500000L / LEDTICK_PERIOD;
    led[myled].ontime  = 200000L / LEDTICK_PERIOD;
    led[myled].rest    = 200000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

void led_XooooXoooo(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 650000L / LEDTICK_PERIOD;
    led[myled].offtime = 650000L / LEDTICK_PERIOD;
    led[myled].ontime  =  30000L / LEDTICK_PERIOD;
    led[myled].rest    =  30000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

void led_XoooooXooooo(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 1000000L / LEDTICK_PERIOD;
    led[myled].offtime = 1000000L / LEDTICK_PERIOD;
    led[myled].ontime  =   30000L / LEDTICK_PERIOD;
    led[myled].rest    =   30000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

void led_oXXXoXXX(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 200000L / LEDTICK_PERIOD;
    led[myled].offtime = 200000L / LEDTICK_PERIOD;
    led[myled].ontime  = 500000L / LEDTICK_PERIOD;
    led[myled].rest    = 500000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

void led_oXXXXXoXXXXX(unsigned char myled)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = 1;
    led[myled].pause   = 100000L / LEDTICK_PERIOD;
    led[myled].offtime = 100000L / LEDTICK_PERIOD;
    led[myled].ontime  = 600000L / LEDTICK_PERIOD;
    led[myled].rest    = 600000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }



// make a series of flashes, then a longer pause
void led_fast_flash(unsigned char myled, unsigned char count)
  {
    unsigned temp;
    led[myled].act_flash = 1;
    led[myled].flashes = count;
    led[myled].pause   = 700000L / LEDTICK_PERIOD;
    led[myled].offtime = 240000L / LEDTICK_PERIOD;
    led[myled].ontime  = 120000L / LEDTICK_PERIOD;
    led[myled].rest    = 120000L / LEDTICK_PERIOD;
    led[myled].state = 1;
    temp = led[myled].bitmask;
    ATOMIC_BLOCK(ATOMIC_FORCEON)
      {
        PORTD &= ~temp;
      }
  }

