//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-------------------------------------------------------------------------------
//
// file:      lights.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-04-14 V0.01 kw start
//            2012-02-12 V0.02 kw ported to lightControl
//            2012-04-12 V0.03 kw ported to mobalist
// 
//===============================================================================
//
// purpose:   powerful light controller
//
//===============================================================================
//
// content:   routines for LED control (Outputport LED)
//            (we need CORTOS as scheduler)
//            Dimming, Blinking ist done here
//            
// Howto:     1. State-Engine, which controls states (On/off, number off flashes)
//            2. Dimmengine, which tries to fullfill states with the given
//               ramps and brightness values.
//
//===============================================================================
//
// Config:    In this module we assume that LEDs are connected to TLC5941
//
//===============================================================================


#include "config.h"
#include "hardware.h"
#include "cortos.h"

#include "prbs8.h"                  // random numbers

#include "timer.h"
#include "lights.h"

#include "pwm.h"                    // control block


#include "bidib_messages.h"         // defines for state control

#define LIGHT_STATE_A 0
#define LIGHT_STATE_B 1
#define FLICKER_TO_A  2


// define some handy delta values

#define DELTA_FAST    100
#define DELTA_200ms   (100 / (200000L / LIGHT_TICK_PERIOD))  // ramp up in 200ms
#define DELTA_400ms   (100 / (400000L / LIGHT_TICK_PERIOD))  // ramp up in 400ms
#define DELTA_1s      (100 / (1000000L / LIGHT_TICK_PERIOD))  // ramp up in 1s

#if (DELTA_1s == 0)
  #warning severe DELTA_1s is too small - use a larger LIGHT_TICK_PERIOD
#endif

//===============================================================================

t_light_cfg light_cfg[NUM_OF_LIGHTS] EEMEM =
  {
    //  brightness_off, brightness_on, dimm_off, dimm_on 
    {        0,              255,           2,       6 },   // 0
    {        0,              255,           2,       6 },   // 1
    {        0,              255,           2,       6 },   // 2
    {        0,              255,           2,       6 },   // 3
    {        0,              255,           2,       6 },   // 4
    {        0,              255,           2,       6 },   // 5
    {        0,              255,           2,       6 },   // 6
    {        0,              255,           2,       6 },   // 7
    {        0,              255,           2,       6 },   // 8
    {        0,              255,           2,       6 },   // 9
    {        0,              255,           2,       6 },   // 10
    {        0,              255,           2,       6 },   // 11
    {        0,              255,           2,       6 },   // 12
    {        0,              255,           2,       6 },   // 13
    {        0,              255,           2,       6 },   // 14
    {        0,              255,           2,       6 },   // 15
    {        0,              255,           2,       6 },   // 16
    {        0,              255,           2,       6 },   // 17
    {        0,              255,           2,       6 },   // 18
    {        0,              255,           2,       6 },   // 19
    {        0,              255,           2,       6 },   // 20
    {        0,              255,           2,       6 },   // 21
    {        0,              255,           2,       6 },   // 22
    {        0,              255,           2,       6 },   // 23
  };

t_light light[NUM_OF_LIGHTS];      // 32


//===============================================================================

// run_lights implements the state engine for external leds
//  1: check out whether we should approch a new lightval
//  2: check blinking and state changes
//  3. load new values to tlc5941_grayscales
//  return code: desired sleep time in systicks


t_cr_task run_lights(void)
  {
    unsigned char i;

    for (i=0; i<NUM_OF_LIGHTS; i++)
      {
        // part 1: check state changes

        if (light[i].rest != 0)
          {
            if (--light[i].rest == 0)
              {
                 switch(light[i].state)
                  {
                    default:
                    case LIGHT_STATE_A:   // 'on' Value
                        if (light[i].act_flash == light[i].flashes)
                          {
                            light[i].rest = light[i].pause;
                            light[i].act_flash = 0;
                          }
                        else
                          {
                            light[i].rest = light[i].time_B;
                          }
                        light[i].state = LIGHT_STATE_B;
                        light[i].target = light[i].light_B_val;
                        light[i].delta = light[i].delta_B;
                        break;
                    case LIGHT_STATE_B:
                        light[i].act_flash++;
                        light[i].rest = light[i].time_A;
                        
                        light[i].state = LIGHT_STATE_A;
                        light[i].target = light[i].light_A_val;
                        light[i].delta = light[i].delta_A;
                        break;
                    case FLICKER_TO_A:
                        if (light[i].act_flash !=0) 
                          {
                            light[i].act_flash = light[i].act_flash >> 1;  // load next bit
                            if (light[i].act_flash == 0) 
                              {
                                // all flicker bits done - turn on port (do not load .rest)
                                light[i].target = light[i].light_A_val;
                                light[i].delta = light[i].delta_A;
                              }
                            else
                              {
                                if (light[i].act_flash & 0x01)
                                  { 
                                    light[i].rest = light[i].time_A;
                                    light[i].target = light[i].light_A_val;
                                    light[i].delta = light[i].delta_A;
                                  }
                                else
                                  { // new delay = random extra + time_B
                                    light[i].rest = (light[i].act_flash & 0x07) + light[i].time_B;
                                    light[i].target = light[i].light_B_val;
                                    light[i].delta = light[i].delta_B;
                                  }
                              }
                          }
                        break;
                  }
              }
          }

        // part 2: check dimming
        
        unsigned char my_bright;
        
        my_bright = light[i].volume;                            // aktueller Wert
   
        if (my_bright == light[i].target);                      // passt
        else if (my_bright > light[i].target)
          {                                                     // wir muessen runter
            if (light[i].target + light[i].delta > my_bright)
              {
                light[i].volume = light[i].target;
                // grayscales[i] = light[i].volume;
              }
            else
              {
                light[i].volume = my_bright - light[i].delta;
                // grayscales[i] = light[i].volume;
              }
          }
        else 
          {                                                     // wir muessen hoch
            if (my_bright + light[i].delta >= light[i].target)
              {
                light[i].volume = light[i].target;
                // grayscales[i] = light[i].volume;
              }
            else
              {
                light[i].volume = my_bright + light[i].delta;
                // grayscales[i] = light[i].volume;
              }
          }
      }
    pwm_update();   // neuen Interruptvektor berechnen
    semaphor_set(C_PWM_Update);   // copy anfordern
    return(LIGHT_TICK_PERIOD / SYSTICK_PERIOD);
  }


void start_light(unsigned char i)
  {
    if (light[i].state == LIGHT_STATE_A)
      {
        light[i].volume = light[i].light_A_val +1;  // force update
        light[i].target = light[i].light_A_val;
        light[i].delta = light[i].delta_A;
      }
    else 
      {
        light[i].volume = light[i].light_B_val +1;
        light[i].target = light[i].light_B_val;
        light[i].delta = light[i].delta_B;
      }
  }


// the following single led controls only use the dimm engine,
// not the state engine

void turn_light_on_soft(unsigned char my_light)
  {
    light[my_light].rest = 0;
    light[my_light].target = FULL_BRIGHT;
    if (light[my_light].volume == FULL_BRIGHT) light[my_light].volume--;
    light[my_light].delta = DELTA_1s;
  }



void turn_light_on_fast(unsigned char my_light)
  {
    light[my_light].rest = 0;
    light[my_light].target = FULL_BRIGHT;
    if (light[my_light].volume == FULL_BRIGHT) light[my_light].volume--;
    light[my_light].delta = DELTA_FAST;
  }


void turn_light_off_soft(unsigned char my_light)
  {
    light[my_light].rest = 0;
    light[my_light].target = 0;
    if (light[my_light].volume == 0) light[my_light].volume = 1;
    light[my_light].delta = DELTA_1s;
  }

void turn_light_off_fast(unsigned char my_light)
  {
    light[my_light].rest = 0;
    light[my_light].target = 0;
    if (light[my_light].volume == 0) light[my_light].volume = 1;
    light[my_light].delta = DELTA_FAST;
  }



#define BLINK_SOFT  (1000000L / LIGHT_TICK_PERIOD)
#define BLINK_FAST  (300000L / LIGHT_TICK_PERIOD)
#define BLINK       (500000L / LIGHT_TICK_PERIOD)


const t_light DAT_light_blink_soft PROGMEM = 
  {
        LIGHT_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_SOFT,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_SOFT,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_SOFT,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        120,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        10,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_SOFT,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        5,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        5,              // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };

void blink_light_soft(unsigned char my_light)
  {
    memcpy_P (&light[my_light], &DAT_light_blink_soft, sizeof(DAT_light_blink_soft));
  }

const t_light DAT_light_blink_fast PROGMEM = 
  {
        LIGHT_STATE_A,    // unsigned char state;        // Aktueller Status (0 = phase A, 1 = phase B)
        BLINK_FAST,     // unsigned char rest;         // Zeit bis zum wechsel in Ticks (20ms)
        101,            // unsigned char volume;       // aktueller Istwert
        100,            // unsigned char target;       // aktueller Zielwert
        1,              // unsigned char delta;        // aktuelle Schrittweite
        
        0,              // unsigned char flashes;      // Anzahl Pulse (wie oft wird A angefahren)
        0,              // unsigned char act_flash;    // Counter dazu
        BLINK_FAST,     // unsigned char pause;        // wenn != 0: pulse repetierend
    
        BLINK_FAST,     // unsigned char time_A;       // Zeit in phase A (Einschaltzeit) 
        120,            // unsigned char light_A_val;  // aktueller Zielwert in Richtung A
        34,             // unsigned char delta_A;      // aktuelles Increment in Richtung A

        BLINK_FAST,     // unsigned char time_B;       // Zeit in phase B (Ausschaltzeit)
        0,              // unsigned char light_B_val;  // aktueller Zielwert in Richtung B
        34,             // unsigned char delta_B;      // aktuelles Decrement in Richtung B
   };

void blink_light_red_fast(unsigned char my_light)
  {
    memcpy_P (&light[my_light], &DAT_light_blink_fast, sizeof(DAT_light_blink_fast));
  }



void blink_light_fast(unsigned char my_light)
  {
    memcpy_P (&light[my_light], &DAT_light_blink_fast, sizeof(DAT_light_blink_fast));
  }



void init_lights(void)
  {
    unsigned char i;

    for(i = 0; i < NUM_OF_LIGHTS; i++) turn_light_off_fast(i);  

    for(i = 0; i < NUM_OF_LIGHTS; i++)
      {
        light[i].light_B_val = eeprom_read_byte(&light_cfg[i].brightness_off) / 2;
        light[i].light_A_val = eeprom_read_byte(&light_cfg[i].brightness_on) / 2;
        light[i].delta_B     = eeprom_read_byte(&light_cfg[i].dimm_off);
        light[i].delta_A     = eeprom_read_byte(&light_cfg[i].dimm_on);
      }

/*
    light[2].delta_A = 127;
    light[2].delta_B = 127;
    
    light[3].delta_A = 12;
    light[3].delta_B = 4;
    light[4].delta_A = 12;
    light[4].delta_B = 4;
    
    light[5].delta_A = 127;
    light[5].delta_B = 127;
    light[6].delta_A = 127;
    light[6].delta_B = 127;
    light[7].delta_A = 127;
    light[7].delta_B = 127;
*/

  }


//===============================================================================
//
// BiDiB-Calls
//
//===============================================================================
//

void set_light(unsigned char cmd, unsigned char port)
  {
    if (port >= sizeof(light)/sizeof(light[0])) return; // out of range

    switch(cmd)
      {
        case BIDIB_PORT_TURN_OFF:
            light[port].rest = 0;
            light[port].delta = DELTA_FAST;
            light[port].target = light[port].light_B_val;
            break;
        case BIDIB_PORT_DIMM_OFF:
            light[port].rest = 0;
            light[port].delta = light[port].delta_B;
            light[port].target = light[port].light_B_val;
            break;
        case BIDIB_PORT_TURN_ON:
            light[port].rest = 0;
            light[port].delta = DELTA_FAST;
            light[port].target = light[port].light_A_val;
            break;
        case BIDIB_PORT_DIMM_ON:
            light[port].rest = 0;
            light[port].delta = light[port].delta_A;
            light[port].target = light[port].light_A_val;
            break;
        case BIDIB_PORT_TURN_ON_NEON:
            light[port].state = FLICKER_TO_A;
            light[port].act_flash = prbs8();
            light[port].rest = prbs8() & 0x0f;
            light[port].rest += 1;
            light[port].time_A = 20500L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 20500L / LIGHT_TICK_PERIOD ;
            light[port].pause = 0 ;
            break;
        case BIDIB_PORT_BLINK_A:
            light[port].state = LIGHT_STATE_A;
            light[port].rest = 500000L / LIGHT_TICK_PERIOD ;
            light[port].target = light[port].light_A_val;
            light[port].delta = light[port].delta_A;
            light[port].time_A = 500000L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 500000L / LIGHT_TICK_PERIOD ;
            light[port].pause = 500000L / LIGHT_TICK_PERIOD ;
            break;
        case BIDIB_PORT_BLINK_B:
            light[port].state = LIGHT_STATE_B;
            light[port].rest = 500000L / LIGHT_TICK_PERIOD ;
            light[port].target = light[port].light_B_val;;
            light[port].delta = light[port].delta_B;
            light[port].flashes = 1;
            light[port].time_A = 500000L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 500000L / LIGHT_TICK_PERIOD ;
            light[port].pause = 500000L / LIGHT_TICK_PERIOD ;
            break;
        case BIDIB_PORT_FLASH_A:
            light[port].state = LIGHT_STATE_A;
            light[port].rest = 250000L / LIGHT_TICK_PERIOD ;
            light[port].target = light[port].light_A_val;;
            light[port].delta = light[port].delta_A;
            light[port].flashes = 1;
            light[port].time_A = 250000L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 250000L / LIGHT_TICK_PERIOD ;
            light[port].pause = 250000L / LIGHT_TICK_PERIOD ;
            break;
        case BIDIB_PORT_FLASH_B:
            light[port].state = LIGHT_STATE_B;
            light[port].rest = 250000L / LIGHT_TICK_PERIOD ;
            light[port].target = light[port].light_B_val;
            light[port].delta = light[port].delta_B;
            light[port].flashes = 1;
            light[port].time_A = 250000L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 250000L / LIGHT_TICK_PERIOD ;
            light[port].pause = 250000L / LIGHT_TICK_PERIOD ;
            break;
        case BIDIB_PORT_DOUBLE_FLASH:
            light[port].state = LIGHT_STATE_A;
            light[port].rest = 31000L / LIGHT_TICK_PERIOD ;
            light[port].target = light[port].light_A_val;
            light[port].delta = light[port].delta_A;
            light[port].flashes = 2;
            light[port].time_A = 31000L / LIGHT_TICK_PERIOD ;
            light[port].time_B = 122000L / LIGHT_TICK_PERIOD ;
            light[port].pause = 500000L / LIGHT_TICK_PERIOD ;
            break;
        default:
            break;
      }
  }

// Die Helligkeit wird in einem Range von 0..255 bergeben, wir
// haben intern aber nur 128 Stufen, also werden die Werte hier halbiert.

void config_light(t_bidib_lport_cfg* newcfg)
  {
    if (newcfg->portnum >= sizeof(light)/sizeof(light[0])) return; // out of range

    if (newcfg->brightness_off != light[newcfg->portnum].light_B_val)
      {
        eeprom_write_byte(&light_cfg[newcfg->portnum].brightness_off, newcfg->brightness_off);
        light[newcfg->portnum].light_B_val = newcfg->brightness_off / 2;
      }
    if (newcfg->brightness_on != light[newcfg->portnum].light_A_val)
      {
        eeprom_write_byte(&light_cfg[newcfg->portnum].brightness_on, newcfg->brightness_on);
        light[newcfg->portnum].light_A_val = newcfg->brightness_on / 2;
      }
    if (newcfg->dimm_off != light[newcfg->portnum].delta_B)
      {
        light[newcfg->portnum].delta_B = newcfg->dimm_off;
        eeprom_write_byte(&light_cfg[newcfg->portnum].dimm_off, newcfg->dimm_off);
      }
    if (newcfg->dimm_on != light[newcfg->portnum].delta_A)
      {
        light[newcfg->portnum].delta_A = newcfg->dimm_on;
        eeprom_write_byte(&light_cfg[newcfg->portnum].dimm_on, newcfg->dimm_on);
      }
  }

// return: 0: out of range
//         1: struct is filled

unsigned char config_light_query(t_bidib_lport_cfg* newcfg)
  {
    unsigned char portnum = newcfg->portnum;
    
    if (portnum >= sizeof(light)/sizeof(light[0])) return(0); // out of range

    newcfg->brightness_off = light[portnum].light_B_val * 2;
    newcfg->brightness_on = light[portnum].light_A_val * 2;
    newcfg->dimm_off = light[portnum].delta_B;
    newcfg->dimm_on = light[portnum].delta_A;
    return(1);
  }
