//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-------------------------------------------------------------------------------
//
// file:      macro.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-04-14 V0.01 kw start
//            2012-02-12 V0.02 kw ported to lightControl
//            2012-04-19 V0.04 kw start macro neu mit flag
//            2012-05-08 V0.05 kw changed to new bidib implementation
//            2012-07-30 V0.06 kw BIDIB_MACRO_RESTORE
//            2012-10-16 V0.07 kw BIDIB_ACCESSORY
//            2012-10-18 V0.08 kw bugfix bei tcode
//            2013-09-29 V0.11 kw added get_port_stat, added new macro codes
//            2013-10-10 V0.12 kw changes bidib_do_accessory
// 
//===============================================================================
//
// purpose:   powerful light controller
//
//===============================================================================
//
// content:   routines for Macro control (BiDiB Level 2)
//            (we need CORTOS as scheduler)
//            
// Howto:     
//===============================================================================

#include "config.h"
#include "hardware.h"
#include "cortos.h"
#include "prbs8.h"


#include "timer.h"
#include "lights.h"
#include "macro.h"
#include "bidib_messages.h"     // define for macro handling
#include "bidib_client_parser.h"    // bidib_* (for notify)

#define MACRO_BASE_TICK   20500L   // = 20500us = 20ms - see www.bidib.org

#define TCODE_MASK 0b11000000
#define TCODE_MIN 0
#define TCODE_HR  0b10000000
#define TCODE_DAY 0b01000000
#define TCODE_MIN_VOID  (TCODE_MIN | 0x3F)
#define TCODE_HR_VOID   (TCODE_HR  | 0x3F)
#define TCODE_DAY_VOID  (TCODE_DAY | 0x3F)

//===============================================================================
// in ANSI C wird bei unions nur das grosste Element definiert, 
// darum hier der shift bei den Nibbles noetig

#if (MOBALIST_MODEL == STANDARD)
    #include "default_macro_model_standard.h"
#elif  (MOBALIST_MODEL == TURNOUT)
    #include "default_macro_model_turnout.h"
#else
    #error MOBALIST_MODEL is void
#endif




t_macro macro[NUM_OF_MACROS];
/*

typedef struct
  {
    uint8_t slow_down;              // this is the ratio between this cam and the cam tick.
    struct
      {
        uint8_t delay;
        uint8_t port;
        uint8_t command;
      } item[NUM_OF_ITEMS_PER_MACRO];
  } t_macro;
*/

typedef struct
  {
    struct
      {
        uint8_t active:1;           // 0=stopped, 1=cam is active
        uint8_t sysstate:1;         // 1=macro is in a sys state waiting for something
        uint8_t start_pending:1;    // 1=(re-)start is requested
        uint8_t stop_pending:1;     // 1=stop soon
        uint8_t stop_disabled:1;    // 1=do not stop me (crtical section)
      };
    uint8_t  slow_down;            // ratio between systick and our tick
    uint8_t  slow_down_cnt;        // the running var associated 
    uint8_t  local_time;           // local time counter (between positions)
    uint8_t  position;             // pointer to next position
    uint8_t  randomwait;           // if should wait for a random time - this is the time
    uint8_t  repeat;
    uint8_t  notify_accessory;     // dieser (virtuelle) AccessoryDekoder wird benachrichtigt. (255 = void)
  } t_ctrl;

t_ctrl ctrl[NUM_OF_MACROS];

// Access-Routines


    static unsigned char read_delay(unsigned char c)
      {
        return(macro[c].item[ctrl[c].position].delay);
      }
    static t_lstate read_lstate(unsigned char c)
      {
        return(macro[c].item[ctrl[c].position].lstate);
      }
    static unsigned char read_lvalue(unsigned char c)
      {
        return(macro[c].item[ctrl[c].position].lvalue);
      }
    static unsigned char read_slow_down(unsigned char c)
      {
        return(macro[c].slow_down);
      }
    static unsigned char read_repeat(unsigned char c)
      {
        return(macro[c].repeat);
      }


//--------------------------------------------------------------------------------------------
// actions (downstream stuff)
//--------------------------------------------------------------------------------------------


// returns state of input (no debouncing)
// TRUE: key is pressed (connected GND)
uint8_t check_input(uint8_t number)
  {
    /*
    uint8_t mask;
    mask = (1 << (uint8_t)(number & 0x07));     // clamp to 8 bits
    if (PIND & mask) return(0);
    else return(1);
    */
    // here only two bits:
    if (number == 0)
      {
        if (JUMPER5_FITTED()) return (1);
        else return(0);
      }
    else if (number == 1)
      {
        if (JUMPER6_FITTED()) return (1);
        else return(0);
      }
    return(0);
  }


// parse a IO-command
void xperform(t_lstate lstate, uint8_t lvalue)
  {
    switch(lstate.type)
      {
        case BIDIB_OUTTYPE_SPORT:               // standard port
            break;            
        case BIDIB_OUTTYPE_LPORT:
            set_light(lstate.cmd, lvalue);      // 1.. 24 Light
            break;            
        case BIDIB_OUTTYPE_SERVO:
            break;
        case BIDIB_OUTTYPE_SOUND:
            break;         
        case BIDIB_OUTTYPE_MOTOR:
            break;            
        case BIDIB_OUTTYPE_ANALOG:
            break;            
        default:
            break;
      }
  }


// converter             from                 to
void convert_bidib2macro(t_bidib_port* bd_lc, t_macro_ctrl* m_lc)
  {
     m_lc->lstate.type = bd_lc->type;
     switch(bd_lc->type)
      {
        case BIDIB_OUTTYPE_SPORT:               // standard port
            m_lc->lstate.cmd = bd_lc->portstat;
            m_lc->lvalue = bd_lc->portnum;
            break;
            
        case BIDIB_OUTTYPE_LPORT:               // light port
            m_lc->lstate.cmd = bd_lc->portstat;
            m_lc->lvalue = bd_lc->portnum;
            break;
            
        case BIDIB_OUTTYPE_SERVO:
            m_lc->lstate.cmd = bd_lc->portnum;
            m_lc->lvalue = bd_lc->portstat;
            break;
            
        case BIDIB_OUTTYPE_SOUND:
            m_lc->lstate.cmd = bd_lc->portnum;
            m_lc->lvalue = bd_lc->portstat;
            break;
            
        case BIDIB_OUTTYPE_MOTOR:
            m_lc->lstate.cmd = bd_lc->portnum;
            m_lc->lvalue = bd_lc->portstat;
            break;
            
        case BIDIB_OUTTYPE_ANALOG:
            m_lc->lstate.cmd = bd_lc->portnum;
            m_lc->lvalue = bd_lc->portstat;
            break;
            
        default:
            m_lc->lstate.cmd = 0xF;
            m_lc->lvalue = 0xFF;
            break;
      }
  }

// converter
void convert_macro2bidib( t_macro_ctrl* m_lc, t_bidib_port* bd_lc)
  {
     bd_lc->type = m_lc->lstate.type;
     switch(bd_lc->type)
      {
        case BIDIB_OUTTYPE_SPORT:               // standard port
            bd_lc->portstat = m_lc->lstate.cmd;
            bd_lc->portnum = m_lc->lvalue;
            break;
        case BIDIB_OUTTYPE_LPORT:
            bd_lc->portstat = m_lc->lstate.cmd;
            bd_lc->portnum = m_lc->lvalue;
            break;
        case BIDIB_OUTTYPE_SERVO:
            bd_lc->portnum = m_lc->lstate.cmd;
            bd_lc->portstat = m_lc->lvalue;
            break;
        case BIDIB_OUTTYPE_SOUND:
            bd_lc->portnum = m_lc->lstate.cmd;
            bd_lc->portstat = m_lc->lvalue;
            break;
        case BIDIB_OUTTYPE_MOTOR:
            bd_lc->portnum = m_lc->lstate.cmd;
            bd_lc->portstat = m_lc->lvalue;
            break;
        case BIDIB_OUTTYPE_ANALOG:
            bd_lc->portnum = m_lc->lstate.cmd;
            bd_lc->portstat = m_lc->lvalue;
            break;
        default:
            bd_lc->portstat = 0xFF;
            bd_lc->portnum = 0xFF;
            break;
      }
  }




//--------------------------------------------------------------------------------------------
// sequencer tools
//--------------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------------
// message system (16 semaphores)

unsigned int message_array; 

// set message bit
void set_message(uint8_t number)
  {
    uint16_t mask;
    mask = (1 << (uint8_t)(number & 0x0F));
    message_array |= mask;
  }

// clear message bit
void clear_message(uint8_t number)
  {
    uint16_t mask;
    mask = (1 << (uint8_t)(number & 0x0F));
    message_array &= ~mask;
  }


// returns state of message bit
uint8_t check_message(uint8_t number)
  {
    uint16_t mask;
    mask = (1 << (uint8_t)(number & 0x0F));
    if(message_array & mask) return(1);
    else return(0);
  }

//-------------------------------------------------------------------------------------
// this moves the macro.position forward until either:
// - the sequence is waiting for an event
// - the sequence is terminated
// - the sequence is in normal time delay
// Warning: the state engine could get in an endless loop (and stack overflow), if
//          a sequence is restarted with zero delay at the beginning
//          (so we block out sequence_start if there is no local_time)

unsigned char start_sequence(uint8_t c);

void advance_macro(uint8_t c)
  {
    while(1)
      {
        ctrl[c].position++;

        if (ctrl[c].position == NUM_OF_ITEMS_PER_MACRO)
          {
            // end reached, stop this cam
            ctrl[c].active = 0;
            return; 
          }
    
        // now check this position

        uint8_t delay = read_delay(c);
        t_lstate lstate = read_lstate(c);
        uint8_t lvalue = read_lvalue(c);
        
        switch(delay)
          {
            case 0:
                xperform(lstate, read_lvalue(c));
                // no return here, we stay in the loop and advance macro again
                break;
            case 1 ... 254:
                // we are in a normal position, and we have to wait (slow down)
                return;
                break;
            case 255:       // this is a sys command, no delay
                if (ctrl[c].stop_pending && !ctrl[c].stop_disabled )
                  {
                    ctrl[c].active = 0;     // somebody want us to stop, so we do it
                    return; 
                  }
                switch(lstate.byte)
                  {
                    case BIDIB_MSYS_ACC_OKAY_NF:
                            if (ctrl[c].notify_accessory !=255)
                              {
                                bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                           BIDIB_ACC_STATE_NO_FB_AVAILABLE + BIDIB_ACC_STATE_DONE,
                                                           0);
                                ctrl[c].notify_accessory = 255;
                              }
                            break;
                    case BIDIB_MSYS_ACC_OKAY_QIN0:
                            #if (NUM_OF_INPUTS > 0)             // wenn keine Inputs, dann keine Abfrage
                                if ((check_input(lvalue))) 
                                  {
                                    if (ctrl[c].notify_accessory !=255)
                                      {
                                        bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                                   BIDIB_ACC_STATE_ERROR,
                                                                   BIDIB_ACC_STATE_ERROR_POSITION);
                                        ctrl[c].notify_accessory = 255;
                                      }
                                  }
                                else
                            #endif
                                  {
                                    if (ctrl[c].notify_accessory !=255)
                                      {
                                        bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                                   BIDIB_ACC_STATE_DONE,
                                                                   0);
                                        ctrl[c].notify_accessory = 255;
                                      }
                                  }
                            // no return, advance again
                            break;
                    case BIDIB_MSYS_ACC_OKAY_QIN1:
                            #if (NUM_OF_INPUTS > 0)             // wenn keine Inputs, dann keine Abfrage
                                if (!(check_input(lvalue))) 
                                  {
                                    if (ctrl[c].notify_accessory !=255)
                                      {
                                        bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                                   BIDIB_ACC_STATE_ERROR,
                                                                   BIDIB_ACC_STATE_ERROR_POSITION);
                                        ctrl[c].notify_accessory = 255;
                                      }
                                  }
                                else
                            #endif
                                  {
                                    if (ctrl[c].notify_accessory !=255)
                                      {
                                        bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                                   BIDIB_ACC_STATE_DONE,
                                                                   0);
                                        ctrl[c].notify_accessory = 255;
                                      }
                                  }
                            // no return, advance again
                            break;
                    case BIDIB_MSYS_DELAY_FIXED:
                            ctrl[c].randomwait = lvalue;    // load the delay
                            ctrl[c].sysstate = 1;
                            return;
                    case BIDIB_MSYS_DELAY_RANDOM:
                            ctrl[c].randomwait = prbs8() % lvalue;   // get a random number, limit to lvalue
                            ctrl[c].sysstate = 1;
                            return;
                    case BIDIB_MSYS_INPUT_QUERY0:
                            ctrl[c].sysstate = 1;
                            return;
                            break;
                    case BIDIB_MSYS_INPUT_QUERY1:
                            ctrl[c].sysstate = 1; 
                            return;
                            break;
                    case BIDIB_MSYS_FLAG_CLEAR:
                            clear_message(lvalue);
                            break;
                    case BIDIB_MSYS_FLAG_SET:
                            set_message(lvalue);
                            break;
                    case BIDIB_MSYS_FLAG_QUERY:
                            ctrl[c].sysstate = 1;
                            return;
                            break;
                    case BIDIB_MSYS_END_CRITCAL:
                            ctrl[c].stop_disabled = 0;
                            break;
                    case BIDIB_MSYS_BEGIN_CRITCAL:
                            ctrl[c].stop_disabled = 1;
                            break;
                    case BIDIB_MSYS_STOP_MACRO:
                            if ((lvalue) < NUM_OF_MACROS)
                              {
                                ctrl[lvalue].stop_pending = 1; 
                              }
                            // no return, advance again
                            break;
                    case BIDIB_MSYS_START_MACRO:
                            if ((lvalue) == c)
                              {
                                // start our own sequence again - so first make it inactive here
                                ctrl[c].active = 0;
                                start_sequence(c);
                                return;         // and stop processing           
                              }
                            start_sequence(lvalue);
                            // no return, advance again
                            break;
                    case BIDIB_MSYS_END_OF_MACRO:
                            if (ctrl[c].repeat)
                              {
                                ctrl[c].repeat--;
                                if (ctrl[c].repeat)
                                  {
                                    ctrl[c].position = -1;       // give a offset of -1
                                    // no return, advance again ->  also restart (seamless)
                                  }
                                else
                                  {
                                    ctrl[c].active = 0;     // stop it
                                    // Comment: Kufer, 04.10.2013
                                    // we should removed here, there is an extra command to do it
                                    if (ctrl[c].notify_accessory !=255)
                                      {
                                        bidib_send_accessory_state(ctrl[c].notify_accessory, 
                                                           BIDIB_ACC_STATE_NO_FB_AVAILABLE + BIDIB_ACC_STATE_DONE,
                                                           0);
                                        ctrl[c].notify_accessory = 255;
                                      }
                                    return;                 // exit here
                                  }
                              }
                            else  // if repeat is 0 -> repeat forever
                              {
                                ctrl[c].active = 0;     // stop it anyway
                                                        // das brauchen wir, damit das nicht
                                                        // in eine Endlos-Loop laufen kann.
                                                        // d.h. hier mind. ein Step Pause
                                start_sequence(c);
                                return;                 // exit here, terminate processing
                              }
                            break;
                    default:
                            break; 
                  }
                break;
          }
      }
  }

unsigned char start_sequence(uint8_t c)
  {
    if (c >= NUM_OF_MACROS) return(BIDIB_MACRO_NOTEXIST);
    if (ctrl[c].active) return(BIDIB_MACRO_RUNNING);  // already started

    ctrl[c].start_pending = 1;
    ctrl[c].notify_accessory = 255;
    return(BIDIB_MACRO_START);
  }

//--------------------------------------------------------------------------------------------
// sequencer task (to be called periodically every tick)
//--------------------------------------------------------------------------------------------

t_cr_task run_macro(void)
  {
    uint8_t c;
    for (c = 0; c < NUM_OF_MACROS; c++)
      {
        if (ctrl[c].start_pending)
          {
            ctrl[c].start_pending = 0;
            ctrl[c].active = 1; 
            ctrl[c].sysstate = 0;
            ctrl[c].stop_pending = 0;
            ctrl[c].stop_disabled = 0;
            ctrl[c].slow_down = read_slow_down(c);
            ctrl[c].repeat = read_repeat(c);
            ctrl[c].slow_down_cnt = ctrl[c].slow_down;
            ctrl[c].local_time = 0;
            ctrl[c].position = -1;       // give a offset of -1

            // now check for init states
            advance_macro(c);
          }
        if (ctrl[c].active)
          {
            if (ctrl[c].stop_pending && !ctrl[c].stop_disabled )
              {
                ctrl[c].active = 0;     // somebody want us to stop, so we do it
              }
            else if (ctrl[c].sysstate)
              {
                t_lstate lstate = read_lstate(c);
                uint8_t delay = read_delay(c);
                uint8_t lvalue = read_lvalue(c);

                switch(delay)
                  {
                    case 0 ... 254: break;
                    case 255:
                        switch(lstate.byte)
                          {
                            case BIDIB_MSYS_ACC_OKAY_NF:
                                    break;
                            case BIDIB_MSYS_ACC_OKAY_QIN0:          // notify acc, no delay
                                    break;
                            case BIDIB_MSYS_ACC_OKAY_QIN1:          // notify acc, no delay
                                    break;
                            case BIDIB_MSYS_DELAY_FIXED:
                            case BIDIB_MSYS_DELAY_RANDOM:
                                    if (ctrl[c].slow_down_cnt) ctrl[c].slow_down_cnt--;
                                    if (ctrl[c].slow_down_cnt == 0)                 // perform one tick
                                      {
                                        ctrl[c].slow_down_cnt = ctrl[c].slow_down;
                                        
                                        if (ctrl[c].local_time == ctrl[c].randomwait)
                                          {
                                            ctrl[c].local_time = 0;
                                            // random delay over
                                            ctrl[c].sysstate = 0;
                                            advance_macro(c);
                                          }
                                        else
                                          {
                                          	ctrl[c].local_time++;
                                          }
                                      }
                                    break;
                            case BIDIB_MSYS_INPUT_QUERY0:
                                    if (!(check_input(lvalue))) 
                                      {
                                        ctrl[c].sysstate = 0;
                                        advance_macro(c);
                                      }                    
                                    break;
                            case BIDIB_MSYS_INPUT_QUERY1:
                                    if ((check_input(lvalue))) 
                                      {
                                        ctrl[c].sysstate = 0;
                                        advance_macro(c);
                                      }                    
                                    break;
                            case BIDIB_MSYS_FLAG_CLEAR:
                                    break;
                            case BIDIB_MSYS_FLAG_SET:
                                    break;
                            case BIDIB_MSYS_FLAG_QUERY:
                                    if (check_message(lvalue)) 
                                      {
                                        ctrl[c].sysstate = 0;
                                        advance_macro(c);
                                      }                                         
                                    break;
                            case BIDIB_MSYS_END_CRITCAL:
                                    break;
                            case BIDIB_MSYS_BEGIN_CRITCAL:
                                    break;
                            case BIDIB_MSYS_STOP_MACRO:
                                    break;
                            case BIDIB_MSYS_START_MACRO:
                                    break;
                            case BIDIB_MSYS_END_OF_MACRO:
                                    break;
                            default:
                                    break;
                          }
                        break;
                  }
              }
            else   // normal operation
              {
                if (ctrl[c].slow_down_cnt) ctrl[c].slow_down_cnt--;
                if (ctrl[c].slow_down_cnt == 0)                 // perform one tick
                  {
                    ctrl[c].slow_down_cnt = ctrl[c].slow_down;
                    ctrl[c].local_time++;
                    if (ctrl[c].local_time == read_delay(c))
                      {
                        ctrl[c].local_time = 0;
                        // target reached
                        t_lstate lstate = read_lstate(c);
                        xperform(lstate, read_lvalue(c));
                        advance_macro(c);
                      }
                  }
              }
          }
      }
    return(MACRO_BASE_TICK / SYSTICK_PERIOD);       // ready again after 20ms
  }

static void restore_macro(unsigned char m)
  {
    unsigned char i;

    macro[m].slow_down = eeprom_read_byte(&macro_save[m].slow_down);
    macro[m].repeat = eeprom_read_byte(&macro_save[m].repeat);
    macro[m].tcode_min = eeprom_read_byte(&macro_save[m].tcode_min);
    macro[m].tcode_hr = eeprom_read_byte(&macro_save[m].tcode_hr);
    macro[m].tcode_day = eeprom_read_byte(&macro_save[m].tcode_day);
    for(i = 0; i < NUM_OF_ITEMS_PER_MACRO; i++)
      {
        macro[m].item[i].delay = eeprom_read_byte(&macro_save[m].item[i].delay);
        macro[m].item[i].lvalue = eeprom_read_byte(&macro_save[m].item[i].lvalue);
        macro[m].item[i].lstate.byte = eeprom_read_byte(&macro_save[m].item[i].lstate.byte);
      }
  }

void init_macro(void)
  {
    unsigned char m;

    for(m = 0; m < NUM_OF_MACROS; m++)
      {
        restore_macro(m);
      }
    for(m = 0; m < NUM_OF_MACROS; m++)
      {
        ctrl[m].notify_accessory = 255;     // clear all notify
      }
  }

static void save_macro(unsigned char m)
  {
    unsigned char i;

    if (macro[m].slow_down != eeprom_read_byte(&macro_save[m].slow_down))
        eeprom_write_byte(&macro_save[m].slow_down, macro[m].slow_down);

    if (macro[m].repeat != eeprom_read_byte(&macro_save[m].repeat))
        eeprom_write_byte(&macro_save[m].repeat, macro[m].repeat);

    if (macro[m].tcode_min != eeprom_read_byte(&macro_save[m].tcode_min))
        eeprom_write_byte(&macro_save[m].tcode_min, macro[m].tcode_min);

    if (macro[m].tcode_hr != eeprom_read_byte(&macro_save[m].tcode_hr))
        eeprom_write_byte(&macro_save[m].tcode_hr, macro[m].tcode_hr);

    if (macro[m].tcode_day != eeprom_read_byte(&macro_save[m].tcode_day))
        eeprom_write_byte(&macro_save[m].tcode_day, macro[m].tcode_day);

    for(i = 0; i < NUM_OF_ITEMS_PER_MACRO; i++)
      {
        if (macro[m].item[i].delay != eeprom_read_byte(&macro_save[m].item[i].delay))
            eeprom_write_byte(&macro_save[m].item[i].delay, macro[m].item[i].delay);

        if (macro[m].item[i].lvalue != eeprom_read_byte(&macro_save[m].item[i].lvalue))
            eeprom_write_byte(&macro_save[m].item[i].lvalue, macro[m].item[i].lvalue);

        if (macro[m].item[i].lstate.byte != eeprom_read_byte(&macro_save[m].item[i].lstate.byte))
            eeprom_write_byte(&macro_save[m].item[i].lstate.byte, macro[m].item[i].lstate.byte);
      }

  }
    


//===============================================================================
//
//  Config and Execute Interface for BiDiB
//
//===============================================================================


void macro_set_point(t_macro_point_cfg* newpoint)
  {
    unsigned char macro_i,i,delay;
    macro_i = newpoint->macro_i;
    i = newpoint->index;
    if (macro_i >= NUM_OF_MACROS)   return; // out of range
    if (i >= NUM_OF_ITEMS_PER_MACRO)   return; // out of range

    delay = newpoint->delay;
    macro[macro_i].item[i].delay = delay;

    if (delay == 255)
      {
        macro[macro_i].item[i].lstate.byte = newpoint->port.type;
        macro[macro_i].item[i].lvalue = newpoint->port.portnum;
      }
    else
      {
        convert_bidib2macro(&newpoint->port, 
                           (t_macro_ctrl*) &macro[macro_i].item[i].lstate);
      }
    // macro[newpoint->macro_i].item[newpoint->index].lvalue = newpoint->lvalue;
    // macro[newpoint->macro_i].item[newpoint->index].lstate = newpoint->lstate;
  }

unsigned char macro_get_point(t_macro_point_cfg* newpoint)
  {
    unsigned char macro_i,i,delay;
    macro_i = newpoint->macro_i;
    i = newpoint->index;
    if ((macro_i >= NUM_OF_MACROS) || (i >= NUM_OF_ITEMS_PER_MACRO))
      {
        newpoint->delay = 255;
        newpoint->port.type = 255;
        return(0); // out of range
      }

    delay = macro[macro_i].item[newpoint->index].delay;
    newpoint->delay = delay;

    if (delay == 255)
      {
        newpoint->port.type = macro[macro_i].item[i].lstate.byte;
        newpoint->port.portnum = macro[macro_i].item[i].lvalue;
        newpoint->port.portstat = 0;   // default
      }
    else
      {
        convert_macro2bidib((t_macro_ctrl*) &macro[macro_i].item[i].lstate, &newpoint->port);
        //newpoint->lvalue = macro[newpoint->macro_i].item[newpoint->index].lvalue;
        //newpoint->lstate = macro[newpoint->macro_i].item[newpoint->index].lstate;
      }
    return(1);
  }

unsigned char macro_handle(uint8_t macro_i, uint8_t opcode)
  {
    unsigned char i;
    
    if (macro_i >= NUM_OF_MACROS) return(BIDIB_MACRO_NOTEXIST);
    switch(opcode)
      {
        case BIDIB_MACRO_OFF:
            ctrl[macro_i].active = 0;
            return(BIDIB_MACRO_OFF);
            break;
        case BIDIB_MACRO_START:
            return(start_sequence(macro_i));
            break;
        case BIDIB_MACRO_RESTORE:
            restore_macro(macro_i);
            return(BIDIB_MACRO_RESTORE);
            break;
        case BIDIB_MACRO_SAVE:
            save_macro(macro_i);
            return(BIDIB_MACRO_SAVE);
            break;
        case BIDIB_MACRO_DELETE:
            ctrl[macro_i].active = 0;
            for (i=0; i<NUM_OF_ITEMS_PER_MACRO; i++)
              {
                macro[macro_i].item[i].delay = 255;
                macro[macro_i].item[i].lstate.byte = 255;
              }
            return(BIDIB_MACRO_DELETE);
            break;
        default:
            return(BIDIB_MACRO_NOTEXIST);
            break;
      }
  }

static unsigned char _macro_match_min(uint8_t condition, uint8_t tcode)
  {
    if ((condition & TCODE_MASK) != TCODE_MIN) return(0);   // failed, no min trigger
    if (condition == tcode) return (1);                      // match
    if (condition == TCODE_MIN + 60) return (1);            // trigger every minute
    if (condition == TCODE_MIN + 61)
      {
        if (tcode == TCODE_MIN + 0) return (1); 
        if (tcode == TCODE_MIN + 30) return (1); 
      }
    if (condition == TCODE_MIN + 62)
      {
        if (tcode == TCODE_MIN + 0) return (1); 
        if (tcode == TCODE_MIN + 15) return (1); 
        if (tcode == TCODE_MIN + 30) return (1); 
        if (tcode == TCODE_MIN + 45) return (1); 
      }
    return(0);
  }
    
static unsigned char _macro_match_hr(uint8_t condition, uint8_t tcode)
  {
    if ((condition & TCODE_MASK) != TCODE_HR) return(0);   // failed, no hr trigger
    if (condition == tcode) return (1);                      // match
    if (condition == TCODE_HR + 24) return (1);            // trigger every hour
    if (condition == TCODE_HR + 25)
      {
        if ((tcode >= TCODE_HR + 8) && (tcode <= TCODE_HR + 17)) return (1); 
      }
    return(0);
  }

static unsigned char _macro_match_day(uint8_t condition, uint8_t tcode)
  {
    if ((condition & TCODE_MASK) != TCODE_DAY) return(0);   // failed, no day trigger
    if (condition == tcode) return (1);                      // match
    if (condition == TCODE_DAY + 7) return (1);             // trigger every day
    return(0);
  }

void macro_eval_timecode(uint8_t* tcode)
  {
    uint8_t m;

    for (m = 0; m < NUM_OF_MACROS; m++)
      {
        if (  _macro_match_min(macro[m].tcode_min, tcode[0]) 
           && _macro_match_hr(macro[m].tcode_hr, tcode[1]) 
           && _macro_match_day(macro[m].tcode_day, tcode[2]) )
          {
            (start_sequence(m));
          }
      }
  }

unsigned char macro_para_set(t_macro_para* newpara)
  {
    uint8_t i;
    if (newpara->macro_i >= NUM_OF_MACROS)
      {
        newpara->value.as_uint8[0] = 255;
        newpara->value.as_uint8[1] = 255;
        newpara->value.as_uint8[2] = 255;
        newpara->value.as_uint8[3] = 255;
        return(255);
      }
    switch(newpara->index)
      {
        case BIDIB_MACRO_PARA_SLOWDOWN:
            macro[newpara->macro_i].slow_down = newpara->value.as_uint8[0];
            return(0);
            break;
        case BIDIB_MACRO_PARA_REPEAT:
            macro[newpara->macro_i].repeat = newpara->value.as_uint8[0];
            return(0);
            break;
        case BIDIB_MACRO_PARA_START_CLK:
            for (i=0; i<4; i++)
              {
                switch(newpara->value.as_uint8[i] & 0b11000000)
                  {
                    case 0b00000000:
                        macro[newpara->macro_i].tcode_min = newpara->value.as_uint8[i];
                        break;
                    case 0b10000000:
                        macro[newpara->macro_i].tcode_hr = newpara->value.as_uint8[i];
                        break;
                    case 0b01000000:
                        macro[newpara->macro_i].tcode_day = newpara->value.as_uint8[i];
                        break;
                    default:
                        break;
                  }
              }
            return(0);
            break;
        default:
            newpara->value.as_uint8[0] = 255;
            newpara->value.as_uint8[1] = 255;
            newpara->value.as_uint8[2] = 255;
            newpara->value.as_uint8[3] = 255;
            return(255);
            break;
      }    
  }

unsigned char macro_para_get(t_macro_para* newpara)
  {
    if (newpara->macro_i >= NUM_OF_MACROS)
      {
        newpara->value.as_uint8[0] = 255;
        newpara->value.as_uint8[1] = 255;
        newpara->value.as_uint8[2] = 255;
        newpara->value.as_uint8[3] = 255;
        return(255);
      }
    switch(newpara->index)
      {
        case BIDIB_MACRO_PARA_SLOWDOWN:
            newpara->value.as_uint8[0] = macro[newpara->macro_i].slow_down;
            newpara->value.as_uint8[1] = 0;
            newpara->value.as_uint8[2] = 0;
            newpara->value.as_uint8[3] = 0;
            return(0);
            break;
        case BIDIB_MACRO_PARA_REPEAT:
            newpara->value.as_uint8[0] = macro[newpara->macro_i].repeat;
            newpara->value.as_uint8[1] = 0;
            newpara->value.as_uint8[2] = 0;
            newpara->value.as_uint8[3] = 0;
            return(0);
            break;
        case BIDIB_MACRO_PARA_START_CLK:
            newpara->value.as_uint8[0] = macro[newpara->macro_i].tcode_min;
            newpara->value.as_uint8[1] = macro[newpara->macro_i].tcode_hr;
            newpara->value.as_uint8[2] = macro[newpara->macro_i].tcode_day;
            newpara->value.as_uint8[3] = 255;
            return(0);
            break;
        default:
            newpara->value.as_uint8[0] = 255;
            newpara->value.as_uint8[1] = 255;
            newpara->value.as_uint8[2] = 255;
            newpara->value.as_uint8[3] = 255;
            return(255);
            break;
      }    
  }

//===============================================================================
//
//  Accessory Frontend
//
//===============================================================================

static unsigned int eval_macro_exec_time(unsigned char m)
  {
    unsigned char  retval,i = 0;
    unsigned int ex_time = 0;
    unsigned long ltemp;

    if (macro[m].repeat != 1) return(0);        // keine Aussage mglich

    while(i < NUM_OF_ITEMS_PER_MACRO)
      {
        if (macro[m].item[i].delay == 255) break;  // Systemfunktion - abbruch
        ex_time += macro[m].item[i].delay;
        i++;
      }

    ex_time *= macro[m].slow_down;           // Einheit MACRO_BASE_TICK
                                             // angegeben in Ticks (20ms), etwa 1...500
    
    // umrechnen auf 0,1s / gemaess bidib/railcom spec, d.h. div 5; 0.2 = 51/256   
    ltemp = (unsigned long) ex_time * 51L; 
    ex_time = ltemp / 256L;

    if (ex_time > 1270) retval = 0xff;
    else if (ex_time > 127)
      {
        // umrechnen auf 1s
        ltemp = (unsigned long) ex_time * 26L; 
        ex_time = ltemp / 256L;
        retval = ex_time;
        retval |= 0x80;             // dann MSB setzen, als Zeichen fuer scale 1s
      }
    else
      {
        retval = ex_time;
      }
    return(retval);
  }


static unsigned char _get_accessory_total(unsigned char acc)
  {
    unsigned char i=0;
    while (i<NUM_OF_ASPECTS_PER_ACCESSORY)
      {
        if (CVram.accessory_map[acc].mapped_macro[i] == 255) return(i);
        i++;
      }
    return(i);
  }

void bidib_do_accessory(unsigned char acc_num, unsigned char aspect)
  {
    unsigned char macro_i, total, wait;
    
    if (acc_num >= NUM_OF_ACCESSORY) 
      {
        bidib_send_accessory_state(acc_num, BIDIB_ACC_STATE_ERROR, BIDIB_ACC_STATE_ERROR_VOID);
        return;
      }
    total = _get_accessory_total(acc_num);

    if (aspect >= total)
      {
        bidib_send_accessory_state(acc_num, BIDIB_ACC_STATE_ERROR, BIDIB_ACC_STATE_ERROR_VOID);
        return;
      }
    macro_i = CVram.accessory_map[acc_num].mapped_macro[aspect];
    
    if (macro_i >= NUM_OF_MACROS)
      {
        bidib_send_accessory_state(acc_num, BIDIB_ACC_STATE_ERROR, BIDIB_ACC_STATE_ERROR_VOID);
        return;
      }
    // ASPECT speichern
    CVram.accessory_map[acc_num].current_aspect = aspect;
    eeprom_write_byte(&CV.accessory_map[acc_num].current_aspect, aspect); 

    wait = eval_macro_exec_time(macro_i);
    if (wait)
      {
        if (start_sequence(macro_i) == BIDIB_MACRO_START)
          {
            ctrl[macro_i].notify_accessory = acc_num;
          }
        bidib_send_accessory_state(acc_num, BIDIB_ACC_STATE_WAIT, wait);
      }
    else
      {
        start_sequence(macro_i);
        bidib_send_accessory_state(acc_num, BIDIB_ACC_STATE_DONE, 0);
      }
  }

void init_accessory(void)
  {
    unsigned char acc;
    unsigned char aspect, macro_i;
    //  if (CVram.restore_accessory == 1)   // das nehmen wir hier immer an.
    if (1 == 1)
      {
        for (acc=0; acc < NUM_OF_ACCESSORY; acc++)
          {
            aspect = CVram.accessory_map[acc].current_aspect;
            if (aspect < NUM_OF_ASPECTS_PER_ACCESSORY)
              {
                macro_i = CVram.accessory_map[acc].mapped_macro[aspect];
                start_sequence(macro_i);
              }
          }
      }
  }

#if (SIMULATION == 2323)

volatile int my_clock;

void test_sequenzer(void)
  {
    uint8_t i;
    start_sequence(0);
    for(i=0; i<10; i++)
      {
        PORTB = my_clock;
        sequencer();
        my_clock++;
      }

    start_sequence(1);
    for(i=0; i<10; i++)
      {
        PORTB = my_clock;
        sequencer();
        my_clock++;
      }
    
  }

#endif

