//==================================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2010 Wolfgang Kufer
//
//----------------------------------------------------------------------------------
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//----------------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2010-05-01 V0.01 kw started 
//
//----------------------------------------------------------------------------------
//
// purpose:   
//
// ====== >>>> config.h  is the central definition file for the project
//
//==================================================================================
//

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"                 // general definitions the decoder, cv's
#include "hardware.h"               // port definitions for target
#include "cortos.h"                 // cooperative scheduler and fifo
#include "timer.h"                  // systick and PWM value
#include "led.h"                    // state engine for LEDs (blinking, flashing ...)

#include "lights.h"                 // control engine for LEDs
#include "pwm.h"                    // pwm engine for LEDs
#include "macro.h"                  
//

#include "keyboard.h"               // only four keys (jumper)

#include "prbs8.h"                  // for Power Up Delay & Backoff
#include "bidib_client_if.h"        // BiDiB
#include "bidib_client_parser.h"    // BiDiB
#include "event_handler.h"          // we have only some internal event ..



#define SIMULATION  0               // 0: real application 
                                    // 1: skip intro
                                    // 2: test timing engine
                                    // 3: test action


#if (SIMULATION != 0)
   #warning SIMULATION is on!



#endif

// unsigned char bid_test[] = {0x0a, 0x02, 0x00, 0x62, 0x4b, 0x00, 0x03, 0x3f, 0xbf, 0x7f, 0xff, 0x3a};
// test_process_bidib_message(bid_test);
// PORTA = 0;





//===============================================================================
// Seriennummer
//
// Fuer BiDIB ist eine eindeutige Seriennummer erforderlich, diese wird an
// den beiden hchsten Bytes des EEPROM abgelegt. Das wird durch eine
// eigene section .ee_serial und entsprechende Linkeranweisungen festgelegt.
//
// Hierzu sind folgende Aenderungen erforderlich:
//
// Project/Makefile:
// Linker (add line): (assign address)
//
//    LDFLAGS += -Wl,-section-start=.ee_serial=0x8101fe
// 
// Locater (add line): (do not load to flash)
//
//    HEX_FLASH_FLAGS += -R .ee_serial
//
// In das eeprom wird diese section nicht aufgenommen, das wird separat beigestellt.
//
// Wenn man nur die Seriennummer erzeugen will, dann muss man:
//
//    HEX_EEPROM_FLAGS = -j .eeprom 
//
// durch folgende Zeile ersetzen:
//
//    HEX_EEPROM_FLAGS = -j .ee_serial
//
// Das erzeugte eep-file fr das eeprom sieht dann so aus:
//    :0201FE0055AA00
//
//-------------------------------------------------------------------------------


unsigned int e_serial[] __attribute__((section(".ee_serial"))) = {11};


int main(void)
  { 
    signed char main_systick;

    init_hardware();                        // setup hardware ports

    // get serial number from eeprom
    
    my_unique_id.product_serial = eeprom_read_word((unsigned int*)(EEPROM_SIZE-2));
    // my_unique_id.product_serial = eeprom_read_word(e_serial);
    

    if ((my_unique_id.product_serial == 0) || (my_unique_id.product_serial == 0xFFFF))
      {
        #if (SIMULATION == 0)
        while (1)
          { LED_MSG_ON(); LED_PROG_ON();
            _delay_ms(150);
            LED_MSG_OFF(); LED_PROG_OFF();
            _delay_ms(150);
          }
        #endif
      }

    uint16_t j; uint8_t *dest;
    dest = (uint8_t *) &CVram;
    for (j=0; j<sizeof(CV); j++)                // copy CVs to Mirror in RAM
      {
        uint8_t data;
        data = eeprom_read_byte((void *) j);
        *dest++ = data; 
      }

    LED_PROG_ON();
    prbs8();
    seed = seed ^ (unsigned char) my_unique_id.product_serial;
    prbs8();
    seed = seed ^ (unsigned char) (my_unique_id.product_serial >> 8);
    prbs8();
    seed = seed ^ (unsigned char) my_unique_id.product_id;
    prbs8();

    #if (SIMULATION == 0)
      {
        unsigned char i;
        for (i=0;i<seed;i++)
          {
            _delay_ms(1L);
          }
      }
    #endif    

    init_macro();                           // load makro from eeprom

    init_accessory();                       // start all current_aspects (if enabled)

    
    init_lights();                          // preload light configuration

    init_timer();                           // system timer (this is timer 2)

    pwm_init();

    init_led_status();                      // setup led states + backlight

    init_keyboard();                        // okay, not really a _keyboard_ :-)

    init_event_handler();

    init_bidib_client();                    // BiDiB Interface (Parser)

    
    led_half(LED_PROG);                    // up

    // macro_handle(0,BIDIB_MACRO_RUN); // testhalber

    //----------- at this point, all hardware init is done
    //            and we are up and ready to run

    main_systick = get_systick();

    sei();
    
    while(1)
      {
        if ((signed char)(get_systick() - main_systick) > 0)
          {
            main_systick = get_systick();
            // here we go with hook functions
            // _mydelay_us(10);
            // nop();
            
          }    
        scheduler();                        // run task engine (from cortos.c)
      }
  }


