//===============================================================================
//
// OpenDCC - BiDiB for MobaLiSt
//
// Copyright (c) 2012 Wolfgang Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-------------------------------------------------------------------------------
//
// file:      pwm.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2009-01-18 V0.01 kw  started
//            2012-10-30 V0.02 kw  Umbau auf Wechselpuffer (wegen Timing)
//            2013-10-22 V0.03 kw  bei COMPB TCNT=0 hinterher
//
//-------------------------------------------------------------------------------
//
// purpose:   LED controller with soft pwm
// content:   library_functions
//
//-------------------------------------------------------------------------------
//
// PWM: Timer1 luft von Null an hoch, dabei wird beim Start des Timers
//      auf allen Ports, die eine Helligkeit > 0 haben, eine 1 ausgegeben.
//      Mittels Compareregister A wird zu den jeweils vorberechneten Zeiten
//      ein neuer Bitvektor auf die Port geladen und so die PWM realisiert.
//      Ports, deren Zeit abgelaufen ist, werden wieder auf 0 geschaltet.
//
//      Das Ende (und Restart) wird hierbei mittels Compare B definiert, 
//      damit wird dann der Timer wieder auf Null gesetzt und die Laufregister
//      fr Compare A zurckgenommen.
//
//      Vorteile: geringe CPU Last, extrem schlanker Interrupt fr Compare A
//                und damit hohe Auflsung der PWM.
//      Nachteil: man mu ein Datenarray mit Vektoren fr Timer und Bits pflegen.
//
//      Hinweis:  Um auch beim Restart (und fallweisem Wechsel der Timingtabelle)
//                eine geringe Latenz zu haben, wird die Timingtabelle mittels
//                Wechselpuffer kontrolliert. Die jeweils aktive Tabelle wird
//                ber das Semaphor C_PWM_Buffer definiert.
//
//===============================================================================
//
// Doku zum Timing und Programmierung:
// - Limit: 16 Bit fr Compare:
//      Atmega luft hier mit 16MHz, wir benutzen Timer 1, Vorteiler 8.
//      Damit hat der Timer mit 65536 etwa 30 Hz minimale Wiederholrate.
//      (Aktuell ist Compare B bei 25000 angezogen, damit ergeben sich
//      80Hz Wiederholrate)
// 
// - Limit: Laufzeit der ISR:
//      Die ISR bentigt etwa 40 Takte, das sind bei 16MHz etwa 2,5us
//      D.h. das Delta zweier Helligkeitswerte darf diesen Wert nicht
//      unterschreiten.
// Gewhlt:
// Vorteiler 8
// Compare B: 25000 -> das ergibt eine Refreshrate von 80Hz
// Kurve: mixed exponentiell und linear, siehe Spreadsheet
//        grayscale and pwm-steps.xls; Werte: Y(i+1) = Y(i)*exp(0.0146)+1.5*i

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action
                        
#include "config.h"         
#include "cortos.h"         
#include "hardware.h"         
#include "pwm.h"         
#include "lights.h"              // get light[i].volume         




const unsigned int PROGMEM pwmstep[PWMSTEPS] =
  {
      0,        4,       8,        13,       19,       26,       34,       45,
     58,       72,       88,      106,      125,      147,      170,      195,
    222,      250,      281,      314,      348,      385,      424,      464,
    507,      552,      599,      649,      700,      754,      810,      868, 
    929,      992,     1058,     1126,     1197,     1270,     1345,     1424,
   1505,     1588,     1675,     1764,     1856,     1950,     2048,     2149, 
   2252,     2359,     2469,     2581,     2697,     2817,     2939,     3065, 
   3194,     3326,     3462,     3602,     3745,     3891,     4041,     4195,   
   4353,     4515,     4680,     4849,     5023,     5200,     5381,     5567,
   5757,     5951,     6150,     6353,     6560,     6772,     6989,     7210, 
   7436,     7667,     7903,     8143,     8389,     8640,     8896,     9157,
   9424,     9696,     9974,    10257,    10546,    10840,    11141,    11447, 
   11759,   12078,    12403,    12733,    13071,    13414,    13765,    14122,
   14485,   14856,    15233,    15618,    16010,    16409,    16815,    17229,
   17650,   18079,    18516,    18961,    19414,    19875,    20344,    20822,
   21308,   21803,    22307,    22819,    23341,    23871,    24412,    24961
  };

#define MAX_PWMSTEPS 24980
// 24980 * 500ns = 12490us = 12,5ms => 80 Hz

//========================================================================
// 2. Global Variables and types
//========================================================================

// This union allows to access 16 bits as word or as two bytes.
// This approach is (probably) more efficient than shifting.


// uint8_t grayscales[NUM_OF_LIGHTS];
// our target is: light[i].volume

typedef struct
  {
    unsigned int delay;
    unsigned char port1;
    unsigned char port2;
    unsigned char port3;
  } t_pwm_array;
 
// Wechselpuffer    
t_pwm_array  pwm_array0[NUM_OF_LIGHTS+1];            // one more -> Lattenzaun
t_pwm_array  pwm_array1[NUM_OF_LIGHTS+1];




//========================================================================
// 3. PWM Engine
//========================================================================

// a) simple engine with a lot CPU cycles (not used)

#if (1==0)
#define T_PWM 1000

static uint8_t pwm_cnt=0;
static uint8_t dim_cnt=0; 

ISR(TIMER1_COMPA_vect)   // old
  {

	OCR1A += (uint16_t)T_PWM;

    sei();

    // note PWM_PORT1 is invers order, layout!
	if (light[0].volume > pwm_cnt) PWM_PORT1 |= (1<<7);	// hat der Kanal den PWM-Wert erreicht?
    if (light[1].volume > pwm_cnt) PWM_PORT1 |= (1<<6);	// dann entsprechendes Bit setzen
    if (light[2].volume > pwm_cnt) PWM_PORT1 |= (1<<5);
    if (light[3].volume > pwm_cnt) PWM_PORT1 |= (1<<4);
    if (light[4].volume > pwm_cnt) PWM_PORT1 |= (1<<3);
    if (light[5].volume > pwm_cnt) PWM_PORT1 |= (1<<2);
    if (light[6].volume > pwm_cnt) PWM_PORT1 |= (1<<1);
    if (light[7].volume > pwm_cnt) PWM_PORT1 |= (1<<0);
	
    if (light[8].volume > pwm_cnt) PWM_PORT2 |= (1<<0);
    if (light[9].volume > pwm_cnt) PWM_PORT2 |= (1<<1);
    if (light[10].volume > pwm_cnt) PWM_PORT2 |= (1<<2);
    if (light[11].volume > pwm_cnt) PWM_PORT2 |= (1<<3);
    if (light[12].volume > pwm_cnt) PWM_PORT2 |= (1<<4);
    if (light[13].volume > pwm_cnt) PWM_PORT2 |= (1<<5);
    if (light[14].volume > pwm_cnt) PWM_PORT2 |= (1<<6);
    if (light[15].volume > pwm_cnt) PWM_PORT2 |= (1<<7);
    
    if (light[16].volume > pwm_cnt) PWM_PORT3 |= (1<<7);
    if (light[17].volume > pwm_cnt) PWM_PORT3 |= (1<<6);
    if (light[18].volume > pwm_cnt) PWM_PORT3 |= (1<<5);
    if (light[19].volume > pwm_cnt) PWM_PORT3 |= (1<<4);
    if (light[20].volume > pwm_cnt) PWM_PORT3 |= (1<<3);
    if (light[21].volume > pwm_cnt) PWM_PORT3 |= (1<<2);
    if (light[22].volume > pwm_cnt) PWM_PORT3 |= (1<<1);
    if (light[23].volume > pwm_cnt) PWM_PORT3 |= (1<<0);
	
    pwm_cnt++;
    if (pwm_cnt==PWM_STEPS)
      {
        pwm_cnt=0;
        PWM_PORT1 = 0;
        PWM_PORT2 = 0;
        PWM_PORT3 = 0;
      }
  }
#endif



// b) advanced transitional timing

// Dieser Pointer zeigt auf das nchste auszugebene OCR + Portpattern
// Geladen von: COMPB und Init
// Incrementiert durch COMPA; der berlauf wird durch COMPB verhindert.
// 
static t_pwm_array* isr_ptr;


ISR(TIMER1_COMPA_vect)
  {
    OCR1A = isr_ptr->delay;     // Naechsten Schaltpunkt laden
    PWM_PORT1 = isr_ptr->port1; // Ports ausgeben
    PWM_PORT2 = isr_ptr->port2;
    PWM_PORT3 = isr_ptr->port3;
    isr_ptr++;                  // auf naechste Transition stellen
  }

// COMPB macht reload der Daten und restart
// Umkopieren wuerde 70us dauern und damit das BiDiB-Timing zerstoeren
// Wir arbeiten daher mit Wechselpuffer
ISR(TIMER1_COMPB_vect) 
  {
    if (semaphor_query(C_PWM_Buffer))
      {
        OCR1A = pwm_array1[0].delay;     // und den Start einladen
        PWM_PORT1 = pwm_array1[0].port1;
        PWM_PORT2 = pwm_array1[0].port2;
        PWM_PORT3 = pwm_array1[0].port3;
        isr_ptr = &pwm_array1[1];        // der erste COMPA Interrupt soll da beginnen
        TCNT1 = 0;
      }
    else
      {
        OCR1A = pwm_array0[0].delay;     // und den Start einladen
        PWM_PORT1 = pwm_array0[0].port1;
        PWM_PORT2 = pwm_array0[0].port2;
        PWM_PORT3 = pwm_array0[0].port3;
        isr_ptr = &pwm_array0[1];        // der erste COMPA Interrupt soll da beginnen
        TCNT1 = 0;
      }
  }


#if (1==0)
ISR(TIMER1_COMPB_vect)  // alte Version, noch ohne Wechselpuffer
  {
    if (isr_semaphor_get(C_PWM_Update))
      {
        sei();
        unsigned char i;                // copy new data
        unsigned char* src = (unsigned char*) pwm_array;
        unsigned char* dest = (unsigned char*) isr_pwm_array;
            
        for (i=0; i<sizeof(pwm_array); i++)
          {
            *dest++ = *src++;
          }
      }
      
    sei();
    isr_ptr = &isr_pwm_array[1];        // der erste COMPA Interrupt soll da beginnen

    TCNT1 = 0;
    OCR1A = isr_pwm_array[0].delay;     // und den Start einladen
    PWM_PORT1 = isr_pwm_array[0].port1;
    PWM_PORT2 = isr_pwm_array[0].port2;
    PWM_PORT3 = isr_pwm_array[0].port3;
  }
#endif


//---------------------------------------------------------------------------------


#if (1==0)
// unser array durchsuchen und auf pwm_array umladen
// 
// note: PWM_PORT1 and PWM_PORT3 is invers order, layout!
//       we shift in invers order for these ports
// runtime: for 24 LEDs, this routine needs ~1ms on a 16MHz AVR.
void pwm_update_old(void)
  {
    uint8_t i, j;
    uint8_t min, new_min, mask;
    unsigned char register port;
    uint16_t min_delay;
 
    min = 0;                // mit min 0 beginnen, d.h. erst mal gucken, 
                            // gibt es channels, die nie leuchten?
    j = 0;

    while (min < 255)
      {
        min_delay = 0xFFFF;                 // das erst mal auf maximum
        port = 0;
        mask = 0x80;
        for(i=0; i < 8; i++)
          {
            if (light[i].volume > min) port |= mask;
            mask >>= 1;
          }
        pwm_array[j].port1 = port;

        port = 0;
        mask = 0x01;
        for(i=8; i < 16; i++)
          {
            if (light[i].volume > min) port |= mask;
            mask <<= 1;
          }
        pwm_array[j].port2 = port;

        port = 0;
        mask = 0x80;
        for(i=16; i < 24; i++)
          {
            if (light[i].volume > min) port |= mask;
            mask >>= 1;
          }
        pwm_array[j].port3 = port;    
        // nachste Schwelle suchen

        new_min = 255;
        for(i=0; i < (NUM_OF_LIGHTS); i++)
          {
            if (light[i].volume < new_min)
              {
                if (light[i].volume > min)
                  {
                    new_min = light[i].volume;
                    min_delay = pgm_read_word( &pwmstep[new_min] );
                  }
              }
          }
        min = new_min;
        pwm_array[j].delay = min_delay;  
        j++;
      }
  }
#endif
 
// -- die gleiche Routine nochmal, nur sind hier die Schleifen ausgewalzt:
//    --> Laufzeit sinkt auf 700us.
// unser array an grayscales (=light[i].volume) durchsuchen und 
// ein pwm_array erzeugen
// 
// note: PWM_PORT1 and PWM_PORT3 is invers order, layout!
//       we shift in invers order for these ports
// runtime: for 24 LEDs, this routine needs ~700us on a 16MHz AVR.
void pwm_update(void)
  {
    uint8_t i;
    uint8_t min, new_min;
    unsigned char register port;
    uint16_t min_delay;

    t_pwm_array* target_ptr;
 
    min = 0;                // mit min 0 beginnen, d.h. erst mal gucken, 
                            // gibt es channels, die nie leuchten?
    
    if (semaphor_query(C_PWM_Buffer))   target_ptr = &pwm_array0[0];    // wenn ISR auf 1 greift, dann ist 0 frei
    else                                target_ptr = &pwm_array1[0];

    while (min < 255)
      {
        min_delay = 0xFFFF;                 // das erst mal auf maximum
        
        port = 0;
        if (light[0].volume > min) port |= 0x80;
        if (light[1].volume > min) port |= 0x40;
        if (light[2].volume > min) port |= 0x20;
        if (light[3].volume > min) port |= 0x10;
        if (light[4].volume > min) port |= 0x08;
        if (light[5].volume > min) port |= 0x04;
        if (light[6].volume > min) port |= 0x02;
        if (light[7].volume > min) port |= 0x01;
        target_ptr->port1 = port;

        port = 0;
        if (light[8].volume > min) port |= 0x01;
        if (light[9].volume > min) port |= 0x02;
        if (light[10].volume > min) port |= 0x04;
        if (light[11].volume > min) port |= 0x08;
        if (light[12].volume > min) port |= 0x10;
        if (light[13].volume > min) port |= 0x20;
        if (light[14].volume > min) port |= 0x40;
        if (light[15].volume > min) port |= 0x80;
        target_ptr->port2 = port;

        port = 0;
        if (light[16].volume > min) port |= 0x80;
        if (light[17].volume > min) port |= 0x40;
        if (light[18].volume > min) port |= 0x20;
        if (light[19].volume > min) port |= 0x10;
        if (light[20].volume > min) port |= 0x08;
        if (light[21].volume > min) port |= 0x04;
        if (light[22].volume > min) port |= 0x02;
        if (light[23].volume > min) port |= 0x01;
        target_ptr->port3 = port;    
        
        // nachste Schwelle suchen
        new_min = 255;
        for(i=0; i < (NUM_OF_LIGHTS); i++)
          {
            if (light[i].volume < new_min)
              {
                if (light[i].volume > min)
                  {
                    new_min = light[i].volume;
                    min_delay = pgm_read_word( &pwmstep[new_min] );
                  }
              }
          }
        min = new_min;
        target_ptr->delay = min_delay;  
        target_ptr++;
      }
    //-- jetzt die ISR auf die neuen Daten umlenken

    if (semaphor_query(C_PWM_Buffer))   semaphor_get(C_PWM_Buffer);   // wenn das 1 war, jetzt 0
    else                                semaphor_set(C_PWM_Buffer); 


  }
 

void pwm_init(void)
  {
    unsigned char i;

    for ( i = 0; i < sizeof(light)/sizeof(light[0]); i++)
      {
        light[i].volume = 0;
      }

    pwm_update();   // recalc pwm array
    pwm_update();   // recalc pwm array ( zweimal, damit sind beide Buffer voll)

    // setup timer 1

    //---------------------------------------------------------------------------------

    // we run Timer1 with 16MHz, Prescaler 1: This is 62ns timer grid;
    // we run Timer1 with 16MHz, Prescaler 8: This is 500ns timer grid;


    #define T1_PRESCALER 8

    #if   (T1_PRESCALER==1)
      #define T1_PRESCALER_BITS   ((0<<CS12)|(0<<CS11)|(1<<CS10))
    #elif (T1_PRESCALER==8)
      #define T1_PRESCALER_BITS   ((0<<CS12)|(1<<CS11)|(0<<CS10))
    #elif (T1_PRESCALER==64)
      #define T1_PRESCALER_BITS   ((0<<CS12)|(1<<CS11)|(1<<CS10))
    #elif (T1_PRESCALER==256)
      #define T1_PRESCALER_BITS   ((1<<CS12)|(0<<CS11)|(0<<CS10))
    #elif (T1_PRESCALER==1024)
      #define T1_PRESCALER_BITS   ((1<<CS12)|(0<<CS11)|(1<<CS10))
    #else
      #error T1_PRESCALER is not defined or is not a supported value
    #endif

    // init_timer1(void)
    TCCR1A = (0 << COM1A1) | (0 << COM1A0)  // 00=normal port  01=normal port   10=clear on compare match, set at TOP   11 set on compare match
           | (0 << COM1B1) | (0 << COM1B0)
           | (0 << FOC1A) | (0 << FOC1B)
           | (0 << WGM11) | (0 << WGM10);

    TCCR1B = (0 << ICNC1)       // input noise canceler
           | (0 << ICES1)       // input cature edge
           | (0 << WGM13)
           | (0 << WGM12)
           | T1_PRESCALER_BITS;
    OCR1A = 0xFFFF;
    OCR1B = MAX_PWMSTEPS;

    TCNT1 = 0;
    OCR1A = pwm_array0[0].delay;     // und den Start einladen
    PWM_PORT1 = pwm_array0[0].port1;
    PWM_PORT2 = pwm_array0[0].port2;
    PWM_PORT3 = pwm_array0[0].port3;
    isr_ptr = &pwm_array0[1];        // der erste COMPA Interrupt soll da beginnen
    TCNT1 = 0;

#if (__AVR_ATmega32__)
    TIMSK |= (1<<OCIE1B);   											// Interrupt freischalten
#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)
    TIMSK1 |= (1<<OCIE1B);   											// Interrupt freischalten
#endif    TIMSK |= (1<<OCIE1A);
  }

