//==================================================================================
//
// OpenDCC - BiDiB for MobaLiSt - Timer Modul
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      timer.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-12-12 V0.01 kw started (copied from throttle project)
//            2009-07-29 V0.02 kw added measurement time
//            2012-04-10 V0.03 kw ported to mobalist
//
//------------------------------------------------------------------------
//
// purpose:   start up timers, supply systick
//
//==================================================================================
//
#include <inttypes.h>
#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "datatypes.h"
#include "config.h"
#include "timer.h"
// #include "hardware.h"

//---------------------------------------------------------------------------------
// assign timer resources
//
//          prescaler   tick    mode    interrupt   hardware    software
// timer0:   8          0.5us   STD     COMP        -           unused
// timer1:   8          0.5us   STD     COMP        -           LED-PWM output
// timer2:  64          4.0us   PWM     OVF         -           systick
//---------------------------------------------------------------------------------



#if   (T0_PRESCALER==1)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(0<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==8)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==64)
  #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(1<<CS00))
#elif (T0_PRESCALER==256)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(0<<CS00))
#elif (T0_PRESCALER==1024)
  #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(1<<CS00))
#else
  #error T0_PRESCALER is not defined or is not a supported value
#endif

#if (__AVR_ATmega32__)

void init_timer0(void)
  {
    TCCR0 = (0<<FOC0)   // reserved set to 0
          | (0<<WGM00)  // all normal mode (Top = 0xFF)
          | (0<<COM01)  // no output compare
          | (0<<COM00)
          | (0<<WGM01)
          | T0_PRESCALER_BITS;

    TCNT0 = 0;

    TIMSK =  (0<<TOIE0)
            | (0<<OCIE0);
  }

#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)

void init_timer0(void)
  {
        TCCR0A = (0<< COM0A1)   // 00 = normal port mode
               | (0<< COM0A0)
               | (0<< COM0B1)   // 00 = normal port mode
               | (0<< COM0B0)
               | (0<< WGM01)    // WGM = 000: normale mode
               | (0<< WGM00);
        TCCR0B = (0<< FOC0A)    // 0 = no forced compare match
               | (0<< FOC0B)    // 0 = no forced compare match
               | (0<< WGM02)    // 
               | T0_PRESCALER_BITS;

    TCNT0 = 0;

    TIMSK0 =  (0<<TOIE0)
            | (0<<OCIE0A);
  }

#else
    #error check processor include
#endif


//---------------------------------------------------------------------------------

 
//---------------------------------------------------------------------------
// Define the TIMER2, this is a 8 Bit Timer, Overflow at 256
//
//
// we run Timer2 with 16MHz, Prescaler 64: This is 4us timer grid;
// Complete cycle 1.6384ms -> Systick
// Knob and keyboard can hook to this tick
//
// #define T2_PRESCALER   64 (in timer.h)

/*
typedef union
{
    uint16_t as_uint16;
    int16_t  as_int16;
    uint8_t  as_uint8[2];
    int8_t   as_int8[2];
} t_data16;  */

volatile t_data16  tick;

// volatile signed char systick;           // gets incremented every 1.6ms
// volatile signed int  systick_i;         // gets incremented every 1.6ms
volatile signed int  systick_10ms;      // gets incremented every 10ms
                                        // atomic access required, range 0..655sec

//---------------------------------------------------------------------------

#if   (T2_PRESCALER==1)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==8)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==32)
  #define T2_PRESCALER_BITS   ((0<<CS22)|(1<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==64)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==128)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(0<<CS21)|(1<<CS20))
#elif (T2_PRESCALER==256)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(0<<CS20))
#elif (T2_PRESCALER==1024)
  #define T2_PRESCALER_BITS   ((1<<CS22)|(1<<CS21)|(1<<CS20))
#else
  #error T2_PRESCALER is not defined or is not a supported value
#endif


#if (__AVR_ATmega32__)

void init_timer2(void)
  {
    TCCR2 = (0<<FOC2)    // reserved set to 0
           | (1<<WGM20)    // Mode 3: Fast PWM (Top = 0xFF)
           | (0<<COM21)   // 1,0 Clear OC2 on Compare Match, set OC2 at TOP.
           | (0<<COM20)
           | (1<<WGM21)
           | T2_PRESCALER_BITS;

    TCNT2 = 0;

    TIMSK |= (1<<TOIE2);  // enable overflow interrupt

    // systick = 0;
    // systick_i = 0;
    systick_10ms = 0;

    tick.as_int16 = 0;
  }

#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)

void init_timer2(void)
  {
    TCCR2A = (0<< COM2A1)   // 00 = normal port mode
           | (0<< COM2A0)
           | (0<< COM2B1)   // 00 = normal port mode
           | (0<< COM2B0)
           | (1<< WGM21)    // WGM = 011: Fast PWM (Top = 0xFF)
           | (1<< WGM20);
    TCCR2B = (0<< FOC2A)    // 0 = no forced compare match
           | (0<< FOC2B)    // 0 = no forced compare match
           | (0<< WGM22)    // WGM = 011: Fast PWM (Top = 0xFF)
           | T2_PRESCALER_BITS;
    TCNT2=0x00;


    TIMSK2 |= (1<<TOIE2);  // enable overflow interrupt

    // systick = 0;
    // systick_i = 0;
    systick_10ms = 0;

    tick.as_int16 = 0;
  }

#else
    #error check processor include
#endif

signed char local_systick;

ISR(TIMER2_OVF_vect)
  {
    // We get an Interrupt every 1024us
    // We do a software extension to systick -> systick_10ms
    // 
    signed char compare;

    sei();
    tick.as_int16++;

    compare = tick.as_int8[0] - local_systick;
    if (compare >= 10)
      {
        systick_10ms++;
        local_systick += 10;
      }

        
    /*
    systick++;
    systick_i++;

    compare = systick - local_systick;
    if (compare >= 6)
      {
        systick_10ms++;
        local_systick += 6;
      }
    */
    // hook for keyboard
    // hook for rtos kernel
  }  


// this returns a 16 bit int, resolution 6,4us, total range ~400ms
// this is to be used for delay calculations
// here we access hardware (TCNT) and interrupt variable, so we
// must ensure the sync 

#if (__AVR_ATmega32__)
    #define my_TIFR   TIFR
#elif (__AVR_ATmega644P__ || __AVR_ATmega644__)
    #define my_TIFR   TIFR2
#else
    #error check processor include
#endif






signed int get_systick_us(void)
  {
    t_data16 result;
    signed char sw_tick;

    cli();
    sw_tick = tick.as_int8[0];
    result.as_uint8[0] = TCNT2;
    if (result.as_uint8[0] < 250)
      {  // low counter value, may be we missed a tick)
        if (my_TIFR & (1<<TOV2))  // pending overflow interrupt not yet processed
          sw_tick++;
      }
    result.as_int8[1] = sw_tick;
    sei();
    return(result.as_int16);
  }

signed int get_systick_10ms(void)
  {
    signed int mstick;

    cli();
    mstick = systick_10ms;       // atomic access
    sei();
    return(mstick);
  }

signed int get_systick_int16(void)
  {
    signed int stick;

    cli();
    stick = tick.as_int16;            // atomic access
    sei();
    return(stick);
  }

signed char get_systick(void)
  {
    signed char stick;

    cli();
    stick = tick.as_int8[0];            // atomic access
    sei();
    return(stick);
  }


void init_timer(void)
  {
    init_timer0(); 
    
    init_timer2();
  }



